/*******************************************************************************
 * This file is part of CSDS.
 * Copyright (c) 2021 Loic Hausammann (loic.hausammann@epfl.ch)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 ******************************************************************************/
/**
 * @brief This file contains functions for the cache.
 */
#ifndef CSDS_CACHE_H
#define CSDS_CACHE_H

/* Include config */
#include "../config.h"

/* Include CSDS headers */
#include "csds_error.h"
#include "csds_fields.h"
#include "csds_inline.h"
#include "csds_part_type.h"
#include "csds_particle.h"
#include "csds_tools.h"

/* Standard headers */
#include <stddef.h>
#include <stdint.h>
#include <stdlib.h>

/** @brief Structure for the cache.
 * It allows an efficient evolution of the particles.
 *
 * The pointers are always in this order:
 *  - The type of particles
 *  - The particle
 *
 * The index of a field (e.g. last.fields) is given by:
 * part_index * cache->last_time_offset + cache->offset[field_index]
 *
 * Each field contains: the field, its first derivative and its second derivative.
 * If the field does not have derivatives, they will simply be empty (size 0).
 */
struct csds_cache {
  /* The current index to use */
  size_t current_index[csds_type_count];

  /* The size of the arrays */
  int64_t capacity[csds_type_count];
  int64_t size[csds_type_count];

  /* Variables of the last record */
  struct {
    /* The fields read */
    char *fields[csds_type_count];

    /* The time of the record */
    double *time[csds_type_count];
  } last;

  /* Variables of the next record */
  struct {
    /* The fields read */
    char *fields[csds_type_count];

    /* The time of the record */
    double *time[csds_type_count];

    /* The offset of the record */
    size_t *offset[csds_type_count];
  } next;

  /* The list of fields contained in the cache. */
  struct field_information *fields[csds_type_count];

  /* Offset of the fields */
  int *offset;

  /* Total size of the fields */
  int total_size;

  /* The number of fields */
  int n_fields;

  /* Last time offset requested */
  size_t last_time_offset;
};

/** @brief Reset the current index */
INLINE static void csds_cache_reset_current_index(struct csds_cache *cache) {
  for (int i = 0; i < csds_type_count; i++) {
    cache->current_index[i] = 0;
  }
}

/** @brief Initialize an empty structure. */
INLINE static void csds_cache_empty_init(struct csds_cache *cache) {
  for (int i = 0; i < csds_type_count; i++) {
    cache->capacity[i] = 0;
    cache->size[i] = 0;
    cache->current_index[i] = 0;
    cache->last.fields[i] = NULL;
    cache->last.time[i] = NULL;
    cache->next.fields[i] = NULL;
    cache->next.time[i] = NULL;
    cache->next.offset[i] = NULL;
    cache->fields[i] = NULL;
  }

  cache->offset = NULL;
  cache->total_size = 0;
  cache->n_fields = 0;
  cache->last_time_offset = 0;
}

/**
 * @brief Initialize the structure.
 *
 * @param cache The #csds_cache
 * @param n_parts The number of particles to allocate
 * @param fields The fields that will be read
 * @param n_fields The size of fields
 * @param over_allocation The over allocation (required for creation / deletion
 * of particles)
 * @param h The logfile's header.
 * @param init_size The initial size of the arrays if no particles are present
 * @param time_offset The time offset of the current reading
 */
INLINE static void csds_cache_init(struct csds_cache *cache,
                                   const uint64_t *number_particles,
                                   const enum field_enum *fields, int n_fields,
                                   float over_allocation, struct header *h,
                                   int64_t *init_size, size_t time_offset) {

  cache->last_time_offset = time_offset;

  /* Compute the size of the arrays */
  for (int i = 0; i < csds_type_count; i++) {
    cache->capacity[i] = number_particles[i] == 0
                             ? init_size[i]
                             : number_particles[i] * over_allocation;
    cache->size[i] = 0;
  }

  /* Set the fields */
  cache->n_fields = n_fields;
  for (int part = 0; part < csds_type_count; part++) {

    /* Do we need the fields? */
    if (cache->capacity[part] == 0) {
      cache->fields[part] = NULL;
      continue;
    }

    /* Allocate the fields */
    cache->fields[part] = (struct field_information *)malloc(
        n_fields * sizeof(struct field_information));
    if (cache->fields[part] == NULL) error("Failed to allocate the fields");

    /* Copy the header's field into the cache */
    for (int field = 0; field < n_fields; field++) {
      const struct field_information *tmp = header_get_field_from_name(
          h, field_enum_get_name(fields[field]), (enum part_type)part);
      cache->fields[part][field] = *tmp;
    }
  }

  /* Allocate the offsets */
  cache->offset = (int *)malloc(n_fields * sizeof(int));
  if (cache->offset == NULL) {
    error("Failed to allocate the offset");
  }

  /* Get the size of all fields together and set the offsets */
  cache->total_size = 0;
  for (int i = 0; i < csds_type_count; i++) {
    if (number_particles[i] != 0) {
      for (int field = 0; field < n_fields; field++) {
        const struct field_information *info = &cache->fields[i][field];
        cache->offset[field] = cache->total_size;
        cache->total_size += info->field.size;

        /* Add the first derivative */
        if (info->first.field != field_enum_none) {
          cache->total_size += info->first.size;
        }
        /* Add the second derivative */
        if (info->second.field != field_enum_none) {
          cache->total_size += info->second.size;
        }
      }

      /* We have everything we need */
      break;
    }
  }

  /* Allocate the arrays for each particle type */
  for (int i = 0; i < csds_type_count; i++) {
    if (cache->capacity[i] != 0) {
      size_t size = cache->capacity[i];
      cache->last.fields[i] = (char *)malloc(size * cache->total_size);
      cache->last.time[i] = (double *)malloc(size * sizeof(double));
      cache->next.time[i] = (double *)malloc(size * sizeof(double));
      cache->next.offset[i] = (size_t *)malloc(size * sizeof(size_t));
      cache->next.fields[i] = (char *)malloc(size * cache->total_size);

      /* Check the allocation */
      if (cache->last.fields[i] == NULL || cache->last.time[i] == NULL ||
          cache->next.fields[i] == NULL || cache->next.time[i] == NULL ||
          cache->next.offset[i] == NULL) {
        error("Failed to allocate the cache memory");
      }
    } else {
      cache->last.fields[i] = NULL;
      cache->last.time[i] = NULL;
      cache->next.time[i] = NULL;
      cache->next.offset[i] = NULL;
      cache->next.fields[i] = NULL;
    }
  }

  /* Reset the current index */
  csds_cache_reset_current_index(cache);
}

#define CACHE_FREE(var)         \
  {                             \
    if (var != NULL) free(var); \
    var = NULL;                 \
  }

/**
 * @brief Free the allocated memory.
 *
 * @param cache The #csds_cache
 */
INLINE static void csds_cache_free(struct csds_cache *cache) {
  /* Free the particle arrays */
  for (int i = 0; i < csds_type_count; i++) {
    CACHE_FREE(cache->fields[i]);
    CACHE_FREE(cache->last.fields[i]);
    CACHE_FREE(cache->last.time[i]);
    CACHE_FREE(cache->next.time[i]);
    CACHE_FREE(cache->next.offset[i]);
    CACHE_FREE(cache->next.fields[i]);
  }
  /* Free the particle type arrays */
  CACHE_FREE(cache->offset);

  /* Set everything to 0 */
  csds_cache_empty_init(cache);
}

/** @brief Get the field index in the cache */
INLINE static int csds_cache_get_field_index(const struct csds_cache *cache,
                                             enum field_enum field,
                                             enum part_type type) {
  int field_index = -1;
  for (int i = 0; i < cache->n_fields; i++) {
    if (field == cache->fields[type][i].field.field) {
      field_index = i;
      break;
    }
  }

  /* Check that we found a field */
  if (field_index < 0) {
    error("The field %s was not found in the cache",
          field_enum_get_name(field));
  }
  return field_index;
}

/**
 * @brief Save a field inside the cache.
 *
 * @param cache The #csds_cache
 * @param index The index of the particle in the cache (-1 to get a new one)
 * @param type The type of particles
 * @param time_before The time of the previous record
 * @param before The field of the previous record
 * @param time_after The time of the next record
 * @param after The field of the next record
 * @param offset_next The offset of the next record
 * @param field The field stored in after and before
 */
INLINE static int64_t csds_cache_save(struct csds_cache *cache, int64_t index,
                                      enum part_type type, double time_before,
                                      struct csds_reader_field *before,
                                      double time_after,
                                      struct csds_reader_field *after,
                                      size_t offset_next,
                                      enum field_enum field) {

  const int new_part = index < 0;

  /* Get a new index if needed */
  if (new_part) {
#pragma omp atomic capture
    {
      index = cache->current_index[type];
      cache->current_index[type]++;
    }
#pragma omp atomic update
    cache->size[type]++;
  }

  /* Ensures that we have enough place */
  if (index >= cache->capacity[type]) {
    error("The cache is too small. Please increase Cache:OverAllocation.");
  }

  /* Get the index of the field */
  int field_index = csds_cache_get_field_index(cache, field, type);

  /* Check if the data are compatible with the previous writing */
  if (!new_part) {
    if (time_before != cache->last.time[type][index] ||
        time_after != cache->next.time[type][index] ||
        offset_next != cache->next.offset[type][index]) {
      error("The data within the cache are not compatible");
    }
  }

  /* Copy the data into the arrays */
  if (new_part) {
    cache->last.time[type][index] = time_before;
    cache->next.time[type][index] = time_after;
    cache->next.offset[type][index] = offset_next;
  }

  /* Copy the fields into the arrays */
  size_t shift = index * cache->total_size + cache->offset[field_index];
  memcpy(cache->last.fields[type] + shift, before->field,
         field_enum_get_size(field));
  memcpy(cache->next.fields[type] + shift, after->field,
         field_enum_get_size(field));
  shift += field_enum_get_size(field);

  /* Copy the first derivative into the arrays */
  enum field_enum first = field_enum_get_first_derivative(field);
  if (first != field_enum_none) {
    if (before->first_deriv == NULL || after->first_deriv == NULL) {
      error("First derivative is missing");
    }
    memcpy(cache->last.fields[type] + shift, before->first_deriv,
           field_enum_get_size(first));
    memcpy(cache->next.fields[type] + shift, after->first_deriv,
           field_enum_get_size(first));
    shift += field_enum_get_size(first);
  }

  /* Copy the second derivative into the arrays */
  enum field_enum second = field_enum_get_second_derivative(field);
  if (second != field_enum_none) {
    if (before->second_deriv == NULL || after->second_deriv == NULL) {
      error("Second derivative is missing");
    }
    memcpy(cache->last.fields[type] + shift, before->second_deriv,
           field_enum_get_size(second));
    memcpy(cache->next.fields[type] + shift, after->second_deriv,
           field_enum_get_size(second));
    shift += field_enum_get_size(second);
  }

  return index;
}

/**
 * @brief Remove a particle from the cache.
 *
 * @param cache The #csds_cache
 * @param index The index of the particle in the cache
 * @param type The type of particles
 */
INLINE static void csds_cache_remove_particle(struct csds_cache *cache,
                                              int64_t index,
                                              enum part_type type) {

  /* Decrease the counters */
  cache->current_index[type]--;
  cache->size[type]--;

  /* Swap with the last particle */
  int64_t size = cache->size[type];
  memcpy(cache->last.fields[type] + index * cache->total_size,
         cache->last.fields[type] + size * cache->total_size,
         cache->total_size);
  memcpy(cache->next.fields[type] + index * cache->total_size,
         cache->next.fields[type] + size * cache->total_size,
         cache->total_size);
  cache->last.time[type][index] = cache->last.time[type][size];
  cache->next.time[type][index] = cache->next.time[type][size];
  cache->next.offset[type][index] = cache->next.offset[type][size];
}
#endif  // CSDS_CACHE_H
