/*******************************************************************************
 * This file is part of CSDS.
 * Copyright (c) 2021 Loic Hausammann (loic.hausammann@epfl.ch)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 ******************************************************************************/

/**
 * @file csds_cosmology.h
 * @brief This file contains the functions that deals with the cosmology.
 */

#ifndef CSDS_COSMOLOGY_H
#define CSDS_COSMOLOGY_H

/* local includes */
#include "csds_inline.h"

/* Some standard headers */
#include <math.h>

struct csds_parser;

/**
 * @brief Structure containing the cosmological parameters.
 */
struct csds_cosmology {

  /*! Cold dark matter density parameter */
  double Omega_cdm;

  /*! Baryon density parameter */
  double Omega_b;

  /*! Cosmological constant density parameter */
  double Omega_lambda;

  /*! Total radiation density parameter (photons and other relics) */
  double Omega_r;

  /*! Curvature density parameter */
  double Omega_k;

  /*! Dark-energy equation of state at z=0 */
  double w_0;

  /*! Dark-energy evolution parameter */
  double w_a;

  /*! Hubble constant at z = 0 (in internal units) */
  double H0;
};

/**
 * @brief Compute the dark energy equation of state.
 *
 * @param cosmo The cosmological model
 * @param scale The scale factor
 *
 * @return The equation of state.
 */
INLINE static double csds_cosmology_get_w_tilde(
    const struct csds_cosmology *cosmo, const double scale) {
  const double w_tilde =
      (scale - 1.) * cosmo->w_a - (1. + cosmo->w_0 + cosmo->w_a) * log(scale);
  return exp(3. * w_tilde);
}

/**
 * @brief Compute E(a) for the derivatives of the scale factor.
 *
 * @param cosmo The cosmological model
 * @param scale The scale factor
 *
 * @return The first derivative of the scale factor.
 */
INLINE static double csds_cosmology_get_E(const struct csds_cosmology *cosmo,
                                          const double scale) {

  const double a_inv = 1. / scale;
  const double a_inv2 = a_inv * a_inv;
  const double Omega_m = cosmo->Omega_b + cosmo->Omega_cdm;
  const double dark_eos = csds_cosmology_get_w_tilde(cosmo, scale);
  const double E = sqrt(cosmo->Omega_r * a_inv2 * a_inv2 + /* Radiation */
                        Omega_m * a_inv * a_inv2 +         /* Matter */
                        cosmo->Omega_k * a_inv2 +          /* Curvature */
                        cosmo->Omega_lambda * dark_eos);
  return E;
}

/**
 * @brief Compute the first time derivative of the scale factor
 *
 * @param cosmo The cosmological model
 * @param scale The scale factor
 *
 * @return The first derivative of the scale factor.
 */
INLINE static double csds_cosmology_get_scale_factor_derivative(
    const struct csds_cosmology *cosmo, const double scale) {
  const double E = csds_cosmology_get_E(cosmo, scale);
  return cosmo->H0 * scale * E;
}

/**
 * @brief Compute the first time derivative of the scale factor
 *
 * @param cosmo The cosmological model
 * @param scale The scale factor
 *
 * @return The first derivative of the scale factor.
 */
INLINE static double csds_cosmology_get_scale_factor_second_derivative(
    const struct csds_cosmology *cosmo, const double scale) {

  /* Compute some powers of a */
  const double a_inv = 1. / scale;
  const double a_inv2 = a_inv * a_inv;
  const double a_inv4 = a_inv2 * a_inv2;

  /* Compute E */
  const double E = csds_cosmology_get_E(cosmo, scale);

  /* Compute the derivative of E */
  double dark_eos_der = (cosmo->w_a) - (1 + cosmo->w_0 + cosmo->w_a) * a_inv;
  dark_eos_der *= csds_cosmology_get_w_tilde(cosmo, scale);
  const double Omega_m = cosmo->Omega_b + cosmo->Omega_cdm;
  const double E_der =
      (-4 * cosmo->Omega_r * a_inv4 * a_inv - 3 * Omega_m * a_inv4 -
       2 * cosmo->Omega_k * a_inv2 * a_inv + dark_eos_der);

  /* Compute the full expression */
  const double a_dot = csds_cosmology_get_scale_factor_derivative(cosmo, scale);
  return cosmo->H0 * (a_dot * E + 0.5 * scale * E_der / E);
}

/**
 * @brief Add the cosmological factors for the position.
 * See Hausammann et al. 2021 for details.
 *
 * @param cosmo The cosmological model.
 * @param scale The scale factor.
 * @param vel The velocity to modify.
 * @param acc The acceleration to modify.
 */
INLINE static void csds_cosmology_add_factor_position(
    const struct csds_cosmology *cosmo, const double scale, float *vel,
    float *acc) {

  const double a_dot = csds_cosmology_get_scale_factor_derivative(cosmo, scale);
  const double scale2 = scale * scale;
  if (acc) {
    const double a_dot_inv = 1. / a_dot;
    const double scale2_inv = 1. / scale2;
    const double a_ddot =
        csds_cosmology_get_scale_factor_second_derivative(cosmo, scale);
    *acc = (*acc - 2 * a_dot * *vel) * scale;
    *acc -= scale2 * a_ddot * *vel * a_dot_inv;
    *acc *= scale2_inv * scale2_inv * a_dot_inv * a_dot_inv;
  }
  *vel /= scale2 * a_dot;
}

/**
 * @brief Add the cosmological factors for the velocities.
 * \f$ \frac{dv'}{da} = \frac{g'}{\dot{a}a} \f$.
 *
 * @param cosmo The cosmological model.
 * @param scale The scale factor.
 * @param acc The acceleration to modify.
 */
INLINE static void csds_cosmology_add_factor_velocity(
    const struct csds_cosmology *cosmo, const double scale, float *acc) {

  const double a_dot = csds_cosmology_get_scale_factor_derivative(cosmo, scale);
  *acc /= scale * a_dot;
}

void csds_cosmology_init(struct csds_cosmology *cosmo,
                         struct csds_parser *params);
#endif  // CSDS_CSDS_PARAMETERS_H
