/*******************************************************************************
 * This file is part of CSDS and was copied from SWIFT.
 * Copyright (c) 2019 Loic Hausammann (loic.hausammann@epfl.ch)
 *               SWIFT
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 ******************************************************************************/
/**
 * @brief This file contains functions that help to deal with messages.
 */
#ifndef CSDS_ERROR_H
#define CSDS_ERROR_H

#include "../config.h"
#include "csds_clocks.h"
#include "csds_inline.h"

#ifdef HAVE_PYTHON
#define NPY_NO_DEPRECATED_API NPY_1_7_API_VERSION
#include <Python.h>
#endif

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* Use exit when not developing, avoids core dumps. */
#ifdef CSDS_DEBUG_CHECKS
#define csds_abort(errcode) abort()
#else
#define csds_abort(errcode) exit(errcode)
#endif

/**
 * @brief Error macro. Prints the message given in argument and aborts.
 *
 */
#define error(s, ...)                                                      \
  ({                                                                       \
    fflush(stdout);                                                        \
    fprintf(stderr, "%s %s:%s():%i: " s "\n", clocks_get_timesincestart(), \
            __FILE__, __FUNCTION__, __LINE__, ##__VA_ARGS__);              \
    csds_abort(1);                                                         \
  })

/**
 * @brief Macro to print a localized message with variable arguments.
 *
 */
#define message(s, ...)                                                 \
  ({                                                                    \
    printf("%s %s: " s "\n", clocks_get_timesincestart(), __FUNCTION__, \
           ##__VA_ARGS__);                                              \
  })

#ifndef HAVE_PYTHON
#define error_python(s, ...) error(s, ##__VA_ARGS__);
#else
#include "frameobject.h"

/**
 * @brief Print the python trace back
 */
INLINE static void csds_loader_print_traceback(void) {

  PyGILState_STATE gstate = PyGILState_Ensure();

  PyFrameObject *frame = PyEval_GetFrame();
  int lineno = PyFrame_GetLineNumber(frame);
  PyObject *py_filename = frame->f_code->co_filename;
  const char *filename = PyUnicode_AsUTF8(py_filename);
  message("File %s, line: %i", filename, lineno);

  Py_DECREF(py_filename);
  Py_DECREF(frame);
  PyGILState_Release(gstate);
}

#define error_python(s, ...)       \
  ({                               \
    csds_loader_print_traceback(); \
    error(s, ##__VA_ARGS__);       \
  })
#endif

#endif  // CSDS_ERROR_H
