/*******************************************************************************
 * This file is part of CSDS.
 * Copyright (c) 2021 Loic Hausammann (loic.hausammann@epfl.ch)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 ******************************************************************************/
/**
 * @brief This file contains functions that deals with the fields
 */
#ifndef CSDS_FIELDS_H
#define CSDS_FIELDS_H

/* Include CSDS headers */
#include "csds_definitions.h"
#include "csds_python_tools.h"
#include "csds_tools.h"

#ifndef GEAR_CHEMISTRY_ELEMENT_COUNT
#define GEAR_CHEMISTRY_ELEMENT_COUNT 10
#endif

/**
 * @brief All the fields that are implemented */
enum field_enum {
  field_enum_special_flags = 0, /* Should be the first */
  field_enum_coordinates,
  field_enum_velocities,
  field_enum_accelerations,
  field_enum_masses,
  field_enum_smoothing_lengths,
  field_enum_internal_energies,
  field_enum_entropies,
  field_enum_particles_ids,
  field_enum_densities,
  field_enum_sphenix_secondary_fields,
  field_enum_birth_time,
  field_enum_gear_star_formation,
  field_enum_gear_chemistry_part,
  field_enum_gear_chemistry_spart,
  field_enum_none, /* Should be just before count */
  field_enum_count
};

/**
 * @brief Names of each field.
 */
extern const char *fields_names[];

/**
 * @brief Structure defining a single field without its derivatives
 */
struct single_field {
  /* The enum corresponding to the field */
  enum field_enum field;

  /* Its mask */
  mask_type mask;

  /* Its size */
  int size;

  /* Its position within the array h->fields */
  int position;
};

/**
 * @brief Structure containing all the information
 * of a field including the derivatives.
 */
struct field_information {
  /* The field */
  struct single_field field;

  /* Its first derivative */
  struct single_field first;

  /* Its second derivative */
  struct single_field second;
};

/**
 * @brief Get the #field_enum from the field name.
 */
INLINE static enum field_enum field_enum_get(const char *name) {
  for (enum field_enum i = field_enum_special_flags; i < field_enum_count;
       i++) {
    if (strcmp(name, fields_names[i]) == 0) return i;
  }

  error_python("Field %s not found", name);
}

/**
 * @brief Get the name of the field.
 */
INLINE static const char *field_enum_get_name(enum field_enum field) {
  return fields_names[field];
}

/**
 * @brief Get the #field_enum for the first derivative.
 */
INLINE static enum field_enum field_enum_get_first_derivative(
    enum field_enum field) {
  switch (field) {
    case field_enum_coordinates:
      return field_enum_velocities;
    case field_enum_velocities:
      return field_enum_accelerations;

    default:
      return field_enum_none;
  }
}

/**
 * @brief Get the #field_enum for the second derivative.
 */
INLINE static enum field_enum field_enum_get_second_derivative(
    enum field_enum field) {
  switch (field) {
    case field_enum_coordinates:
      return field_enum_accelerations;

    default:
      return field_enum_none;
  }
}

/**
 * @brief Get the size of the field.
 */
INLINE static unsigned int field_enum_get_size(enum field_enum field) {
  switch (field) {

    case field_enum_special_flags:
      return CSDS_SPECIAL_FLAGS_SIZE;

    case field_enum_coordinates:
      return 3 * sizeof(double);

    case field_enum_velocities:
    case field_enum_accelerations:
      return 3 * sizeof(float);

    case field_enum_masses:
    case field_enum_smoothing_lengths:
    case field_enum_internal_energies:
    case field_enum_entropies:
    case field_enum_birth_time:
    case field_enum_densities:
      return sizeof(float);

    case field_enum_particles_ids:
      return sizeof(long long);

    case field_enum_sphenix_secondary_fields:
      return 7 * sizeof(float);

    case field_enum_gear_star_formation:
      return 2 * sizeof(float) + sizeof(long long);

    case field_enum_gear_chemistry_part:
      return 2 * GEAR_CHEMISTRY_ELEMENT_COUNT * sizeof(double);

    case field_enum_gear_chemistry_spart:
      return GEAR_CHEMISTRY_ELEMENT_COUNT * sizeof(double);

    default:
      error_python("The field %s is not implemented", fields_names[field]);
  }
}

#ifdef HAVE_PYTHON
/**
 * @brief Set a #csds_python_field for the field.
 */
INLINE static struct csds_python_field field_enum_get_python_field(
    enum field_enum field) {
  switch (field) {
    case field_enum_coordinates:
      return csds_loader_python_field(/* Dimension */ 3, NPY_DOUBLE);

    case field_enum_velocities:
    case field_enum_accelerations:
      return csds_loader_python_field(/* Dimension */ 3, NPY_FLOAT32);

    case field_enum_masses:
    case field_enum_smoothing_lengths:
    case field_enum_internal_energies:
    case field_enum_entropies:
    case field_enum_birth_time:
    case field_enum_densities:
      return csds_loader_python_field(/* Dimension */ 1, NPY_FLOAT32);

    case field_enum_particles_ids:
      return csds_loader_python_field(/* Dimension */ 1, NPY_LONGLONG);

    case field_enum_sphenix_secondary_fields: {
      struct csds_python_field python_field =
          csds_loader_python_field(/* Dimension */ 7, CUSTOM_NPY_TYPE);

      csds_loader_python_field_add_subfield(&python_field, "Entropies",
                                            /* offset */ 0,
                                            /* Dimension */ 1, NPY_FLOAT32);

      csds_loader_python_field_add_subfield(&python_field, "Pressures",
                                            /* offset */ sizeof(float),
                                            /* Dimension */ 1, NPY_FLOAT32);

      csds_loader_python_field_add_subfield(
          &python_field, "ViscosityParameters", /* offset */ 2 * sizeof(float),
          /* Dimension */ 1, NPY_FLOAT32);

      csds_loader_python_field_add_subfield(
          &python_field, "DiffusionParameters", /* offset */ 3 * sizeof(float),
          /* Dimension */ 1, NPY_FLOAT32);

      csds_loader_python_field_add_subfield(&python_field,
                                            "LaplacianInternalEnergies",
                                            /* offset */ 4 * sizeof(float),
                                            /* Dimension */ 1, NPY_FLOAT32);

      csds_loader_python_field_add_subfield(
          &python_field, "VelocityDivergences", /* offset */ 5 * sizeof(float),
          /* Dimension */ 1, NPY_FLOAT32);

      csds_loader_python_field_add_subfield(
          &python_field, "VelocityDivergenceTimeDifferentials",
          /* offset */ 6 * sizeof(float),
          /* Dimension */ 1, NPY_FLOAT32);
      return python_field;
    }

    case field_enum_gear_star_formation: {
      struct csds_python_field python_field =
          csds_loader_python_field(/* Dimension */ 3, CUSTOM_NPY_TYPE);

      csds_loader_python_field_add_subfield(&python_field, "BirthDensities",
                                            /* offset */ 0,
                                            /* Dimension */ 1, NPY_FLOAT32);

      csds_loader_python_field_add_subfield(&python_field, "BirthMasses",
                                            /* offset */ sizeof(float),
                                            /* Dimension */ 1, NPY_FLOAT32);

      csds_loader_python_field_add_subfield(&python_field, "ProgenitorIDs",
                                            /* offset */ 2 * sizeof(float),
                                            /* Dimension */ 1, NPY_LONGLONG);

      return python_field;
    }

    case field_enum_gear_chemistry_part: {
      struct csds_python_field python_field =
          csds_loader_python_field(/* Dimension */ 2, CUSTOM_NPY_TYPE);

      csds_loader_python_field_add_subfield(
          &python_field, "SmoothedMetalMassFractions", /* offset */ 0,
          GEAR_CHEMISTRY_ELEMENT_COUNT, NPY_DOUBLE);

      csds_loader_python_field_add_subfield(
          &python_field, "MetalMassFractions",
          GEAR_CHEMISTRY_ELEMENT_COUNT * sizeof(double),
          GEAR_CHEMISTRY_ELEMENT_COUNT, NPY_DOUBLE);
      return python_field;
    }

    case field_enum_gear_chemistry_spart:
      return csds_loader_python_field(GEAR_CHEMISTRY_ELEMENT_COUNT, NPY_DOUBLE);

    default:
      error_python("The field %s is not implemented", fields_names[field]);
  }
}
#endif

struct mask_data {
  /* Number of bytes for a mask. */
  int size;

  /* Mask value. */
  mask_type mask;

  /* Name of the mask. */
  char name[CSDS_STRING_SIZE];
};

/**
 * @brief Initialize a derivative of a #field_information.
 *
 * @param der The derivative to initialize.
 * @param field The field corresponding to the derivative.
 * @param all_data All the #mask_data from the header.
 * @param size_all_data Size of all_data.
 */
INLINE static void field_information_set_derivative(struct single_field *der,
                                                    enum field_enum field,
                                                    struct mask_data *all_data,
                                                    int size_all_data) {

  der->field = field;
  der->mask = 0;
  der->size = 0;
  der->position = -1;
  /* Set the first derivative */
  if (der->field != field_enum_none) {
    struct mask_data *data = NULL;
    const char *name_first = field_enum_get_name(der->field);
    /* Get the first derivative in all_data */
    for (int i = 0; i < size_all_data; i++) {
      if (strcmp(name_first, all_data[i].name) == 0) {
        data = &all_data[i];
        break;
      }
    }
    /* The first derivative does not exist in the header. */
    if (data == NULL) {
      der->field = field_enum_none;
      return;
    }

    der->mask = data->mask;
    der->size = field_enum_get_size(der->field);
    if (der->size != data->size) {
      error_python("Inconsistent size for %s (%i != %i)", fields_names[field],
                   der->size, data->size);
    }
  }
}

/**
 * @brief Initialize a #field_information.
 *
 * @param info The #field_information to initialize.
 * @param data The #mask_data of the field.
 * @param all_data All the #mask_data from the header.
 * @param size_all_data Size of all_data.
 */
INLINE static void field_information_init(struct field_information *info,
                                          struct mask_data *data,
                                          struct mask_data *all_data,
                                          int size_all_data) {

  /* Set the field */
  enum field_enum field = field_enum_get(data->name);
  info->field.field = field;
  info->field.mask = data->mask;
  info->field.size = field_enum_get_size(field);
  if (info->field.size != data->size) {
    error_python("Inconsistent size for %s (%i != %i)", fields_names[field],
                 info->field.size, data->size);
  }
  info->field.position = -1;

  /* Get the derivatives */
  enum field_enum first = field_enum_get_first_derivative(field);
  enum field_enum second = field_enum_get_second_derivative(field);

  /* Set the two derivatives */
  field_information_set_derivative(&info->first, first, all_data,
                                   size_all_data);
  field_information_set_derivative(&info->second, second, all_data,
                                   size_all_data);
}

/**
 * @brief Set the position of the field and its derivatives
 * according to the array.
 *
 * @param info An array of #field_information.
 * @param size The size of the array.
 */
INLINE static void field_information_set_positions(
    struct field_information *info, const unsigned int size) {
  for (unsigned int i = 0; i < size; i++) {
    /* Set the field */
    info[i].field.position = i;

    /* Set the first derivative */
    if (info[i].first.field != field_enum_none) {
      for (unsigned int j = 0; j < size; j++) {
        if (info[i].first.field == info[j].field.field) {
          info[i].first.position = j;
          break;
        }
      }
    }

    /* Set the second derivative */
    if (info[i].second.field != field_enum_none) {
      for (unsigned int j = 0; j < size; j++) {
        if (info[i].second.field == info[j].field.field) {
          info[i].second.position = j;
          break;
        }
      }
    }
  }
}

#endif  // CSDS_FIELDS_H
