// Copyright 2020 Joshua J Baker. All rights reserved.
// Use of this source code is governed by an MIT-style
// license that can be found in the LICENSE file.

#include <stdio.h>
#include <string.h>

/* CSDS headers */
#include "csds_error.h"
#include "csds_hashmap.h"

/**
 * @brief returns a new hash map.
 *
 * The csds_hashmap must be freed with csds_hashmap_free().
 *
 * @param cap The default lower capacity of the csds_hashmap. Setting this to
 * zero will default to 16.
 */
struct csds_hashmap *csds_hashmap_new(size_t cap) {

  /* Get the size as a power of 2 */
  size_t ncap = 16;
  if (cap < ncap) {
    cap = ncap;
  } else {
    while (ncap < cap) {
      ncap *= 2;
    }
    cap = ncap;
  }

  /* A bucket is composed of the bucket + the data */
  size_t bucketsz = sizeof(struct bucket) + sizeof(struct index_data);
  while (bucketsz & (sizeof(uintptr_t) - 1)) {
    bucketsz++;
  }
  /* We allocate some extra space for temporary data */
  size_t size = sizeof(struct csds_hashmap) + 2 * bucketsz;
  struct csds_hashmap *map = malloc(size);
  if (!map) {
    return NULL;
  }

  /* Set all the attributes */
  memset(map, 0, sizeof(struct csds_hashmap));
  map->bucketsz = bucketsz;
  map->spare = ((char *)map) + sizeof(struct csds_hashmap);
  map->edata = (char *)map->spare + bucketsz;
  map->cap = cap;
  map->nbuckets = cap;
  map->mask = map->nbuckets - 1;

  /* Allocate the array of buckets */
  map->buckets = malloc(map->bucketsz * map->nbuckets);
  if (!map->buckets) {
    free(map);
    return NULL;
  }

  /* Ensures that all the buckets are empty */
  memset(map->buckets, 0, map->bucketsz * map->nbuckets);

  /* Set the limits for the size */
  map->growat = map->nbuckets / hashmap_overallocation;
  map->shrinkat = map->nbuckets * 0.10;
  return map;
}

/**
 * @brief quickly clears the map.
 * When the update_cap is provided, the map's capacity will be updated to match
 * the currently number of allocated buckets. This is an optimization to ensure
 * that this operation does not perform any allocations.
 */
void csds_hashmap_clear(struct csds_hashmap *map, int update_cap) {
  map->count = 0;
  if (update_cap) {
    map->cap = map->nbuckets;
  } else if (map->nbuckets != map->cap) {
    void *new_buckets = malloc(map->bucketsz * map->cap);
    if (new_buckets) {
      free(map->buckets);
      map->buckets = new_buckets;
    }
    map->nbuckets = map->cap;
  }
  memset(map->buckets, 0, map->bucketsz * map->nbuckets);
  map->mask = map->nbuckets - 1;
  map->growat = map->nbuckets / hashmap_overallocation;
  map->shrinkat = map->nbuckets * 0.10;
}

/**
 * @brief Resize the hash table.
 */
static int resize(struct csds_hashmap *map, size_t new_cap) {

  /* Allocate a new hashmap */
  struct csds_hashmap *map2 = csds_hashmap_new(new_cap);
  if (!map2) {
    return 0;
  }

  /* Copy the buckets */
  for (size_t i = 0; i < map->nbuckets; i++) {
    struct bucket *entry = bucket_at(map, i);
    /* Skip empty buckets */
    if (!entry->dib) {
      continue;
    }
    entry->dib = 1;
    size_t j = entry->hash & map2->mask;
    /* Copy the bucket into map2 */
    while (1) {
      struct bucket *bucket = bucket_at(map2, j);
      /* Empty bucket in map2 */
      if (bucket->dib == 0) {
        memcpy(bucket, entry, map->bucketsz);
        break;
      }
      /* Bucket not empty shift the previous data */
      if (bucket->dib < entry->dib) {
        memcpy(map2->spare, bucket, map->bucketsz);
        memcpy(bucket, entry, map->bucketsz);
        memcpy(entry, map2->spare, map->bucketsz);
      }
      j = (j + 1) & map2->mask;
      entry->dib += 1;
    }
  }

  /* First cleanup */
  free(map->buckets);

  /* Copy the data */
  map->buckets = map2->buckets;
  map->nbuckets = map2->nbuckets;
  map->mask = map2->mask;
  map->growat = map2->growat;
  map->shrinkat = map2->shrinkat;

  /* Cleanup */
  free(map2);
  return 1;
}

/**
 * @brief inserts or replaces an item in the csds_hash map.
 * If an item is
 * replaced then it is returned otherwise NULL is returned. This operation
 * may allocate memory.
 */
void *csds_hashmap_set(struct csds_hashmap *map, struct index_data *item) {
  if (!item) {
    error("item is null");
  }
  /* Increase the size of the map if needed */
  if (map->count == map->growat) {
    if (!resize(map, map->nbuckets * 2)) {
      error("Failed to reallocate memory");
    }
  }

  /* Create a bucket from the item */
  struct bucket *entry = map->edata;
  entry->hash = get_hash(item->id);
  entry->dib = 1;
  memcpy(bucket_item(entry), item, sizeof(struct index_data));

  /* Place the new bucket into the array */
  size_t i = entry->hash & map->mask;
  while (1) {
    struct bucket *bucket = bucket_at(map, i);
    /* Empty bucket */
    if (bucket->dib == 0) {
      memcpy(bucket, entry, map->bucketsz);
      map->count++;
      return NULL;
    }
    /* Replace bucket */
    if (entry->hash == bucket->hash &&
        bucket_item(entry)->id == bucket_item(bucket)->id) {

      memcpy(map->spare, bucket_item(bucket), sizeof(struct index_data));
      memcpy(bucket_item(bucket), bucket_item(entry),
             sizeof(struct index_data));
      return map->spare;
    }
    /* Move current element to next available bucket */
    if (bucket->dib < entry->dib) {
      memcpy(map->spare, bucket, map->bucketsz);
      memcpy(bucket, entry, map->bucketsz);
      memcpy(entry, map->spare, map->bucketsz);
    }
    i = (i + 1) & map->mask;
    entry->dib += 1;
  }
}

/**
 * @brief returns the item based on the provided key. If the item is not
 * found then NULL is returned.
 */
struct index_data *csds_hashmap_get(struct csds_hashmap *map, id_type key) {
  uint64_t hash = get_hash(key);
  size_t i = hash & map->mask;
  while (1) {
    struct bucket *bucket = bucket_at(map, i);
    /* Empty bucket, thus key not found */
    if (!bucket->dib) {
      return NULL;
    }
    /* Key match */
    if (bucket->hash == hash && key == bucket_item(bucket)->id) {
      return bucket_item(bucket);
    }

    /* Update current bucket index */
    i = (i + 1) & map->mask;
  }
}

/**
 * @brief removes an item from the hash map and returns it. If the
 * item is not found then NULL is returned.
 */
void *csds_hashmap_delete(struct csds_hashmap *map, id_type key) {
  uint64_t hash = get_hash(key);
  size_t i = hash & map->mask;
  while (1) {
    struct bucket *bucket = bucket_at(map, i);
    /* Not found */
    if (!bucket->dib) {
      return NULL;
    }
    /* Did we get the correct particle? */
    if (bucket->hash == hash && key == bucket_item(bucket)->id) {
      memcpy(map->spare, bucket_item(bucket), sizeof(struct index_data));
      bucket->dib = 0;
      /* Shift the elements with the same hash */
      while (1) {
        struct bucket *prev = bucket;
        i = (i + 1) & map->mask;
        bucket = bucket_at(map, i);
        if (bucket->dib <= 1) {
          prev->dib = 0;
          break;
        }
        memcpy(prev, bucket, map->bucketsz);
        prev->dib--;
      }
      map->count--;
      if (map->nbuckets > map->cap && map->count <= map->shrinkat) {
        // Ignore the return value. It's ok for the resize operation to
        // fail to allocate enough memory because a shrink operation
        // does not change the integrity of the data.
        resize(map, map->nbuckets / 2);
      }
      return map->spare;
    }
    /* Move to the next element with the same hash */
    i = (i + 1) & map->mask;
  }
}

/**
 * @brief returns the number of items in the hash map.
 */
size_t csds_hashmap_count(struct csds_hashmap *map) { return map->count; }

/**
 * @brief frees the hash map
 */
void csds_hashmap_free(struct csds_hashmap *map) {
  if (!map) return;
  free(map->buckets);
  free(map);
}

/**
 * @brief Writes the hash map
 *
 * @param map The hashmap.
 * @param f The file to use.
 */
void csds_hashmap_write(struct csds_hashmap *map, FILE *f) {
  size_t count = 0;
  /* Loop over all the buckets */
  for (size_t i = 0; i < map->nbuckets; i++) {
    struct bucket *bucket = bucket_at(map, i);
    /* Do only the non-empty buckets */
    if (bucket->dib) {
      count += 1;
      fwrite(bucket_item(bucket), sizeof(struct index_data), 1, f);
    }
  }

  /* Ensure that the correct number of elements
   * have been written. */
  if (count != map->count) {
    error("Written a wrong number of elements.");
  }
}

/**
 * @brief Try to get item in a given bucket.
 * Returns NULL if it does not exist.
 *
 * @param map The hashmap.
 * @param i The request bucket
 */
struct index_data *csds_hashmap_get_from_position(struct csds_hashmap *map,
                                                  size_t i) {
  struct bucket *bucket = bucket_at(map, i);
  if (bucket->dib) {
    return bucket_item(bucket);
  } else {
    return NULL;
  }
}

/**
 * @brief Returns the number of buckets.
 */
size_t csds_hashmap_get_number_buckets(struct csds_hashmap *map) {
  return map->nbuckets;
}
