/*
 * The file was obtained from https://github.com/tidwall/hashmap.c
 * and was slightly adapted.
 */
// Copyright 2020 Joshua J Baker. All rights reserved.
// Use of this source code is governed by an MIT-style
// license that can be found in the LICENSE file.

#ifndef CSDS_HASHMAP_H
#define CSDS_HASHMAP_H

#include <stddef.h>
#include <stdint.h>
#include <stdio.h>

/* CSDS headers */
#include "csds_inline.h"

#define id_type int64_t
#define hashmap_overallocation 4. / 3.
#define hashmap_seed0 1513
#define hashmap_seed1 7654

/**
 * @brief Data structure contained in the csds files.
 */
struct index_data {
  /* Id of the particle. */
  id_type id;

  /* Offset of the particle in the file. */
  uint64_t offset;
};

/**
 * hashmap is an open addressed hash map using robinhood hashing.
 */
struct csds_hashmap {

  /* The capacity of the hashmap. */
  size_t cap;

  /* The size of each bucket. */
  size_t bucketsz;

  /* The number of buckets. */
  size_t nbuckets;

  /* The number of items. */
  size_t count;

  /* A mask to restrict the hash within nbuckets. */
  size_t mask;

  /* The size at which the hashmap increases its size. */
  size_t growat;

  /* The size at which the hashmap decreases its size. */
  size_t shrinkat;

  /* The array of buckets. */
  void *buckets;

  /* A spare buckets for some operations */
  void *spare;

  /* A second spare buckets for setting an item. */
  void *edata;
};

struct bucket {
  /* The hash value of the bucket */
  uint64_t hash : 48;

  /* Number of elements with this hash value */
  uint64_t dib : 16;
};

struct csds_hashmap *csds_hashmap_new(size_t cap);
void csds_hashmap_free(struct csds_hashmap *map);
void csds_hashmap_clear(struct csds_hashmap *map, int update_cap);
size_t csds_hashmap_count(struct csds_hashmap *map);
struct index_data *csds_hashmap_get(struct csds_hashmap *map, id_type key);
void *csds_hashmap_set(struct csds_hashmap *map, struct index_data *item);
void *csds_hashmap_delete(struct csds_hashmap *map, id_type key);
void csds_hashmap_write(struct csds_hashmap *map, FILE *f);
size_t csds_hashmap_get_number_buckets(struct csds_hashmap *map);
struct index_data *csds_hashmap_get_from_position(struct csds_hashmap *map,
                                                  size_t i);

INLINE static struct bucket *bucket_at(struct csds_hashmap *map, size_t index) {
  return (struct bucket *)(((char *)map->buckets) + (map->bucketsz * index));
}

INLINE static struct index_data *bucket_item(struct bucket *entry) {
  char *out = ((char *)entry) + sizeof(struct bucket);
  return (struct index_data *)out;
}

//-----------------------------------------------------------------------------
// SipHash reference C implementation
//
// Copyright (c) 2012-2016 Jean-Philippe Aumasson
// <jeanphilippe.aumasson@gmail.com>
// Copyright (c) 2012-2014 Daniel J. Bernstein <djb@cr.yp.to>
//
// To the extent possible under law, the author(s) have dedicated all copyright
// and related and neighboring rights to this software to the public domain
// worldwide. This software is distributed without any warranty.
//
// You should have received a copy of the CC0 Public Domain Dedication along
// with this software. If not, see
// <http://creativecommons.org/publicdomain/zero/1.0/>.
//
// default: SipHash-2-4
//-----------------------------------------------------------------------------
static uint64_t SIP64(const id_type *key) {
  const uint8_t *in = (uint8_t *)key;
  uint64_t seed0 = hashmap_seed0;
  uint64_t seed1 = hashmap_seed1;
  const int inlen = sizeof(id_type);
#define U8TO64_LE(p)                                           \
  {                                                            \
    (((uint64_t)((p)[0])) | ((uint64_t)((p)[1]) << 8) |        \
     ((uint64_t)((p)[2]) << 16) | ((uint64_t)((p)[3]) << 24) | \
     ((uint64_t)((p)[4]) << 32) | ((uint64_t)((p)[5]) << 40) | \
     ((uint64_t)((p)[6]) << 48) | ((uint64_t)((p)[7]) << 56))  \
  }
#define U64TO8_LE(p, v)                        \
  {                                            \
    U32TO8_LE((p), (uint32_t)((v)));           \
    U32TO8_LE((p) + 4, (uint32_t)((v) >> 32)); \
  }
#define U32TO8_LE(p, v)            \
  {                                \
    (p)[0] = (uint8_t)((v));       \
    (p)[1] = (uint8_t)((v) >> 8);  \
    (p)[2] = (uint8_t)((v) >> 16); \
    (p)[3] = (uint8_t)((v) >> 24); \
  }
#define ROTL(x, b) (uint64_t)(((x) << (b)) | ((x) >> (64 - (b))))
#define SIPROUND       \
  {                    \
    v0 += v1;          \
    v1 = ROTL(v1, 13); \
    v1 ^= v0;          \
    v0 = ROTL(v0, 32); \
    v2 += v3;          \
    v3 = ROTL(v3, 16); \
    v3 ^= v2;          \
    v0 += v3;          \
    v3 = ROTL(v3, 21); \
    v3 ^= v0;          \
    v2 += v1;          \
    v1 = ROTL(v1, 17); \
    v1 ^= v2;          \
    v2 = ROTL(v2, 32); \
  }
  uint64_t k0 = U8TO64_LE((uint8_t *)&seed0);
  uint64_t k1 = U8TO64_LE((uint8_t *)&seed1);
  uint64_t v3 = UINT64_C(0x7465646279746573) ^ k1;
  uint64_t v2 = UINT64_C(0x6c7967656e657261) ^ k0;
  uint64_t v1 = UINT64_C(0x646f72616e646f6d) ^ k1;
  uint64_t v0 = UINT64_C(0x736f6d6570736575) ^ k0;
  const uint8_t *end = in + inlen - (inlen % sizeof(uint64_t));
  for (; in != end; in += 8) {
    uint64_t m = U8TO64_LE(in);
    v3 ^= m;
    SIPROUND;
    SIPROUND;
    v0 ^= m;
  }
  uint64_t b = ((uint64_t)inlen) << 56;
  switch (inlen & 7) {
    case 7:
      b |= ((uint64_t)in[6]) << 48;
      break;
    case 6:
      b |= ((uint64_t)in[5]) << 40;
      break;
    case 5:
      b |= ((uint64_t)in[4]) << 32;
      break;
    case 4:
      b |= ((uint64_t)in[3]) << 24;
      break;
    case 3:
      b |= ((uint64_t)in[2]) << 16;
      break;
    case 2:
      b |= ((uint64_t)in[1]) << 8;
      break;
    case 1:
      b |= ((uint64_t)in[0]);
      break;
    case 0:
      break;
  }
  v3 ^= b;
  SIPROUND;
  SIPROUND;
  v0 ^= b;
  v2 ^= 0xff;
  SIPROUND;
  SIPROUND;
  SIPROUND;
  SIPROUND;
  b = v0 ^ v1 ^ v2 ^ v3;
  uint64_t out = 0;
  U64TO8_LE((uint8_t *)&out, b);
  return out;
}

// hashmap_sip returns a hash value for `data` using SipHash-2-4.
INLINE static uint64_t get_hash(id_type x) {
  return SIP64(&x) & ((1LU << 48) - 1);
}

#endif
