/*******************************************************************************
 * This file is part of CSDS.
 * Copyright (c) 2019 Loic Hausammann (loic.hausammann@epfl.ch)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 ******************************************************************************/
#include "csds_header.h"

#include "csds_logfile.h"
#include "csds_mapped_file.h"
#include "csds_reader.h"
#include "csds_tools.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* Name of each offset direction. */
const char *csds_offset_name[csds_offset_count] = {
    "Forward",
    "Backward",
    "Corrupted",
};

/**
 * @brief Print the properties of the header to stdout.
 *
 * @param h The #header.
 */
void header_print(const struct header *h) {
#ifdef CSDS_DEBUG_CHECKS
  message("Debug checks enabled.");
#endif
  message("First Offset:     %lu.", h->offset_first_record);
  message("Offset direction: %s.", csds_offset_name[h->offset_direction]);

  for (int type = 0; type < csds_type_count; type++) {
    message("Masks for particle type %i:", type);
    for (int k = 0; k < h->number_fields[type]; k++) {
      /* Do the field */
      struct field_information *field = &h->fields[type][k];
      const char *name = field_enum_get_name(field->field.field);
      message("  %20s:\t mask=%03i\t size=%i", name, field->field.mask,
              field->field.size);

      /* Do the first derivative */
      if (field->first.field != field_enum_none) {
        const char *name_first = field_enum_get_name(field->first.field);
        message("  %20s   First derivative: %s", "", name_first);
      }

      /* Do the second derivative */
      if (field->second.field != field_enum_none) {
        const char *name_second = field_enum_get_name(field->second.field);
        message("  %20s   Second derivative: %s", "", name_second);
      }
    }
  }
};

/**
 * @brief free the allocated memory.
 *
 * @param h The #header.
 */
void header_free(struct header *h) {
  for (int i = 0; i < csds_type_count; i++) {
    if (h->fields[i]) {
      free(h->fields[i]);
      h->fields[i] = NULL;
    }
  }
};

/**
 * @brief Update the offset direction in the structure and
 * write it to the logfile.
 *
 * @param h #header file structure.
 * @param new_value The new value to write.
 *
 */
void header_change_offset_direction(struct header *h,
                                    enum csds_offset_direction new_value) {
  h->offset_direction = new_value;
  /* Skip file format and version numbers. */
  size_t offset = CSDS_STRING_SIZE + 2 * sizeof(int);

  csds_mapped_file_write_data(&h->log->log, offset, sizeof(unsigned int),
                              &new_value);
}

/**
 * @brief read the csds header.
 *
 * @param h out: The #header.
 * @param log The #csds_logfile.
 */
void header_read(struct header *h, struct csds_logfile *log) {

  /* Set pointer to log. */
  h->log = log;

  /* Position in the file */
  size_t offset = 0;

  /* read the file format. */
  char file_format[CSDS_STRING_SIZE];
  offset = csds_mapped_file_read_data(&log->log, offset, CSDS_STRING_SIZE,
                                      &file_format);
  if (strcmp(file_format, CSDS_FORMAT_STRING))
    error_python("Wrong file format (%s).", file_format);

  /* Read the major version number. */
  offset = csds_mapped_file_read_data(&log->log, offset, sizeof(int),
                                      &h->major_version);

  /* Read the minor version number. */
  offset = csds_mapped_file_read_data(&log->log, offset, sizeof(int),
                                      &h->minor_version);

  /* Check the mask size */
  if (sizeof(mask_type) != CSDS_MASK_SIZE)
    error_python("The mask sizes are not consistent (%li != %i)",
                 sizeof(mask_type), CSDS_MASK_SIZE);

  struct csds_reader *reader = log->reader;
  if (&reader->log != log) error_python("Wrong link to the reader.");

  if (reader->verbose > 0)
    message("File version %i.%i.", h->major_version, h->minor_version);

  /* Read the offset directions. */
  offset = csds_mapped_file_read_data(&log->log, offset, sizeof(int),
                                      &h->offset_direction);

  if (!header_is_forward(h) && !header_is_backward(h) &&
      !header_is_corrupted(h))
    error_python("Wrong offset value in the header (%i).", h->offset_direction);

  /* Read offset to first record. */
  h->offset_first_record = 0;
  offset = csds_mapped_file_read_data(&log->log, offset, CSDS_OFFSET_SIZE,
                                      &h->offset_first_record);

  /* Read the size of the strings. */
  unsigned int string_length = 0;
  offset = csds_mapped_file_read_data(&log->log, offset, sizeof(unsigned int),
                                      &string_length);

  /* Check if value defined in this file is large enough. */
  if (CSDS_STRING_SIZE < string_length) {
    error_python("Name too large in log file %i.", string_length);
  }

  /* Read the number of masks. */
  unsigned int masks_count = 0;
  offset = csds_mapped_file_read_data(&log->log, offset, sizeof(unsigned int),
                                      &masks_count);

  /* Allocate the masks memory. */
  struct mask_data *masks = malloc(sizeof(struct mask_data) * masks_count);
  if (masks == NULL) {
    error_python("Failed to allocate the memory for the masks.");
  }

  /* Loop over all masks. */
  for (unsigned int i = 0; i < masks_count; i++) {
    /* Read the mask name. */
    offset = csds_mapped_file_read_data(&log->log, offset, string_length,
                                        masks[i].name);

    /* Set the mask value. */
    masks[i].mask = 1 << i;

    /* Read the mask data size. */
    offset = csds_mapped_file_read_data(&log->log, offset, sizeof(unsigned int),
                                        &masks[i].size);

    /* Print the information. */
    if (reader->verbose > 1) {
      message("Field %s found in the logfile", masks[i].name);
    }
  }

  /* Check that the special flag is at the expected place. */
  const int special_flag = 0;
  if (strcmp(masks[special_flag].name, CSDS_SPECIAL_FLAGS_NAME) != 0) {
    error_python("The special flag should be the first mask '%s'",
                 masks[special_flag].name);
  }

  /* Check the special flag mask */
  if (masks[special_flag].mask != CSDS_SPECIAL_FLAGS_MASK) {
    error_python("The mask of the special flag should be %i",
                 CSDS_SPECIAL_FLAGS_MASK);
  }

  /* Check the size of the special flag. */
  if (masks[special_flag].size != CSDS_SPECIAL_FLAGS_SIZE) {
    error_python("The special flag is expected to be %li bytes",
                 CSDS_SPECIAL_FLAGS_SIZE);
  }

  /* Check that the timestamp is at the expected place */
  const int timestamp = 1;
  if (strcmp(masks[timestamp].name, CSDS_TIMESTAMP_NAME) != 0) {
    error_python("The time stamp should be the second mask");
  }

  /* Check the timestamp mask */
  if (masks[timestamp].mask != CSDS_TIMESTAMP_MASK) {
    error_python("The mask of the timestamp should be %i", CSDS_TIMESTAMP_MASK);
  }

  /* Check the size of the timestamp. */
  if (masks[timestamp].size != CSDS_TIMESTAMP_SIZE) {
    error_python(
        "The timestamp is expected to be an integertime_t and a double.");
  }

  /* Read the number of fields per particle  */
  offset = csds_mapped_file_read_data(
      &log->log, offset, sizeof(h->number_fields), h->number_fields);

  /* Read the order of the fields */
  for (int type = 0; type < csds_type_count; type++) {
    if (h->number_fields[type] == 0) {
      h->fields[type] = NULL;
      continue;
    }

    /* Allocate and read the order */
    size_t size = h->number_fields[type] * sizeof(int);
    int *order = (int *)malloc(size);
    offset = csds_mapped_file_read_data(&log->log, offset, size, order);

    /* Add the special flag */
    size = (h->number_fields[type] + 1) * sizeof(struct field_information);

    /* Allocate the memory in the header */
    h->fields[type] = (struct field_information *)malloc(size);

    /* Set the fields */
    for (int k = 0; k < h->number_fields[type]; k++) {
      /* Special flag is at 0, thus skip it */
      field_information_init(&h->fields[type][k + 1], &masks[order[k]], masks,
                             masks_count);
    }

    /* Add the contribution from the special flag */
    field_information_init(&h->fields[type][0],
                           &masks[CSDS_SPECIAL_FLAGS_INDEX], masks,
                           masks_count);
    h->number_fields[type]++;

    /* Set the positions */
    field_information_set_positions(h->fields[type], h->number_fields[type]);

    /* Free the memory */
    free(order);
  }

  /* Check the logfile header's size. */
  if (offset != h->offset_first_record) {
    error_python("Wrong header size (in header %zi, current %zi).",
                 h->offset_first_record, offset);
  }

  /* Cleanup the memory */
  free(masks);
};

/**
 * @brief Count number of bits in a given mask (without the record header).
 *
 * @param h #header file structure.
 * @param mask Mask to compute.
 *
 * @return number of bits in mask.
 */
size_t header_get_record_size_from_mask(const struct header *h,
                                        mask_type mask) {
  size_t count = 0;

  /* Do the time stamp */
  if (mask & CSDS_TIMESTAMP_MASK) {
    count += CSDS_TIMESTAMP_SIZE;
    mask &= ~CSDS_TIMESTAMP_MASK;
  }

  /* Loop over each masks. */
  for (int part_type = 0; part_type < csds_type_count; part_type++) {
    for (int i = 0; i < h->number_fields[part_type]; i++) {
      if (mask & h->fields[part_type][i].field.mask) {
        count += h->fields[part_type][i].field.size;
        mask &= ~h->fields[part_type][i].field.mask;

        /* Early exit */
        if (mask == 0) {
          part_type = csds_type_count;
          break;
        }
      }
    }
  }

  /* Ensure that we found all the masks */
  if (mask != 0) {
    error("A mask was not found");
  }
  return count;
}

/**
 * @brief Find the #field_information from the field name.
 *
 * @param h The #header
 * @param name The name of the field.
 * @param part_type The particle type.
 *
 * @return The first #field_information found.
 */
const struct field_information *header_get_field_from_name(
    const struct header *h, const char *name, enum part_type part_type) {

  const struct field_information *info = NULL;
  for (int i = 0; i < h->number_fields[part_type]; i++) {
    const char *current_name =
        field_enum_get_name(h->fields[part_type][i].field.field);
    if (strcmp(name, current_name) == 0) {
      info = &h->fields[part_type][i];
      break;
    }
  }

  /* Check if we obtained a field. */
  if (info == NULL) {
    error(
        "The logfile does not contain the field %s"
        " for particle type %i",
        name, part_type);
  }

  return info;
}
