/*******************************************************************************
 * This file is part of CSDS.
 * Copyright (c) 2019 Loic Hausammann (loic.hausammann@epfl.ch)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 ******************************************************************************/
#ifndef CSDS_CSDS_HEADER_H
#define CSDS_CSDS_HEADER_H

#include "csds_fields.h"
#include "csds_part_type.h"
#include "csds_tools.h"

#include <stdio.h>
#include <stdlib.h>

/**
 * @brief Names of the offset directions.
 */
extern const char *csds_offset_name[];

struct csds_logfile;

/**
 * @brief This structure contains everything from the file header.
 *
 * This structure is initialized by #header_read and need to be freed
 * with #header_free.
 *
 * The information contained by the header can be easily access with
 * the functions #header_get_record_size_from_mask and #header_get_field_index.
 *
 * The only function that modify the file is #header_change_offset_direction.
 */
struct header {
  /* Dump's major version. */
  int major_version;

  /* Dump's minor version. */
  int minor_version;

  /* Offset of the first record. */
  size_t offset_first_record;

  /* Direction of the offset in the records. */
  enum csds_offset_direction offset_direction;

  /* The corresponding log. */
  struct csds_logfile *log;

  /* Number of mask per particle type */
  int number_fields[csds_type_count];

  /* The fields for each particle type in the correct order.
     By construction, the special flag is the first element in each array. */
  struct field_information *fields[csds_type_count];
};

void header_print(const struct header *h);
void header_free(struct header *h);
void header_read(struct header *h, struct csds_logfile *log);
size_t header_get_record_size_from_mask(const struct header *h, mask_type mask);
void header_change_offset_direction(struct header *h,
                                    enum csds_offset_direction new_value);
const struct field_information *header_get_field_from_name(
    const struct header *h, const char *name, enum part_type part_type);

/**
 * @brief Check if the offset are forward.
 * @param h The #header.
 */
INLINE static int header_is_forward(const struct header *h) {
  return h->offset_direction == csds_offset_forward;
}

/**
 * @brief Check if the offset are backward.
 * @param h The #header.
 */
INLINE static int header_is_backward(const struct header *h) {
  return h->offset_direction == csds_offset_backward;
}

/**
 * @brief Check if the offset are corrupted.
 * @param h The #header.
 */
INLINE static int header_is_corrupted(const struct header *h) {
  return h->offset_direction == csds_offset_corrupted;
}

#endif  // CSDS_CSDS_HEADER_H
