/*******************************************************************************
 * This file is part of CSDS.
 * Copyright (c) 2019 Loic Hausammann (loic.hausammann@epfl.ch)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 ******************************************************************************/

/* Include the corresponding header */
#include "csds_index.h"

/* Include the standard headers */
#include <errno.h>
#include <fcntl.h>

/* Include local headers */
#include "csds_mapped_file.h"
#include "csds_reader.h"
#include "quick_sort.h"

/* Define the place and size of each element in the header. */
/* The time in double precision. */
#define csds_index_time_offset 0
#define csds_index_time_size sizeof(double)
/* The time on the integer timeline. */
#define csds_index_integer_time_offset \
  csds_index_time_offset + csds_index_time_size
#define csds_index_integer_time_size sizeof(integertime_t)
/* The number of particle (for each type). */
#define csds_index_npart_offset \
  csds_index_integer_time_offset + csds_index_integer_time_size
#define csds_index_npart_size sizeof(uint64_t) * csds_type_count
/* The flag used for checking if the file is sorted or not. */
#define csds_index_is_sorted_offset \
  csds_index_npart_offset + csds_index_npart_size
#define csds_index_is_sorted_size sizeof(char)
/* The array containing the offset and ids. Rounding to a multiplier of 8 in
 * order to align the memory */
#define csds_index_data_offset \
  ((csds_index_is_sorted_offset + csds_index_is_sorted_size + 7) & ~7)

/**
 * @brief Read the index file header.
 *
 * @param index The #csds_index.
 * @param filename The filename of the index file.
 */
void csds_index_read_header(struct csds_index *index, const char *filename) {

  /* Open the file */
  if (index->reader->verbose > 1) {
    message("Reading %s", filename);
  }
  csds_index_map_file(index, filename, 0);

  /* Read times */
  memcpy(&index->time, index->index.map + csds_index_time_offset,
         csds_index_time_size);
  memcpy(&index->integer_time,
         index->index.map + csds_index_integer_time_offset,
         csds_index_integer_time_size);

  /* Read the number of particles. */
  memcpy(index->nparts, index->index.map + csds_index_npart_offset,
         csds_index_npart_size);

  /* Read whether the file is sorted. */
  memcpy(&index->is_sorted, index->index.map + csds_index_is_sorted_offset,
         csds_index_is_sorted_size);

  /* Compute the position of the history of new particles. */
  size_t count = 0;
  for (int i = 0; i < csds_type_count; i++) {
    count += index->nparts[i];
  }
  count *= sizeof(struct index_data);

  /* Read the number of new particles. */
  memcpy(&index->nparts_created,
         index->index.map + csds_index_data_offset + count,
         csds_index_npart_size);

  /* Compute the position of the history of particles removed. */
  size_t count_new = 0;
  for (int i = 0; i < csds_type_count; i++) {
    count_new += index->nparts_created[i];
  }
  count_new *= sizeof(struct index_data);

  /* Read the number of particles removed. */
  memcpy(&index->nparts_removed,
         index->index.map + csds_index_data_offset + count +
             csds_index_npart_size + count_new,
         csds_index_npart_size);

  /* Close the file */
  csds_index_free(index);
}

/**
 * @brief Write that the file is sorted.
 *
 * WARNING The file must be mapped.
 *
 * @param index The #csds_index.
 */
void csds_index_write_sorted(struct csds_index *index) {
  /* Set the value */
  const char is_sorted = 1;

  /* Write the value */
  memcpy(index->index.map + csds_index_is_sorted_offset, &is_sorted,
         csds_index_is_sorted_size);
}

/**
 * @brief Get the #index_data of a particle type.
 *
 * The #index_data contains the ids and offset of the particle.
 * This function should be used when looking for the last known particles.
 *
 * @param index The #csds_index.
 * @param type The particle type.
 */
struct index_data *csds_index_get_data(struct csds_index *index, int type) {
  /* Get the offset of particles of this type by skipping entries of lower
   * types. */
  size_t count = 0;
  for (int i = 0; i < type; i++) {
    count += index->nparts[i];
  }
  const size_t type_offset = count * sizeof(struct index_data);

  const char *ret = index->index.map + csds_index_data_offset + type_offset;
  return (struct index_data *)ret;
}

/**
 * @brief Get the #index_data for the particles created.
 *
 * The #index_data contains the ids and offset of the particle.
 * This function should be used when looking for the particles created.
 *
 * @param index The #csds_index.
 * @param type The particle type.
 */
struct index_data *csds_index_get_created_history(struct csds_index *index,
                                                  int type) {

  /* Get the position after the previous array. */
  char *ret = (char *)csds_index_get_data(index, csds_type_count);

  /* Get the offset of particles of this type by skipping entries of lower
   * types. */
  size_t count = 0;
  for (int i = 0; i < type; i++) {
    count += index->nparts_created[i];
  }
  const size_t type_offset = count * sizeof(struct index_data);

  ret += type_offset + csds_index_npart_size;
  return (struct index_data *)ret;
}

/**
 * @brief Get the #index_data for the particles removed.
 *
 * The #index_data contains the ids and offset of the particle.
 * This function should be used when looking for the particles removed.
 *
 * @param index The #csds_index.
 * @param type The particle type.
 */
struct index_data *csds_index_get_removed_history(struct csds_index *index,
                                                  int type) {

  /* Get the position after the previous array. */
  char *ret = (char *)csds_index_get_created_history(index, csds_type_count);

  /* Get the offset of particles of this type by skipping entries of lower
   * types. */
  size_t count = 0;
  for (int i = 0; i < type; i++) {
    count += index->nparts_removed[i];
  }
  const size_t type_offset = count * sizeof(struct index_data);

  ret += type_offset + csds_index_npart_size;
  return (struct index_data *)ret;
}

/**
 * @brief Check if a time array is written at the end.
 *
 * @param index The #csds_index.
 */
int csds_index_contains_time_array(struct csds_index *index) {
  /* Only the first index file should have a time array */
  if (index->integer_time != 0) {
    error_python("Only the first index file can have a time array.");
  }

  /* Get the file size */
  const size_t file_size = index->index.mmap_size;

  /* Get the position after the previous array. */
  char *ret = (char *)csds_index_get_removed_history(index, csds_type_count);
  const size_t before_time_array = ret - index->index.map;

  return file_size != before_time_array;
}

/**
 * @brief Map the file and if required sort it.
 *
 * @param index The #csds_index.
 * @param filename The index filename.
 * @param sorted Does the file needs to be sorted?
 */
void csds_index_map_file(struct csds_index *index, const char *filename,
                         int sorted) {
  /* Un-map previous file */
  if (index->index.map != NULL) {
    error_python("Trying to remap.");
  }

  /* Check if need to sort the file */
  if (sorted && !index->is_sorted) {
    if (index->reader->verbose > 1) {
      message("Sorting the index file.");
    }
    /* Map the index file */
    csds_mapped_file_mmap_file(&index->index, filename,
                               /* read_only */ 0, /* track_mmap */ 0);
    /* Sort the file */
    for (int i = 0; i < csds_type_count; i++) {
      if (index->nparts[i] != 0) {
        struct index_data *data = csds_index_get_data(index, i);
        quick_sort(data, index->nparts[i]);
      }
    }

    /* Write that the file is sorted */
    csds_index_write_sorted(index);

    /* Free the index file before opening it again in read only */
    csds_index_free(index);

    if (index->reader->verbose > 1) {
      message("Sorting done.");
    }
  }

  /* Map the index file */
  csds_mapped_file_mmap_file(&index->index, filename,
                             /* read_only */ 1, /* track_mmap */ 0);
}

/**
 * @brief Cleanup the memory of a csds_index
 *
 * @param index The #csds_index.
 */
void csds_index_free(struct csds_index *index) {
  if (index->index.map == NULL) {
    error_python("Trying to unmap an unexisting map");
  }
  csds_mapped_file_munmap_file(&index->index);
}

/**
 * @brief Get the offset of a given particle
 *
 * @param index The #csds_index.
 * @param id The ID of the particle.
 * @param type The type of the particle.
 *
 * @return The offset of the particle or 0 if not found.
 */
size_t csds_index_get_particle_offset(struct csds_index *index, long long id,
                                      int type) {
  /* Define a few variables */
  const struct index_data *data = csds_index_get_data(index, type);
  size_t left = 0;
  size_t right = index->nparts[type] - 1;

  /* Search for the value (binary search) */
  while (left <= right) {
    size_t m = (left + right) / 2;
    if (data[m].id < id) {
      left = m + 1;
    } else if (data[m].id > id) {
      right = m - 1;
    } else {
      return data[m].offset;
    }
  }

  return 0;
}

/**
 * @brief Initialize the #csds_index.
 *
 * @param index The #csds_index.
 * @param reader The #csds_reader.
 */
void csds_index_init(struct csds_index *index,
                     const struct csds_reader *reader) {
  /* Set the mapped file to NULL */
  index->index.map = NULL;

  /* Set the pointer to the reader */
  index->reader = reader;

  /* Set the time to its default value */
  index->time = -1;
}

/**
 * @brief Give the offset of a particle given its id.
 *
 * @param index The #csds_index.
 * @param id The ID of the particle.
 * @param type The type of particles.
 * @param offset (output) The offset of the particle.
 *
 * @return Did we find the particle?
 */
int csds_index_get_offset(struct csds_index *index, int64_t id, int type,
                          uint64_t *offset) {

  /* Ensure that the file is sorted according to the ids. */
  if (!index->is_sorted) {
    error_python("The index file should be already sorted.");
  }
  /* Ensure that we have some particles */
  if (index->nparts[type] == 0) {
    error_python("Trying to find a particle in an empty index.");
  }

  /* Get the data structure. */
  struct index_data *data = csds_index_get_data(index, type);

  /* Perform a binary search. */
  size_t left = 0;
  size_t right = index->nparts[type] - 1;

  /* Find the id through a bisection method. */
  // TODO use interpolation search (same for the other binary searches)
  while (left <= right) {
    int64_t center = (left + right) >> 1;

    if (id > data[center].id) {
      left = center + 1;
    } else if (id < data[center].id) {
      right = center - 1;
    } else {
      left = right = center;
      break;
    }
  }

  /* Ensure that we got the correct id */
  if (left == right && data[left].id != id) {
    error("Failed to obtain the correct id");
  }

  /* Return the correct value if found. */
  if (left == right) {
    *offset = data[left].offset;
    return 1;
  }

  /* Try to find it in the history */
  // TODO Improve this with a sort?
  data = csds_index_get_created_history(index, type);
  for (uint64_t i = 0; i < index->nparts_created[type]; i++) {
    if (data[i].id == id) {
      *offset = data[i].offset;
      return 1;
    }
  }

  /* The particle was not found */
  return 0;
}
