/*******************************************************************************
 * This file is part of CSDS.
 * Copyright (c) 2019 Loic Hausammann (loic.hausammann@epfl.ch)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 ******************************************************************************/

#ifndef CSDS_CSDS_INDEX_H
#define CSDS_CSDS_INDEX_H

#include "csds_hashmap.h"
#include "csds_mapped_file.h"
#include "csds_tools.h"

/* predefine the structure */
struct csds_reader;

/**
 * @brief Structure dealing with the index files.
 *
 * The structure is initialized with #csds_index_init and
 * then a file can be read with #csds_index_read_header and
 * #csds_index_map_file.
 *
 * The functions #csds_index_get_particle_offset and
 * #csds_index_get_data should be used to access the element
 * stored inside the index file.
 * The first one access a particle through its ids and the second one
 * gives a pointer to the first element that can be looped through.
 */
struct csds_index {
  /* Pointer to the reader */
  const struct csds_reader *reader;

  /* Time of the index file */
  double time;

  /* Integer time of the index file */
  integertime_t integer_time;

  /* Number of particles in the file */
  uint64_t nparts[csds_type_count];

  /* Is the file sorted ? */
  char is_sorted;

  /* The mapped file */
  struct mapped_file index;

  /* Number of particles created. */
  uint64_t nparts_created[csds_type_count];

  /* Number of particles removed. */
  uint64_t nparts_removed[csds_type_count];
};

int csds_index_contains_time_array(struct csds_index *index);
void csds_index_write_sorted(struct csds_index *index);
void csds_index_init(struct csds_index *index,
                     const struct csds_reader *reader);
void csds_index_read_header(struct csds_index *index, const char *filename);
void csds_index_map_file(struct csds_index *index, const char *filename,
                         int sorted);
size_t csds_index_get_particle_offset(struct csds_index *index, long long id,
                                      int type);
void csds_index_free(struct csds_index *index);
void csds_index_sort_file(struct csds_index *index);
int csds_index_get_offset(struct csds_index *index, int64_t id, int type,
                          uint64_t *offset);
struct index_data *csds_index_get_data(struct csds_index *index, int type);
struct index_data *csds_index_get_created_history(struct csds_index *index,
                                                  int type);
struct index_data *csds_index_get_removed_history(struct csds_index *index,
                                                  int type);

#endif  // CSDS_CSDS_INDEX_H
