/*******************************************************************************
 * This file is part of CSDS.
 * Copyright (c) 2020 Loic Hausammann (loic.hausammann@epfl.ch)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 ******************************************************************************/

#ifndef CSDS_CSDS_INTERPOLATION_H
#define CSDS_CSDS_INTERPOLATION_H

#include "csds_parameters.h"
#include "csds_tools.h"

/**
 * @brief Limits the value of x to be between a and b
 *
 * Only wraps once. If x > a + 2(b - a), the returned value will be larger than
 * b. Similarly for x < a - (b - a), the value will be smaller than a.
 */
#define box_wrap(x, a, b)                                              \
  ({                                                                   \
    const __typeof__(x) _x = (x);                                      \
    const __typeof__(a) _a = (a);                                      \
    const __typeof__(b) _b = (b);                                      \
    _x < _a ? (_x + (_b - _a)) : ((_x >= _b) ? (_x - (_b - _a)) : _x); \
  })

/**
 * @brief Compute the quintic hermite spline interpolation.
 * See https://en.wikipedia.org/wiki/Hermite_interpolation for more information.
 * @f$ p(x) = f(x_0) + f'(x_0) (x - x_0) + \frac{1}{2}f''(x_0) (x - x_0)^2 +
 \frac{f(x_1) - f(x_0) - f'(x_0) (x_1 - x_0) - \frac{1}{2} f''(x_0) (x_1 -
 x_0)^2}{(x_1 - x_0)^3} (x - x_0)^3
  + \frac{3 f(x_0) - 3 f(x_1) + \left( 2 f'(x_0) + f'(x_1) \right) (x_1 - x_0) +
 \frac{1}{2} f''(x_0) (x_1 - x_0)^2}{(x_1 - x_0)^4} (x - x_0)^3 (x - x_1)
  + \frac{6 f(x_1) - 6 f(x_0) - 3 \left( f'(x_0) + f'(x_1) \right) (x_1 - x_0) +
 \frac{1}{2}\left( f''(x_1) - f''(x_0) \right) (x_1 - x_0)^2}{(x_1 - x_0)^5} (x
 - x_0)^3 (x - x_1)^2 @f$
 *
 * @param t0 The time at the left of the interval.
 * @param x0 The function at the left of the interval.
 * @param v0 The first derivative at the left of the interval.
 * @param a0 The second derivative at the left of the interval.
 * @param t1 The time at the right of the interval.
 * @param x1 The function at the right of the interval.
 * @param v1 The first derivative at the right of the interval.
 * @param a1 The second derivative at the right of the interval.
 * @param t The time of the interpolation.
 *
 * @return The function evaluated at t.
 */
INLINE static double interpolate_quintic_hermite_spline(
    const double t0, const double x0, const float v0, const float a0,
    const double t1, const double x1, const float v1, const float a1,
    const double t) {

  /* Generates recurring variables  */
  /* Time differences */
  const double dt = t1 - t0;
  const double dt2 = dt * dt;
  const double dt3 = dt2 * dt;
  const double dt4 = dt3 * dt;
  const double dt5 = dt4 * dt;

  const double t_t0 = t - t0;
  const double t_t0_2 = t_t0 * t_t0;
  const double t_t0_3 = t_t0_2 * t_t0;
  const double t_t1 = t - t1;
  const double t_t1_2 = t_t1 * t_t1;

  /* Derivatives */
  const double v0_dt = v0 * dt;
  const double a0_dt2 = 0.5 * a0 * dt2;
  const double v1_dt = v1 * dt;
  const double a1_dt2 = 0.5 * a1 * dt2;

  /* Do the first 3 terms of the hermite spline */
  double x = x0 + v0 * t_t0 + 0.5 * a0 * t_t0_2;

  /* Cubic term */
  x += (x1 - x0 - v0_dt - a0_dt2) * t_t0_3 / dt3;

  /* Quartic term */
  x += (3. * x0 - 3. * x1 + v1_dt + 2. * v0_dt + a0_dt2) * t_t0_3 * t_t1 / dt4;

  /* Quintic term */
  x += (6. * x1 - 6. * x0 - 3. * v0_dt - 3. * v1_dt + a1_dt2 - a0_dt2) *
       t_t0_3 * t_t1_2 / dt5;

  return x;
}

/**
 * @brief Compute the cubic hermite spline interpolation.
 * See https://en.wikipedia.org/wiki/Hermite_interpolation for more information.
 * @f$ p(x) = f(x_0) + f'(x_0) (x - x_0) + \frac{f(x_1) - f(x_0) - f'(x_0) (x_1
 - x_0)}{(x_1 - x_0)^2} (x - x_0)^2
 + \frac{2 f(x_0) - 2 f(x_1) + f'(x_0) (x_1 - x_0) + f'(x_1) (x_1 - x_0)}{(x_1
 - x_0)^3} (x - x_0)^2 (x - x_1) @f$
 *
 * @param t0 The time at the left of the interval.
 * @param v0 The first derivative at the left of the interval.
 * @param a0 The second derivative at the left of the interval.
 * @param t1 The time at the right of the interval.
 * @param v1 The first derivative at the right of the interval.
 * @param a1 The second derivative at the right of the interval.
 * @param t The time of the interpolation.
 *
 * @return The function evaluated at t.
 */
#define interpolate_cubic_hermite_spline_definition(type_v, type_a)       \
  INLINE static type_v interpolate_cubic_hermite_##type_v(                \
      const double t0, const type_v v0, const type_a a0, const double t1, \
      const type_v v1, const type_a a1, const double t) {                 \
                                                                          \
    /* Generates recurring variables  */                                  \
    /* Time differences */                                                \
    const double dt = t1 - t0;                                            \
    const double dt2 = dt * dt;                                           \
    const double dt3 = dt2 * dt;                                          \
                                                                          \
    const double t_t0 = t - t0;                                           \
    const double t_t0_2 = t_t0 * t_t0;                                    \
    const double t_t1 = t - t1;                                           \
                                                                          \
    /* Derivatives */                                                     \
    const type_a a0_dt = a0 * dt;                                         \
    const type_a a1_dt = a1 * dt;                                         \
                                                                          \
    /* Do the first 2 terms of the hermite spline */                      \
    type_v x = v0 + a0 * t_t0;                                            \
                                                                          \
    /* Square term */                                                     \
    x += (v1 - v0 - a0_dt) * t_t0_2 / dt2;                                \
                                                                          \
    /* Cubic term */                                                      \
    x += (2. * v0 - 2. * v1 + a1_dt + a0_dt) * t_t0_2 * t_t1 / dt3;       \
                                                                          \
    return x;                                                             \
  }

interpolate_cubic_hermite_spline_definition(float, float);
interpolate_cubic_hermite_spline_definition(double, float);

/**
 * @brief Interpolates a field in N dimension using the first and second
 * derivatives if possible.
 *
 * The field is in double and both derivatives in float.
 *
 * @param before Pointer to the #csds_reader_field at a time < t.
 * @param after Pointer to the #csds_reader_field at a time > t.
 * @param otuput Pointer to the output value.
 * @param t_before Time of field_before (< t).
 * @param t_after Time of field_after (> t).
 * @param t Requested time.
 * @param periodic Should the periodic boundary be applied?
 * @param params The simulation's #csds_parameters.
 */
INLINE static void interpolate_quintic_double_float_ND(
    const double t_before, const struct csds_reader_field *restrict before,
    const double t_after, const struct csds_reader_field *restrict after,
    void *restrict output, const double t, const int dimension, int periodic,
    const struct csds_parameters *params) {

  /* Get the arrays */
  const double *x_bef = (double *)before->field;
  const float *v_bef = (float *)before->first_deriv;
  const float *a_bef = (float *)before->second_deriv;

  const double *x_aft = (double *)after->field;
  const float *v_aft = (float *)after->first_deriv;
  const float *a_aft = (float *)after->second_deriv;

  double *x = (double *)output;

  /* Loop over the dimensions */
  for (int i = 0; i < dimension; i++) {
    /* Get the current variables */
    double x0 = x_bef[i];
    double x1 = x_aft[i];
    float v0 = v_bef ? v_bef[i] : -1;
    float v1 = v_aft ? v_aft[i] : -1;
    float a0 = a_bef ? a_bef[i] : -1;
    float a1 = a_aft ? a_aft[i] : -1;

    /* Add cosmological factor for velocity and acceleration */
    if (params->with_cosmology && v_bef && v_aft) {
      csds_cosmology_add_factor_position(&params->cosmo, t_before, &v0,
                                         a_bef ? &a0 : NULL);
      csds_cosmology_add_factor_position(&params->cosmo, t_after, &v1,
                                         a_aft ? &a1 : NULL);
    }

    /* Apply the periodic boundaries */
    if (periodic) {
      /* Move towards the origin for more accuracy */
      const double half = 0.5 * params->box_size[i];
      if (x0 > x1 + half)
        x0 -= params->box_size[i];
      else if (x1 > x0 + half)
        x1 -= params->box_size[i];
    }

    /* Use quintic hermite spline. */
    if (v_bef && v_aft && a_bef && a_aft) {
      x[i] = interpolate_quintic_hermite_spline(t_before, x0, v0, a0, t_after,
                                                x1, v1, a1, t);
    }
    /* Use cubic hermite spline. */
    else if (v_bef && v_aft) {
      x[i] = interpolate_cubic_hermite_double(t_before, x0, v0, t_after, x1, v1,
                                              t);
    }
    /* Use linear interpolation. */
    else {
      const double wa = (t - t_before) / (t_after - t_before);
      const double wb = 1. - wa;
      x[i] = wa * x1 + wb * x0;
    }

    /* Apply the periodic boundaries to the output */
    if (periodic) {
      /* We suppose that only 1 wrapping is enough
         otherwise something really bad happened */
      x[i] = box_wrap(x[i], 0., params->box_size[i]);

#ifdef CSDS_DEBUG_CHECKS
      /* Check if the periodic box is correctly applied */
      if (x[i] >= params->box_size[i] || x[i] < 0) {
        error_python(
            "A particle is outside the periodic box (%g):"
            "before=%g, after=%g interpolation=%g",
            params->box_size[i], x0, x1, x[i]);
      }
#endif
    }
  }
}

/**
 * @brief Interpolates a field in N dimension using the first derivative if
 * possible.
 *
 * The field and the first derivatives are floats.
 *
 * @param before Pointer to the #csds_reader_field at a time < t.
 * @param after Pointer to the #csds_reader_field at a time > t.
 * @param otuput Pointer to the output value.
 * @param t_before Time of field_before (< t).
 * @param t_after Time of field_after (> t).
 * @param t Requested time.
 * @param periodic Should the periodic boundary be applied?
 * @param params The simulation's #csds_parameters.
 */
INLINE static void interpolate_cubic_float_ND(
    const double t_before, const struct csds_reader_field *restrict before,
    const double t_after, const struct csds_reader_field *restrict after,
    void *restrict output, const double t, const int dimension,
    ATTR_UNUSED int periodic, const struct csds_parameters *params) {

#ifdef CSDS_DEBUG_CHECKS
  /* The position is expected to be a double => error with periodic */
  if (periodic) {
    error_python("Not implemented yet");
  }
#endif

  /* Compute the interpolation scaling. */
  const float wa = (t - t_before) / (t_after - t_before);
  const float wb = 1. - wa;

  const float *v_bef = (float *)before->field;
  const float *a_bef = (float *)before->first_deriv;
  const float *v_aft = (float *)after->field;
  const float *a_aft = (float *)after->first_deriv;

  for (int i = 0; i < dimension; i++) {
    float *v = (float *)output;
    float a0 = a_bef ? a_bef[i] : -1;
    float a1 = a_aft ? a_aft[i] : -1;

    /* Add cosmological factor for acceleration */
    if (params->with_cosmology && a_bef && a_aft) {
      csds_cosmology_add_factor_velocity(&params->cosmo, t_before, &a0);
      csds_cosmology_add_factor_velocity(&params->cosmo, t_after, &a0);
    }

    /* Use a cubic hermite spline. */
    if (a_bef && a_aft) {
      v[i] = interpolate_cubic_hermite_float(t_before, v_bef[i], a0, t_after,
                                             v_aft[i], a1, t);
    }
    /* Use linear interpolation. */
    else {
      v[i] = wa * v_aft[i] + wb * v_bef[i];
    }
  }
}

/**
 * @brief Interpolates a field in N dimension.
 *
 * The field is in float.
 *
 * @param before Pointer to the #csds_reader_field at a time < t.
 * @param after Pointer to the #csds_reader_field at a time > t.
 * @param otuput Pointer to the output value.
 * @param t_before Time of field_before (< t).
 * @param t_after Time of field_after (> t).
 * @param t Requested time.
 * @param periodic Should the periodic boundary be applied?
 * @param params The simulation's #csds_parameters.
 */
INLINE static void interpolate_linear_float_ND(
    const double t_before, const struct csds_reader_field *restrict before,
    const double t_after, const struct csds_reader_field *restrict after,
    void *restrict output, const double t, const int dimension,
    ATTR_UNUSED int periodic,
    ATTR_UNUSED const struct csds_parameters *params) {

#ifdef CSDS_DEBUG_CHECKS
  /* The position is expected to be a double => error with periodic */
  if (periodic) {
    error_python("Not implemented yet");
  }
#endif

  /* Compute the interpolation scaling. */
  const float wa = (t - t_before) / (t_after - t_before);
  const float wb = 1. - wa;

  /* interpolate vectors. */
  for (int i = 0; i < dimension; i++) {
    float *a = (float *)output;
    const float *a_bef = (float *)before->field;
    const float *a_aft = (float *)after->field;
    a[i] = wa * a_aft[i] + wb * a_bef[i];
  }
}

/**
 * @brief Interpolates a field in N dimension.
 *
 * The field is in double.
 *
 * @param before Pointer to the #csds_reader_field at a time < t.
 * @param after Pointer to the #csds_reader_field at a time > t.
 * @param otuput Pointer to the output value.
 * @param t_before Time of field_before (< t).
 * @param t_after Time of field_after (> t).
 * @param t Requested time.
 * @param periodic Should the periodic boundary be applied?
 * @param params The simulation's #csds_parameters.
 */
INLINE static void interpolate_linear_double_ND(
    const double t_before, const struct csds_reader_field *restrict before,
    const double t_after, const struct csds_reader_field *restrict after,
    void *restrict output, const double t, const int dimension,
    ATTR_UNUSED int periodic,
    ATTR_UNUSED const struct csds_parameters *params) {

#ifdef CSDS_DEBUG_CHECKS
  /* The velocity and acceleration are supposed to be provided => quintic =>
   * error here */
  if (periodic) {
    error_python("Not implemented yet");
  }
#endif

  /* Compute the interpolation scaling. */
  const double wa = (t - t_before) / (t_after - t_before);
  const double wb = 1. - wa;

  /* interpolate vectors. */
  for (int i = 0; i < dimension; i++) {
    double *a = (double *)output;
    const double *a_bef = (double *)before->field;
    const double *a_aft = (double *)after->field;
    a[i] = wa * a_aft[i] + wb * a_bef[i];
  }
}

/**
 * @brief Interpolates a field stored as a float.
 *
 * @param before Pointer to the #csds_reader_field at a time < t.
 * @param after Pointer to the #csds_reader_field at a time > t.
 * @param otuput Pointer to the output value.
 * @param t_before Time of field_before (< t).
 * @param t_after Time of field_after (> t).
 * @param t Requested time.
 * @param periodic Should the periodic boundary be applied?
 * @param params The simulation's #csds_parameters.
 */
INLINE static void interpolate_linear_float(
    const double t_before, const struct csds_reader_field *restrict before,
    const double t_after, const struct csds_reader_field *restrict after,
    void *restrict output, const double t, ATTR_UNUSED int periodic,
    ATTR_UNUSED const struct csds_parameters *params) {

#ifdef CSDS_DEBUG_CHECKS
  /* The position is expected to be a double => error with periodic */
  if (periodic) {
    error_python("Not implemented yet");
  }
#endif

  /* Compute the interpolation scaling. */
  const float wa = (t - t_before) / (t_after - t_before);
  const float wb = 1. - wa;
  ((float *)output)[0] =
      wa * ((float *)after->field)[0] + wb * ((float *)before->field)[0];
}

#endif  // CSDS_CSDS_INTERPOLATION_H
