/*******************************************************************************
 * This file is part of CSDS.
 * Copyright (c) 2019 Loic Hausammann (loic.hausammann@epfl.ch)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 ******************************************************************************/
#include "csds_logfile.h"

#include "csds_mapped_file.h"
#include "csds_reader.h"

/**
 * @brief Initialize the #csds_logfile.
 *
 * If required this function will also reverse the offsets.
 * @param log The #csds_logfile.
 * @param filename the log's filename.
 * @param reader The #csds_reader.
 * @param only_header Read only the header.
 */
void csds_logfile_init_from_file(struct csds_logfile *log, char *filename,
                                 struct csds_reader *reader, int only_header) {

  /* Set the pointer to the reader. */
  log->reader = reader;
  if (&reader->log != log) error_python("Wrong link to the reader.");

  /* Set pointers to zero. */
  time_array_init(&log->times, CSDS_TIME_INIT_SIZE);

  /* Open file, map it and get its size. */
  if (reader->verbose > 1) message("Mapping the log file.");
  csds_mapped_file_mmap_file(&log->log, filename,
                             /* read_only */ 1, /* track_mmap */ 1);

  /* Read the header. */
  if (reader->verbose > 1) message("Reading the header.");
  header_read(&log->header, log);

  /* Print the header. */
  if (reader->verbose > 0) {
    header_print(&log->header);
  }

  /* No need to continue if only the
     header is required. */
  if (only_header) return;

  /* Check if the offset are corrupted. */
  if (header_is_corrupted(&log->header)) {
    error_python("The offsets have been corrupted.");
  }

  /* Reverse the offsets direction. */
  if (header_is_backward(&log->header)) {
    csds_logfile_reverse_offset(log, filename);
  }

  /* Initialize the time array. */
  if (reader->verbose > 1) message("Reading the timestamps.");
  time_array_populate(&log->times, log);

  /* Print the time array. */
  if (reader->verbose > 0) {
    time_array_print(&log->times);
  }
}

/**
 * @brief Free the allocated memory and unmap the file.
 *
 * @param log The #csds_logfile.
 */
void csds_logfile_free(struct csds_logfile *log) {
  csds_mapped_file_munmap_file(&log->log);

  time_array_free(&log->times);

  header_free(&log->header);
}

/**
 * @brief debugging function checking the offset and the mask of a record.
 *
 * Compare the mask with the one pointed by the header.
 * if the record is a particle, check the id too.
 *
 * @param reader The #csds_reader.
 * @param offset position of the record.
 *
 * @return position after the record.
 */
void csds_logfile_check_record_consistency(
    ATTR_UNUSED struct csds_logfile *log) {
#ifdef CSDS_DEBUG_CHECKS
  struct header *header = &log->header;
  struct csds_reader *reader = log->reader;

  if (reader->verbose > 0) {
    message("Check record's headers...");
  }

  /* Prepare the progress bar. */
  float next_percentage = 0.;
  const float delta_percentage = 0.5;
  const ticks init_time = getticks();

  /* check that the record offset points to another record. */
  for (size_t offset_debug = header->offset_first_record;
       offset_debug < log->log.mmap_size;
       offset_debug = tools_check_record_consistency(reader, offset_debug)) {

    /* Check if we should update the progress bar. */
    float current = 100 * ((float)offset_debug) / log->log.mmap_size;
    if (reader->verbose > 0 && current > next_percentage) {
      /* Print the bar */
      tools_print_progress(current, init_time, "Checking offsets");

      /* Compute the next update of the progress bar */
      next_percentage += delta_percentage;
    }
  }

  /* Print final message and close progress bar */
  if (reader->verbose > 0) {
    printf("\n");
    message("Record's headers are correct.");
  }

#else
  error("This function should not be called outside of the debugging checks");
#endif
}

/**
 * @brief Reverse offset in log file
 *
 * @param log The #csds_logfile
 * @param filename The log's filename.
 */
void csds_logfile_reverse_offset(struct csds_logfile *log, char *filename) {

  /* Close and reopen the file in write mode. */
  csds_mapped_file_munmap_file(&log->log);
  csds_mapped_file_mmap_file(&log->log, filename,
                             /* read_only */ 0, /* track_mmap */ 0);

  /* Get pointers */
  struct header *header = &log->header;
  const struct csds_reader *reader = log->reader;
  if (&reader->log != log) error_python("Wrong link to the reader.");

  /* Check if the offsets need to be reversed. */
  if (!header_is_backward(header)) {
    error_python("The offsets are already reversed.");
  }

#ifdef CSDS_DEBUG_CHECKS
  csds_logfile_check_record_consistency(log);
#endif

  message("WARNING: Modifying the logfile, do not kill the job!");

  /* Set the offset direction to a corrupted status. */
  header_change_offset_direction(header, csds_offset_corrupted);

  if (reader->verbose > 0) {
    message("Reversing offsets...");
  }

  /* Prepare the progress message. */
  float next_percentage = 0.;
  const float delta_percentage = 0.2;
  const ticks init_time = getticks();

  /* reverse the record's offset. */
  for (size_t offset = header->offset_first_record; offset < log->log.mmap_size;
       offset = tools_reverse_offset(header, &log->log, offset)) {
    /* Check if we should update the progress. */
    float current = 100 * ((float)offset) / log->log.mmap_size;
    if (reader->verbose > 0 && current > next_percentage) {

      /* Print the remaining time */
      tools_print_progress(current, init_time, "Reversing the offsets");

      /* Compute the next update of the progress */
      next_percentage += delta_percentage;
    }
  }

  /* Print final message */
  if (reader->verbose > 0) {
    printf("\n");
    message("Reversing done.");
  }

  /* Now that the offset are effectively reversed, can set the direction to
     forward. */
  header_change_offset_direction(header, csds_offset_forward);

  message("WARNING: Modification done, you can now safely kill the job.");

#ifdef CSDS_DEBUG_CHECKS
  csds_logfile_check_record_consistency(log);
#endif

  /* Close and reopen the file in read mode. */
  csds_mapped_file_munmap_file(&log->log);
  csds_mapped_file_mmap_file(&log->log, filename,
                             /* read_only */ 1,
                             /* track_mmap */ 1);
}
