/*******************************************************************************
 * This file is part of CSDS.
 * Copyright (c) 2019 Loic Hausammann (loic.hausammann@epfl.ch)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 ******************************************************************************/
/**
 * @file csds_mapped_file.h
 * @brief This file contains basic IO function.
 */
#ifndef CSDS_LOADER_IO_H
#define CSDS_LOADER_IO_H

#include "csds_fields.h"
#include "csds_header.h"
#include "csds_tools.h"

/* Standard headers */
#include <stdio.h>
#include <stdlib.h>
#include <sys/mman.h>
#include <unistd.h>

#define CSDS_VERBOSE_TIMERS -1
#define CSDS_TRACKING_SIZE_BYTES (1024 * 1024 * 1024)

/* Structure for mapping a file. */
struct mapped_file {
  /* Mapped data. */
  char *map;

  /* File size. */
  size_t mmap_size;

#ifdef CSDS_MMAP_TRACKING
  struct {
    /* File for the tracking (should not use mmap file) */
    FILE *f;

    /* Size of the file? */
    size_t size;

    /* Current number of elements  */
    int64_t number_elements;
  } tracking;
#endif
};

#ifdef CSDS_MMAP_TRACKING
/**
 * @brief Structure written in the tracking file.
 */
struct mmap_tracking_element {
  /* Number of ticks for accessing the page */
  int64_t ticks;

  /* Offset accessed */
  long long offset;
};

/** @brief Initialize a tracking element */
INLINE static void mmap_tracking_element_init(struct mmap_tracking_element *el,
                                              size_t offset) {
  el->ticks = getticks();
  el->offset = offset;
}

/** @brief Write a tracking element into the file */
INLINE static void mmap_tracking_element_write(struct mmap_tracking_element *el,
                                               struct mapped_file *file) {
#pragma omp critical
  {
    /* Stop when file becomes too huge */
    if (file->tracking.size < CSDS_TRACKING_SIZE_BYTES) {

      /* Write the element */
      fwrite(el, 1, sizeof(struct mmap_tracking_element), file->tracking.f);

      /* Update the counter */
      file->tracking.number_elements++;

      /* Write it back to the file */
      const long int pos = ftell(file->tracking.f);
      /* See csds_mapped_file_mmap_file for the value */
      fseek(file->tracking.f, 20, SEEK_SET);
      fwrite(&file->tracking.number_elements, 1,
             sizeof(file->tracking.number_elements), file->tracking.f);
      fseek(file->tracking.f, pos, SEEK_SET);

      /* Update the file size */
      file->tracking.size += sizeof(struct mmap_tracking_element);
    }
  }
}
#endif

size_t csds_loader_io_get_file_size(int fd);
void csds_mapped_file_mmap_file(struct mapped_file *map, const char *filename,
                                int read_only, int track_mmap);
void csds_mapped_file_munmap_file(struct mapped_file *map);

/**
 * @brief read a mask with its offset.
 *
 * @param mmap The #mapped_file.
 * @param offset The offset in the file
 * @param mask (output) mask read from the file.
 * @param diff_offset (output) offset difference to previous/next corresponding
 * record.
 *
 * @return The offset after the record header.
 */
INLINE static size_t csds_mapped_file_read_mask(struct mapped_file *mmap,
                                                size_t offset, mask_type *mask,
                                                size_t *diff_offset) {
#ifdef CSDS_MMAP_TRACKING
  struct mmap_tracking_element el;
  mmap_tracking_element_init(&el, offset);
#endif

  /* read mask */
  if (mask) {
    *mask = 0;
    memcpy(mask, mmap->map + offset, CSDS_MASK_SIZE);
  }
  offset += CSDS_MASK_SIZE;

  /* read offset */
  if (diff_offset) {
    *diff_offset = 0;
    memcpy(diff_offset, mmap->map + offset, CSDS_OFFSET_SIZE);
  }
  offset += CSDS_OFFSET_SIZE;

#ifdef CSDS_MMAP_TRACKING
  /* Write the result into the file */
  el.ticks = getticks() - el.ticks;
  if (mmap->tracking.f) {
    mmap_tracking_element_write(&el, mmap);
  }
#endif

  return offset;
}

/**
 * @brief read a single value from a file.
 *
 * @param mmap The #mapped_file.
 * @param offset The offset in the file
 * @param size size of the data to read.
 * @param p pointer where to store the data.

 * @return The offset after the data read.
 */
INLINE static size_t csds_mapped_file_read_data(struct mapped_file *mmap,
                                                size_t offset,
                                                const size_t size, void *p) {
#ifdef CSDS_MMAP_TRACKING
  struct mmap_tracking_element el;
  mmap_tracking_element_init(&el, offset);
#endif

  memcpy(p, mmap->map + offset, size);

#ifdef CSDS_MMAP_TRACKING
  /* Write the result into the file */
  el.ticks = getticks() - el.ticks;
  if (mmap->tracking.f) {
    mmap_tracking_element_write(&el, mmap);
  }
#endif

  return offset + size;
};

/**
 * @brief write a single value in a file.
 *
 * @param mmap The #mapped_file.
 * @param offset The offset in the file
 * @param size size of the data to write.
 * @param p pointer to the data.
 *
 * @return memory after the data written.
 */
INLINE static size_t csds_mapped_file_write_data(struct mapped_file *mmap,
                                                 size_t offset,
                                                 const size_t size,
                                                 const void *p) {

#ifdef CSDS_MMAP_TRACKING
  struct mmap_tracking_element el;
  mmap_tracking_element_init(&el, offset);
#endif

  memcpy(mmap->map + offset, p, size);

#ifdef CSDS_MMAP_TRACKING
  /* Write the result into the file */
  el.ticks = getticks() - el.ticks;
  if (mmap->tracking.f) {
    mmap_tracking_element_write(&el, mmap);
  }
#endif

  return offset + size;
};

/* Defines a few advice for the mmap */
#define CSDS_ADVICE_SEQUENTIAL(log)                                  \
  {                                                                  \
    /* Warn the OS that we will read in a sequential way */          \
    int test_ret = madvise(log.map, log.mmap_size, MADV_SEQUENTIAL); \
    if (test_ret != 0) {                                             \
      error("Failed to advise the mmap");                            \
    }                                                                \
  }

#define CSDS_ADVICE_NORMAL(log)                                  \
  {                                                              \
    /* Warn the OS that we will read in a normal way */          \
    int test_ret = madvise(log.map, log.mmap_size, MADV_NORMAL); \
    if (test_ret != 0) {                                         \
      error("Failed to advise the mmap");                        \
    }                                                            \
  }

#define CSDS_ADVICE_WILL_NEED(mmap, addr, length)                         \
  {                                                                       \
    /* Adjust the size to be at least the page size. */                   \
    const size_t page_size = sysconf(_SC_PAGE_SIZE);                      \
    const size_t mem_align = (addr + (page_size - 1)) & ~(page_size - 1); \
                                                                          \
    /* Warn the OS that we will read soon this address  */                \
    int test_ret =                                                        \
        madvise(mmap + mem_align, length + page_size, MADV_WILLNEED);     \
    if (test_ret != 0) {                                                  \
      error("Failed to advise the mmap");                                 \
    }                                                                     \
  }

#endif  // CSDS_LOADER_IO_H
