/*******************************************************************************
 * This file is part of CSDS.
 * Copyright (c) 2019 Loic Hausammann (loic.hausammann@epfl.ch)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 ******************************************************************************/

/* Include corresponding header */
#include "csds_parameters.h"

/* Include the csds */
#include "csds_parser.h"
#include "csds_reader.h"
#include "csds_tools.h"

/**
 * @brief Initialize the parameters from the yaml file.
 *
 * @param reader The #csds_reader.
 * @param params The #csds_parameters to initialize.
 *
 */
void csds_parameters_init(struct csds_parameters *params,
                          const struct csds_reader *reader) {

  /* Generate filename */
  char filename[CSDS_STRING_SIZE];
  strcpy(filename, reader->basename);
  size_t len = strlen(filename);
  /* Remove the rank number */
  filename[len - 5] = '\0';
  strcat(filename, ".yml");

  /* Initialize the parser */
  struct csds_parser csds_parser;
  parser_read_file(filename, &csds_parser);

#ifdef CSDS_DEBUG_CHECKS
  /* Print the parameters */
  if (reader->verbose > 0) parser_print_params(&csds_parser);
#endif

  /* Read the number of dimension */
  params->dimension = parser_get_param_int(&csds_parser, "Header:Dimension");

  /* Read the box size */
  parser_get_param_double_array(&csds_parser, "Header:BoxSize", 3,
                                params->box_size);

  /* Read the periodicity */
  params->periodic = parser_get_param_int(&csds_parser, "Header:Periodic");

  /* Read if we are running with cosmology */
  params->with_cosmology =
      parser_get_param_int(&csds_parser, "Policy:cosmological integration");

  /* Initialize the cosmology */
  if (params->with_cosmology) csds_cosmology_init(&params->cosmo, &csds_parser);

  /* Read the initial number of particles. */
  for (int i = 0; i < csds_type_count; i++) {
    params->approximate_number_particles[i] = 0;
  }
  params->approximate_number_particles[csds_type_gas] =
      parser_get_param_longlong(&csds_parser, "Header:NumberParts");
  params->approximate_number_particles[csds_type_stars] =
      parser_get_param_longlong(&csds_parser, "Header:NumberSParts");
  params->approximate_number_particles[csds_type_dark_matter] =
      parser_get_param_longlong(&csds_parser, "Header:NumberGParts");

  /* Read the cache parameters */
  params->cache_over_alloc =
      parser_get_opt_param_float(&csds_parser, "Cache:OverAllocation", 1.2);

  long long init_alloc[csds_type_count];
  for (int i = 0; i < csds_type_count; i++) {
    init_alloc[i] = 0;
  }
  parser_get_opt_param_longlong_array(&csds_parser, "Cache:InitialAllocation",
                                      csds_type_count, init_alloc);
  for (int i = 0; i < csds_type_count; i++) {
    params->cache_init_size[i] = init_alloc[i];
  }
}
