/*******************************************************************************
 * This file is part of CSDS and was copied from SWIFT.
 * Copyright (c) 2021 loic.hausammann@epfl.ch
 *               SWIFT
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 ******************************************************************************/

/* Config parameters. */
#include "../config.h"

/* Some standard headers. */
/* Needs to be included so that strtok returns char * instead of a int *. */
#include <ctype.h>
#include <stdlib.h>
#include <string.h>

/* This object's header. */
#include "csds_parser.h"

/* Local headers. */
#include "csds_error.h"

#define FIELD_BUFFER_SIZE 64
#define PARSER_COMMENT_STRING "#"
#define PARSER_COMMENT_CHAR '#'
#define PARSER_VALUE_CHAR ':'
#define PARSER_VALUE_STRING ":"
#define PARSER_START_OF_FILE "---"
#define PARSER_END_OF_FILE "..."

#define CHUNK 10

/* Alias of long long needed and used only for the preprocessor definitions */
typedef long long longlong;

/* Private functions. */
static int is_empty(const char *str);
static int count_indentation(const char *str);
static void parse_line(char *line, struct csds_parser *params);
static void parse_value(char *line, struct csds_parser *params);
static void parse_section_param(char *line, int *isFirstParam,
                                char *sectionName, struct csds_parser *params);
static void find_duplicate_params(const struct csds_parser *params,
                                  const char *param_name);
static void find_duplicate_section(const struct csds_parser *params,
                                   const char *section_name);
static int lineNumber = 0;

/**
 * @brief trim leading white space from a string.
 *
 * Returns pointer to first character.
 *
 * @param s the string.
 * @result the result.
 */
char *trim_leading(char *s) {
  if (s == NULL || strlen(s) < 2) return s;
  while (isspace(*s)) s++;
  return s;
}

/**
 * @brief trim trailing white space from a string.
 *
 * Modifies the string by adding a NULL to the end.
 *
 * @param s the string.
 * @result the result.
 */
char *trim_trailing(char *s) {
  if (s == NULL || strlen(s) < 2) return s;
  char *end = s + strlen(s) - 1;
  while (isspace(*end)) end--;
  *(end + 1) = '\0';
  return s;
}

/**
 * @brief trim leading and trailing white space from a string.
 *
 * Can modify the string by adding a NULL to the end.
 *
 * @param s the string.
 * @result the result.
 */
char *trim_both(char *s) {
  if (s == NULL || strlen(s) < 2) return s;
  return trim_trailing(trim_leading(s));
}

/**
 * @brief Initialize the parser structure.
 *
 * @param file_name Name of file to be read
 * @param params Structure to be populated from file
 */
void parser_init(const char *file_name, struct csds_parser *params) {
  params->paramCount = 0;
  params->sectionCount = 0;
  strcpy(params->fileName, file_name);
}

/**
 * @brief Reads an input file and stores each parameter in a structure.
 *
 * @param file_name Name of file to be read
 * @param params Structure to be populated from file
 */
void parser_read_file(const char *file_name, struct csds_parser *params) {
  /* Open file for reading */
  FILE *file = fopen(file_name, "r");

  /* Line to parsed. */
  char line[PARSER_MAX_LINE_SIZE];

  /* Initialise parameter count. */
  parser_init(file_name, params);

  /* Check if parameter file exits. */
  if (file == NULL) {
    error("Error opening parameter file: %s", file_name);
  }

  /* Read until the end of the file is reached.*/
  while (!feof(file)) {
    if (fgets(line, PARSER_MAX_LINE_SIZE, file) != NULL) {
      lineNumber++;
      parse_line(line, params);
    }
  }

  fclose(file);
}

/**
 * @brief Counts the number of white spaces that prefix a string.
 *
 * @param str String to be checked
 *
 * @return Number of white spaces prefixing str
 */

static int count_indentation(const char *str) {
  int count = 0;

  /* Check if the line contains the character */
  while (*(++str) == ' ') {
    count++;
  }
  return count;
}

/**
 * @brief Checks if a string is empty.
 *
 * @param str String to be checked
 *
 * @return Returns 1 if str is empty, 0 otherwise
 */

static int is_empty(const char *str) {
  int retParam = 1;
  while (*str != '\0') {
    if (!isspace(*str)) {
      retParam = 0;
      break;
    }
    str++;
  }

  return retParam;
}

/**
 * @brief Look for duplicate parameters.
 *
 * @param params Structure that holds the parameters
 * @param param_name Name of parameter to be searched for
 */

static void find_duplicate_params(const struct csds_parser *params,
                                  const char *param_name) {
  for (int i = 0; i < params->paramCount; i++) {
    if (!strcmp(param_name, params->data[i].name)) {
      error("Invalid line:%d '%s', parameter is a duplicate.", lineNumber,
            param_name);
    }
  }
}

/**
 * @brief Look for duplicate sections.
 *
 * @param params Structure that holds the parameters
 * @param section_name Name of section to be searched for
 */

static void find_duplicate_section(const struct csds_parser *params,
                                   const char *section_name) {
  for (int i = 0; i < params->sectionCount; i++) {
    if (!strcmp(section_name, params->section[i].name)) {
      error("Invalid line:%d '%s', section is a duplicate.", lineNumber,
            section_name);
    }
  }
}
/**
 * @brief Parses a line from a file and stores any parameters in a structure.
 *
 * @param line Line to be parsed.
 * @param params Structure to be populated from file.
 */

static void parse_line(char *line, struct csds_parser *params) {
  /* Parse line if it doesn't begin with a comment. */
  if (line[0] != PARSER_COMMENT_CHAR) {
    char trim_line[PARSER_MAX_LINE_SIZE];
    char tmp_str[PARSER_MAX_LINE_SIZE];
    char *token;

    /* Remove comments at the end of a line. */
    token = strtok(line, PARSER_COMMENT_STRING);
    strcpy(tmp_str, token);

    /* Check if the line is just white space. */
    if (!is_empty(tmp_str)) {
      /* Trim '\n' characters from string. */
      token = strtok(tmp_str, "\n");
      strcpy(trim_line, token);

      /* Check if the line contains a value and parse it. */
      if (strchr(trim_line, PARSER_VALUE_CHAR)) {

        /* Trim trailing space before parsing line for a value. */
        char no_space_line[PARSER_MAX_LINE_SIZE];
        strcpy(no_space_line, trim_trailing(trim_line));

        parse_value(no_space_line, params);
      }
      /* Check for invalid lines,not including the start and end of file. */
      else if (strcmp(trim_line, PARSER_START_OF_FILE) &&
               strcmp(trim_line, PARSER_END_OF_FILE)) {
        error("Invalid line:%d '%s'.", lineNumber, trim_line);
      }
    }
  }
}

/**
 * @brief Performs error checking and stores a parameter in a structure.
 *
 * @param line Line containing the parameter
 * @param params Structure to be written to
 *
 */
static void parse_value(char *line, struct csds_parser *params) {
  static int inSection = 0;
  static char section[PARSER_MAX_LINE_SIZE]; /* Keeps track of current section
                                                name. */
  static int isFirstParam = 1;
  char tmpStr[PARSER_MAX_LINE_SIZE];
  char tmpSectionName[PARSER_MAX_LINE_SIZE];

  char *token;

  /* Check that standalone parameters have correct indentation. */
  if (!inSection && line[0] == ' ') {
    error(
        "Invalid line:%d '%s', standalone parameter defined with incorrect "
        "indentation.",
        lineNumber, line);
  }

  /* Check that it is a parameter inside a section.*/
  if (line[0] == ' ' || line[0] == '\t') {
    parse_section_param(line, &isFirstParam, section, params);
  } else {
    /* It is the start of a new section or standalone parameter.
     * Take first token as the parameter name. */
    token = strtok(line, ":\t");
    strcpy(tmpStr, trim_trailing(token));

    /* Take second token as the parameter value. */
    token = trim_both(strtok(NULL, "#\n"));

    /* If second token is NULL or empty then the line must be a section
     * heading. */
    if (token == NULL || strlen(token) == 0) {
      strcpy(tmpSectionName, tmpStr);
      strcat(tmpSectionName, PARSER_VALUE_STRING);

      /* Check for duplicate section name. */
      find_duplicate_section(params, tmpSectionName);

      /* Check for duplicate standalone parameter name used as a section name.
       */
      find_duplicate_params(params, tmpStr);

      strcpy(section, tmpSectionName);
      strcpy(params->section[params->sectionCount].name, tmpSectionName);
      if (params->sectionCount == PARSER_MAX_NO_OF_SECTIONS - 1) {
        error(
            "Maximal number of sections in parameter file reached. Aborting !");
      } else {
        params->sectionCount++;
      }
      inSection = 1;
      isFirstParam = 1;
    } else {
      /* Create string with standalone parameter name appended with ":" to aid
       * duplicate search as section names are stored with ":" at the end.*/
      strcpy(tmpSectionName, tmpStr);
      strcat(tmpSectionName, PARSER_VALUE_STRING);

      /* Check for duplicate parameter name. */
      find_duplicate_params(params, tmpStr);

      /* Check for duplicate section name used as standalone parameter name. */
      find_duplicate_section(params, tmpSectionName);

      /* Must be a standalone parameter so no need to prefix name with a
       * section. */
      strcpy(params->data[params->paramCount].name, tmpStr);
      strcpy(params->data[params->paramCount].value, token);
      if (params->paramCount == PARSER_MAX_NO_OF_PARAMS - 1) {
        error(
            "Maximal number of parameters in parameter file reached. Aborting "
            "!");
      } else {
        params->paramCount++;
      }
      inSection = 0;
      isFirstParam = 1;
    }
  }
}

/**
 * @brief Parses a parameter that appears in a section and stores it in a
 *structure.
 *
 * @param line Line containing the parameter
 * @param isFirstParam Shows if the first parameter of a section has been found
 * @param sectionName String containing the current section name
 * @param params Structure to be written to
 *
 */

static void parse_section_param(char *line, int *isFirstParam,
                                char *sectionName, struct csds_parser *params) {
  static int sectionIndent = 0;
  char tmpStr[PARSER_MAX_LINE_SIZE];
  char paramName[PARSER_MAX_LINE_SIZE];
  char *token;

  /* Count indentation of each parameter and check that it
   * is consistent with the first parameter in the section. */
  if (*isFirstParam) {
    sectionIndent = count_indentation(line);
    *isFirstParam = 0;
  } else if (count_indentation(line) != sectionIndent) {
    error("Invalid line:%d '%s', parameter has incorrect indentation.",
          lineNumber, line);
  }

  /* Take first token as the parameter name and trim leading white space. */
  token = trim_both(strtok(line, ":\t"));
  strcpy(tmpStr, token);

  /* Take second token as the parameter value. */
  token = trim_both(strtok(NULL, "#\n"));

  /* Prefix the parameter name with its section name and
   * copy it into the parameter structure. */
  strcpy(paramName, sectionName);
  strcat(paramName, tmpStr);

  /* Check for duplicate parameter name. */
  find_duplicate_params(params, paramName);

  strcpy(params->data[params->paramCount].name, paramName);
  strcpy(params->data[params->paramCount].value, token);
  if (params->paramCount == PARSER_MAX_NO_OF_PARAMS - 1) {
    error("Maximal number of parameters in parameter file reached. Aborting !");
  } else {
    params->paramCount++;
  }
}

// Retrieve parameter value from structure. TYPE is the data type, float, int
// etc. FMT the format required for that data type, i.e. %f, %d etc. and DESC
// a one word description of the type, "float", "int" etc.
#define PARSER_GET_VALUE(TYPE, FMT, DESC)                                     \
  static int get_param_##TYPE(struct csds_parser *params, const char *name,   \
                              TYPE *def, TYPE *result) {                      \
    char str[PARSER_MAX_LINE_SIZE];                                           \
    for (int i = 0; i < params->paramCount; i++) {                            \
      if (strcmp(name, params->data[i].name) == 0) {                          \
        /* Check that exactly one number is parsed, capture junk. */          \
        if (sscanf(params->data[i].value, " " FMT "%s ", result, str) != 1) { \
          error("Tried parsing " DESC                                         \
                " '%s' but found '%s' with "                                  \
                "illegal trailing characters '%s'.",                          \
                params->data[i].name, params->data[i].value, str);            \
        }                                                                     \
        return 1;                                                             \
      }                                                                       \
    }                                                                         \
    if (def == NULL)                                                          \
      error("Cannot find '%s' in the structure, in file '%s'.", name,         \
            params->fileName);                                                \
    return 0;                                                                 \
  }

/* Instantiations. */
PARSER_GET_VALUE(char, "%c", "char");
PARSER_GET_VALUE(int, "%d", "int");
PARSER_GET_VALUE(float, "%f", "float");
PARSER_GET_VALUE(double, "%lf", "double");
PARSER_GET_VALUE(longlong, "%lld", "long long");

/**
 * @brief Retrieve integer parameter from structure.
 *
 * @param params Structure that holds the parameters
 * @param name Name of the parameter to be found
 * @return Value of the parameter found
 */
int parser_get_param_int(struct csds_parser *params, const char *name) {
  int result = 0;
  get_param_int(params, name, NULL, &result);
  return result;
}

/**
 * @brief Retrieve char parameter from structure.
 *
 * @param params Structure that holds the parameters
 * @param name Name of the parameter to be found
 * @return Value of the parameter found
 */
char parser_get_param_char(struct csds_parser *params, const char *name) {
  char result = 0;
  get_param_char(params, name, NULL, &result);
  return result;
}

/**
 * @brief Retrieve float parameter from structure.
 *
 * @param params Structure that holds the parameters
 * @param name Name of the parameter to be found
 * @return Value of the parameter found
 */
float parser_get_param_float(struct csds_parser *params, const char *name) {
  float result = 0;
  get_param_float(params, name, NULL, &result);
  return result;
}

/**
 * @brief Retrieve double parameter from structure.
 *
 * @param params Structure that holds the parameters
 * @param name Name of the parameter to be found
 * @return Value of the parameter found
 */
double parser_get_param_double(struct csds_parser *params, const char *name) {
  double result = 0;
  get_param_double(params, name, NULL, &result);
  return result;
}

/**
 * @brief Retrieve long long parameter from structure.
 *
 * @param params Structure that holds the parameters
 * @param name Name of the parameter to be found
 * @return Value of the parameter found
 */
long long parser_get_param_longlong(struct csds_parser *params,
                                    const char *name) {
  long long result = 0;
  get_param_longlong(params, name, NULL, &result);
  return result;
}

/**
 * @brief Retrieve string parameter from structure.
 *
 * @param params Structure that holds the parameters
 * @param name Name of the parameter to be found
 * @param retParam (return) Value of the parameter found
 */
void parser_get_param_string(struct csds_parser *params, const char *name,
                             char *retParam) {

  for (int i = 0; i < params->paramCount; i++) {
    if (!strcmp(name, params->data[i].name)) {
      strcpy(retParam, params->data[i].value);
      return;
    }
  }

  error("Cannot find '%s' in the structure.", name);
}

/**
 * @brief Retrieve optional integer parameter from structure.
 *
 * @param params Structure that holds the parameters
 * @param name Name of the parameter to be found
 * @param def Default value of the parameter of not found.
 * @return Value of the parameter found
 */
int parser_get_opt_param_int(struct csds_parser *params, const char *name,
                             int def) {
  int result = 0;
  if (get_param_int(params, name, &def, &result)) return result;
  return def;
}

/**
 * @brief Retrieve optional char parameter from structure.
 *
 * @param params Structure that holds the parameters
 * @param name Name of the parameter to be found
 * @param def Default value of the parameter of not found.
 * @return Value of the parameter found
 */
char parser_get_opt_param_char(struct csds_parser *params, const char *name,
                               char def) {
  char result = 0;
  if (get_param_char(params, name, &def, &result)) return result;
  return def;
}

/**
 * @brief Retrieve optional float parameter from structure.
 *
 * @param params Structure that holds the parameters
 * @param name Name of the parameter to be found
 * @param def Default value of the parameter of not found.
 * @return Value of the parameter found
 */
float parser_get_opt_param_float(struct csds_parser *params, const char *name,
                                 float def) {
  float result = 0;
  if (get_param_float(params, name, &def, &result)) return result;
  return def;
}

/**
 * @brief Retrieve optional double parameter from structure.
 *
 * @param params Structure that holds the parameters
 * @param name Name of the parameter to be found
 * @param def Default value of the parameter of not found.
 * @return Value of the parameter found
 */
double parser_get_opt_param_double(struct csds_parser *params, const char *name,
                                   double def) {
  double result = 0;
  if (get_param_double(params, name, &def, &result)) return result;
  return def;
}

/**
 * @brief Retrieve optional long long parameter from structure.
 *
 * @param params Structure that holds the parameters
 * @param name Name of the parameter to be found
 * @param def Default value of the parameter of not found.
 * @return Value of the parameter found
 */
long long parser_get_opt_param_longlong(struct csds_parser *params,
                                        const char *name, long long def) {
  long long result = 0;
  if (get_param_longlong(params, name, &def, &result)) return result;
  return def;
}

/**
 * @brief Retrieve string parameter from structure.
 *
 * @param params Structure that holds the parameters
 * @param name Name of the parameter to be found
 * @param def Default value of the parameter of not found.
 * @param retParam (return) Value of the parameter found
 */
void parser_get_opt_param_string(struct csds_parser *params, const char *name,
                                 char *retParam, const char *def) {

  for (int i = 0; i < params->paramCount; i++) {
    if (!strcmp(name, params->data[i].name)) {
      strcpy(retParam, params->data[i].value);
      return;
    }
  }
  strcpy(retParam, def);
}

/* Macro defining functions that get primitive types as simple one-line YAML
 * arrays, that is SEC: [v1,v2,v3...] format, with the extension that the []
 * are optional. TYPE is the data type, float etc. FMT a format to parse a
 * single value, so "%f" for a float and DESC the type description
 * i.e. "float".
 */
#define PARSER_GET_ARRAY(TYPE, FMT, DESC)                             \
  static int get_param_##TYPE##_array(struct csds_parser *params,     \
                                      const char *name, int required, \
                                      int nval, TYPE *values) {       \
    char str[PARSER_MAX_LINE_SIZE];                                   \
    char cpy[PARSER_MAX_LINE_SIZE];                                   \
                                                                      \
    for (int i = 0; i < params->paramCount; i++) {                    \
      if (!strcmp(name, params->data[i].name)) {                      \
        char *cp = cpy;                                               \
        strcpy(cp, params->data[i].value);                            \
        cp = trim_both(cp);                                           \
                                                                      \
        /* Strip off [], if present. */                               \
        if (cp[0] == '[') cp++;                                       \
        int l = strlen(cp);                                           \
        if (cp[l - 1] == ']') cp[l - 1] = '\0';                       \
        cp = trim_both(cp);                                           \
                                                                      \
        /* Format that captures spaces and trailing junk. */          \
        char fmt[20];                                                 \
        sprintf(fmt, " %s%%s ", FMT);                                 \
                                                                      \
        /* Parse out values which should now be "v, v, v" with        \
         * internal     whitespace variations. */                     \
        char *p = strtok(cp, ",");                                    \
        for (int k = 0; k < nval; k++) {                              \
          if (p != NULL) {                                            \
            TYPE tmp_value;                                           \
            if (sscanf(p, fmt, &tmp_value, str) != 1) {               \
              error("Tried parsing " DESC                             \
                    " '%s' but found '%s' with "                      \
                    "illegal " DESC " characters '%s'.",              \
                    name, p, str);                                    \
            }                                                         \
            values[k] = tmp_value;                                    \
          } else {                                                    \
            error(                                                    \
                "Array '%s' with value '%s' has too few values, "     \
                "expected %d",                                        \
                name, params->data[i].value, nval);                   \
          }                                                           \
          if (k < nval - 1) p = strtok(NULL, ",");                    \
        }                                                             \
        return 1;                                                     \
      }                                                               \
    }                                                                 \
    if (required)                                                     \
      error("Cannot find '%s' in the structure, in file '%s'.", name, \
            params->fileName);                                        \
    return 0;                                                         \
  }

/* Instantiations. */
PARSER_GET_ARRAY(char, "%c", "char");
PARSER_GET_ARRAY(int, "%d", "int");
PARSER_GET_ARRAY(float, "%f", "float");
PARSER_GET_ARRAY(double, "%lf", "double");
PARSER_GET_ARRAY(longlong, "%lld", "long long");

/**
 * @brief Retrieve char array parameter from structure.
 *
 * @param params Structure that holds the parameters
 * @param name Name of the parameter to be found
 * @param nval number of values expected.
 * @param values Values of the parameter found, of size at least nvals.
 */
void parser_get_param_char_array(struct csds_parser *params, const char *name,
                                 int nval, char *values) {
  get_param_char_array(params, name, 1, nval, values);
}

/**
 * @brief Retrieve optional char array parameter from structure.
 *
 * @param params Structure that holds the parameters
 * @param name Name of the parameter to be found
 * @param nval number of values expected.
 * @param values Values of the parameter found, of size at least nvals. If the
 *               parameter is not found these values will be returned
 *               unmodified, so should be set to the default values.
 * @return whether the parameter has been found.
 */
int parser_get_opt_param_char_array(struct csds_parser *params,
                                    const char *name, int nval, char *values) {
  if (get_param_char_array(params, name, 0, nval, values) != 1) {
    return 0;
  }
  return 1;
}

/**
 * @brief Retrieve int array parameter from structure.
 *
 * @param params Structure that holds the parameters
 * @param name Name of the parameter to be found
 * @param nval number of values expected.
 * @param values Values of the parameter found, of size at least nvals.
 */
void parser_get_param_int_array(struct csds_parser *params, const char *name,
                                int nval, int *values) {
  get_param_int_array(params, name, 1, nval, values);
}

/**
 * @brief Retrieve optional int array parameter from structure.
 *
 * @param params Structure that holds the parameters
 * @param name Name of the parameter to be found
 * @param nval number of values expected.
 * @param values Values of the parameter found, of size at least nvals. If the
 *               parameter is not found these values will be returned
 *               unmodified, so should be set to the default values.
 * @return whether the parameter has been found.
 */
int parser_get_opt_param_int_array(struct csds_parser *params, const char *name,
                                   int nval, int *values) {
  if (get_param_int_array(params, name, 0, nval, values) != 1) {
    return 0;
  }
  return 1;
}

/**
 * @brief Retrieve float array parameter from structure.
 *
 * @param params Structure that holds the parameters
 * @param name Name of the parameter to be found
 * @param nval number of values expected.
 * @param values Values of the parameter found, of size at least nvals.
 */
void parser_get_param_float_array(struct csds_parser *params, const char *name,
                                  int nval, float *values) {
  get_param_float_array(params, name, 1, nval, values);
}

/**
 * @brief Retrieve optional float array parameter from structure.
 *
 * @param params Structure that holds the parameters
 * @param name Name of the parameter to be found
 * @param nval number of values expected.
 * @param values Values of the parameter found, of size at least nvals. If the
 *               parameter is not found these values will be returned
 *               unmodified, so should be set to the default values.
 * @return whether the parameter has been found.
 */
int parser_get_opt_param_float_array(struct csds_parser *params,
                                     const char *name, int nval,
                                     float *values) {
  if (get_param_float_array(params, name, 0, nval, values) != 1) {
    return 0;
  }
  return 1;
}

/**
 * @brief Retrieve double array parameter from structure.
 *
 * @param params Structure that holds the parameters
 * @param name Name of the parameter to be found
 * @param nval number of values expected.
 * @param values Values of the parameter found, of size at least nvals.
 */
void parser_get_param_double_array(struct csds_parser *params, const char *name,
                                   int nval, double *values) {
  get_param_double_array(params, name, 1, nval, values);
}

/**
 * @brief Retrieve optional double array parameter from structure.
 *
 * @param params Structure that holds the parameters
 * @param name Name of the parameter to be found
 * @param nval number of values expected.
 * @param values Values of the parameter found, of size at least nvals. If the
 *               parameter is not found these values will be returned
 *               unmodified, so should be set to the default values.
 * @return whether the parameter has been found.
 */
int parser_get_opt_param_double_array(struct csds_parser *params,
                                      const char *name, int nval,
                                      double *values) {
  if (get_param_double_array(params, name, 0, nval, values) != 1) {
    return 0;
  }
  return 1;
}

/**
 * @brief Retrieve long long array parameter from structure.
 *
 * @param params Structure that holds the parameters
 * @param name Name of the parameter to be found
 * @param nval number of values expected.
 * @param values Values of the parameter found, of size at least nvals.
 */
void parser_get_param_longlong_array(struct csds_parser *params,
                                     const char *name, int nval,
                                     long long *values) {
  get_param_longlong_array(params, name, 1, nval, values);
}

/**
 * @brief Retrieve optional long long array parameter from structure.
 *
 * @param params Structure that holds the parameters
 * @param name Name of the parameter to be found
 * @param nval number of values expected.
 * @param values Values of the parameter found, of size at least nvals. If the
 *               parameter is not found these values will be returned
 *               unmodified, so should be set to the default values.
 * @return whether the parameter has been found.
 */
int parser_get_opt_param_longlong_array(struct csds_parser *params,
                                        const char *name, int nval,
                                        long long *values) {
  if (get_param_longlong_array(params, name, 0, nval, values) != 1) {
    return 0;
  }
  return 1;
}

/**
 * @brief Prints the contents of the parameter structure.
 *
 * @param params Structure that holds the parameters
 */
void parser_print_params(const struct csds_parser *params) {
  printf("\n--------------------------\n");
  printf("|   CSDS Parameter File  |\n");
  printf("--------------------------\n");

  for (int i = 0; i < params->paramCount; i++) {
    printf("Parameter name: %s\n", params->data[i].name);
    printf("Parameter value: %s\n", params->data[i].value);
  }
}
