/*******************************************************************************
 * This file is part of CSDS and was copied from SWIFT.
 * Copyright (c) 2021 loic.hausammann@epfl.ch
 *               SWIFT
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 ******************************************************************************/
#ifndef CSDS_PARSER_H
#define CSDS_PARSER_H

/* Config parameters. */
#include "../config.h"

/* Standard headers */
#include <stdio.h>

#if defined(HAVE_HDF5)
#include <hdf5.h>
#endif

/* Some constants. */
#define PARSER_MAX_LINE_SIZE 256
#define PARSER_MAX_NO_OF_PARAMS 600
#define PARSER_MAX_NO_OF_SECTIONS 64

/* A parameter in the input file */
struct parameter {
  char name[PARSER_MAX_LINE_SIZE];
  char value[PARSER_MAX_LINE_SIZE];
};

struct section {
  char name[PARSER_MAX_LINE_SIZE];
};

/* The array of parameters read from a file */
struct csds_parser {
  struct section section[PARSER_MAX_NO_OF_SECTIONS];
  struct parameter data[PARSER_MAX_NO_OF_PARAMS];
  int sectionCount;
  int paramCount;
  char fileName[PARSER_MAX_LINE_SIZE];
};

/* Public API. */
void parser_init(const char *file_name, struct csds_parser *params);
void parser_read_file(const char *file_name, struct csds_parser *params);
void parser_print_params(const struct csds_parser *params);

char parser_get_param_char(struct csds_parser *params, const char *name);
int parser_get_param_int(struct csds_parser *params, const char *name);
float parser_get_param_float(struct csds_parser *params, const char *name);
double parser_get_param_double(struct csds_parser *params, const char *name);
long long parser_get_param_longlong(struct csds_parser *params,
                                    const char *name);
void parser_get_param_string(struct csds_parser *params, const char *name,
                             char *retParam);

char parser_get_opt_param_char(struct csds_parser *params, const char *name,
                               char def);
int parser_get_opt_param_int(struct csds_parser *params, const char *name,
                             int def);
float parser_get_opt_param_float(struct csds_parser *params, const char *name,
                                 float def);
double parser_get_opt_param_double(struct csds_parser *params, const char *name,
                                   double def);
long long parser_get_opt_param_longlong(struct csds_parser *params,
                                        const char *name, long long def);
void parser_get_opt_param_string(struct csds_parser *params, const char *name,
                                 char *retParam, const char *def);
void parser_get_param_char_array(struct csds_parser *params, const char *name,
                                 int nval, char *values);
void parser_get_param_int_array(struct csds_parser *params, const char *name,
                                int nval, int *values);
void parser_get_param_float_array(struct csds_parser *params, const char *name,
                                  int nval, float *values);
void parser_get_param_double_array(struct csds_parser *params, const char *name,
                                   int nval, double *values);
void parser_get_param_longlong_array(struct csds_parser *params,
                                     const char *name, int nval,
                                     long long *values);
int parser_get_opt_param_char_array(struct csds_parser *params,
                                    const char *name, int nval, char *values);
int parser_get_opt_param_int_array(struct csds_parser *params, const char *name,
                                   int nval, int *values);
int parser_get_opt_param_float_array(struct csds_parser *params,
                                     const char *name, int nval, float *values);
int parser_get_opt_param_double_array(struct csds_parser *params,
                                      const char *name, int nval,
                                      double *values);
int parser_get_opt_param_longlong_array(struct csds_parser *params,
                                        const char *name, int nval,
                                        long long *values);

#endif /* CSDS_PARSER_H */
