/*******************************************************************************
 * This file is part of CSDS.
 * Copyright (c) 2019 Loic Hausammann (loic.hausammann@epfl.ch)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 ******************************************************************************/
#ifndef CSDS_CSDS_PARTICLE_H
#define CSDS_CSDS_PARTICLE_H

/* Include the tools. */
#include "csds_tools.h"

/* Include the other local files. */
#include "csds_definitions.h"
#include "csds_header.h"
#include "csds_interpolation.h"
#include "csds_mapped_file.h"
#include "csds_parameters.h"
#include "csds_time.h"

#include <stdio.h>
#include <stdlib.h>

/**
 * @brief Read a particle (of any type) record in the log file.
 *
 * @param offset offset of the record to read.
 * @param output Buffer for the requested field.
 * @param field_read The field to read.
 * @param derivative The derivative to read (0 for the field).
 * @param mask (out) The mask of the record.
 * @param h_offset (out) Difference of offset with the next record.
 * @param fields The list of available fields
 * @param number_fields The number of fields
 * @param log_map The mmap logfile.
 *
 * @return position after the record.
 */
INLINE static size_t csds_particle_read_field(
    size_t offset, void *output, const struct field_information *field_read,
    const int derivative, mask_type *mask, size_t *h_offset,
    const struct field_information *fields, int number_fields,
    struct mapped_file *log_map) {

  *mask = 0;
  *h_offset = 0;

  /* Read the record's mask. */
  offset = csds_mapped_file_read_mask(log_map, offset, mask, h_offset);

#ifdef CSDS_DEBUG_CHECKS
  /* Check if it is not a time record. */
  if (*mask == CSDS_TIMESTAMP_MASK) {
    error_python("Cannot read a particle from timestep record.");
  }

  if (derivative == 1 && field_read->first.field == field_enum_none) {
    error("Trying to read the non existing first derivative.");
  }
  if (derivative == 2 && field_read->second.field == field_enum_none) {
    error("Trying to read the non existing second derivative.");
  }
#endif

  /* Get the position of the field to read */
  const int position = derivative == 0
                           ? field_read->field.position
                           : derivative == 1 ? field_read->first.position
                                             : field_read->second.position;

#ifdef CSDS_DEBUG_CHECKS
  if (position < 0) {
    error("Position not set (derivative %i of %s)", derivative,
          field_enum_get_name(field_read->field.field));
  }
#endif

  /* Read the record and copy it to a particle. */
  for (int i = 0; i < number_fields; i++) {

    /* Is the mask present? */
    if (!(*mask & fields[i].field.mask)) {
      continue;
    }

    if (position == i) {
      /* Read the data. */
      offset = csds_mapped_file_read_data(log_map, offset, fields[i].field.size,
                                          output);
    } else {
      /* Update the buffer's position. */
      offset += fields[i].field.size;
    }
  }

  return offset;
}

/**
 * @brief Extract the flag and its related data from the data inside a record.
 *
 * Flags and data are stored as a 32-bit unsigned int in which the lower 24 bits
 * contain the data, and the top 8 bits contain the flag value.

 * @param logfile_data The raw data taken from the logfile.
 * @param data (output) The data related to the flag.
 * @param part_type (output) The particle type raising this flag.
 *
 * @return The flag extracted from the raw data.
 */

INLINE static enum csds_special_flags csds_unpack_flags_and_data(
    uint32_t logfile_data, int *flag_data, int *part_type) {
  *part_type = (logfile_data & 0xFF);
  *flag_data = (logfile_data & 0xFFFF00) >> 8;

  const int tmp = ((logfile_data & 0xFF000000) >> 24);
  const enum csds_special_flags flag = (enum csds_special_flags)tmp;

#ifdef CSDS_DEBUG_CHECKS
  if (*part_type >= csds_type_count || *part_type < 0) {
    error("Invalid particle type found (%i)", *part_type);
  }
#endif
  return flag;
}

/**
 * @brief Read the special flag of a particle (of any type) in the log file.
 *
 * @param offset offset of the record to read.
 * @param mask (out) The mask of the record.
 * @param h_offset (out) Difference of offset with the next record.
 * @param data (out) The data of the flag.
 * @param part_type (out) The type of particle raising this flag.
 *
 * @return The special flag.
 */
INLINE static enum csds_special_flags csds_particle_read_special_flag(
    size_t offset, mask_type *mask, size_t *h_offset, int *data, int *part_type,
    struct mapped_file *map) {

  *mask = 0;
  *h_offset = 0;

  /* Read the record's mask. */
  offset = csds_mapped_file_read_mask(map, offset, mask, h_offset);

#ifdef CSDS_DEBUG_CHECKS
  /* Check if it is not a time record. */
  if (*mask == CSDS_TIMESTAMP_MASK) {
    error_python("Cannot read a particle from timestep record.");
  }
#endif

  /* Read the special flag */
  uint32_t packed_data = 0;
  offset = csds_mapped_file_read_data(map, offset, CSDS_SPECIAL_FLAGS_SIZE,
                                      &packed_data);

  return csds_unpack_flags_and_data(packed_data, data, part_type);
}

/**
 * @brief Interpolate a field of the particle at the given time.
 *
 * @param before Pointer to the #csds_reader_field at a time < t.
 * @param after Pointer to the #csds_reader_field at a time > t.
 * @param otuput Pointer to the output value.
 * @param time_before Time of field_before (< t).
 * @param time_after Time of field_after (> t).
 * @param time Requested time.
 * @param field The field to reconstruct.
 * @param params The simulation's #csds_parameters.
 */
INLINE static void csds_particle_interpolate_field(
    const double time_before, const struct csds_reader_field *restrict before,
    const double time_after, const struct csds_reader_field *restrict after,
    void *restrict output, const double time,
    const struct field_information *field,
    const struct csds_parameters *params) {

  switch (field->field.field) {
      /* Coordinates */
    case field_enum_coordinates:
      interpolate_quintic_double_float_ND(
          time_before, before, time_after, after, output, time,
          /* dimension */ 3, params->periodic, params);
      break;

      /* Velocities */
    case field_enum_velocities:
      interpolate_cubic_float_ND(time_before, before, time_after, after, output,
                                 time,
                                 /* dimension= */ 3, /* periodic= */ 0, params);
      break;

      /* Accelerations */
    case field_enum_accelerations:
      interpolate_linear_float_ND(
          time_before, before, time_after, after, output, time,
          /* dimension= */ 3, /* periodic= */ 0, params);
      break;

      /* Interpolation of float */
    case field_enum_masses:
    case field_enum_smoothing_lengths:
    case field_enum_internal_energies:
    case field_enum_entropies:
    case field_enum_densities:
      interpolate_linear_float(time_before, before, time_after, after, output,
                               time, /* periodic= */ 0, params);
      break;

      /* Constant float term */
    case field_enum_birth_time:
      if (*(float *)after->field != *(float *)before->field) {
        error("Interpolating different particles.");
      }
      *(float *)output = *(float *)after->field;
      break;

      /* Particle IDs */
    case field_enum_particles_ids:
      /* Ensure to have the same particle */
      if (*(long long *)after->field != *(long long *)before->field) {
        error("Interpolating different particles.");
      }
      *(long long *)output = *(long long *)after->field;
      break;

      /* Gear star formation */
    case field_enum_gear_star_formation: {
      /* Get some variables */
      const float wa = (time - time_before) / (time_after - time_before);
      const float wb = 1. - wa;
      float *x = (float *)output;
      const float *bef = (float *)before->field;
      const float *aft = (float *)after->field;

      /* Birth densities + Birth masses */
      x[0] = wa * aft[0] + wb * bef[0];
      x[1] = wa * aft[1] + wb * bef[1];

      /* ProgenitorIDs */
      long long *id = (long long *)(x + 2);
      *id = *(long long *)(aft + 2);
      if (*id != *(long long *)(bef + 2)) {
        error("Interpolating different particles.");
      }
      break;
    }

      /* GEAR chemistry (part) */
    case field_enum_gear_chemistry_part:
      interpolate_linear_double_ND(time_before, before, time_after, after,
                                   output, time,
                                   2 * GEAR_CHEMISTRY_ELEMENT_COUNT,
                                   /* periodic */ 0, params);
      break;

      /* GEAR chemistry (spart) */
    case field_enum_gear_chemistry_spart:
      interpolate_linear_double_ND(time_before, before, time_after, after,
                                   output, time, GEAR_CHEMISTRY_ELEMENT_COUNT,
                                   /* periodic */ 0, params);
      break;

      /* Sphenix grouped fields */
    case field_enum_sphenix_secondary_fields: {
      /* Get some variables */
      const float wa = (time - time_before) / (time_after - time_before);
      const float wb = 1. - wa;
      float *x = (float *)output;
      const float *bef = (float *)before->field;
      const float *aft = (float *)after->field;

      /* Entropy + pressure + Viscosity + Diffusion + Laplacian*/
      const int n_linear = 5;
      for (int i = 0; i < n_linear; i++) {
        x[i] = wa * aft[i] + wb * bef[i];
      }

      /* Div v */
      x[n_linear] = interpolate_cubic_hermite_float(
          time_before, bef[n_linear], bef[n_linear + 1], time_after,
          aft[n_linear], aft[n_linear + 1], time);

      /* d Div v / dt */
      x[n_linear + 1] = wa * aft[n_linear + 1] + wb * bef[n_linear + 1];
      /* Use the linear interpolation */
      x[1] = wa * aft[n_linear + 2] + wb * bef[n_linear + 2];
      break;
    }

    default:
      error("Field %s not implemented",
            field_enum_get_name(field->field.field));
  }
}

#endif  // CSDS_CSDS_PARTICLE_H
