/*******************************************************************************
 * This file is part of CSDS.
 * Copyright (c) 2019 Loic Hausammann (loic.hausammann@epfl.ch)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 ******************************************************************************/
#include "csds_fields.h"
#include "csds_header.h"
#include "csds_mapped_file.h"
#include "csds_particle.h"
#include "csds_python_tools.h"
#include "csds_reader.h"
#include "csds_time.h"

#ifdef HAVE_PYTHON
#define NPY_NO_DEPRECATED_API NPY_1_7_API_VERSION

#include <Python.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>

/* Include PyMethodDef */
#include <structmember.h>

static PyTypeObject PyParticleReader_Type;

/**
 * @brief Reader for a single particle type.
 */
typedef struct {
  PyObject_HEAD;

  /* Link to the reader */
  void *reader;

  /* Particle type */
  enum part_type type;
} PyParticleReader;

/**
 * @brief Main class for the reader.
 */
typedef struct {
  PyObject_HEAD;

  /* Add csds stuff here */
  struct csds_reader reader;

  /* Is the csds ready? */
  int ready;

  /* basename of the logfile */
  char *basename;

  /* Verbose level */
  int verbose;

  /* Number of index file to write */
  int number_index;

  /* Are we restarting the index files? */
  int restart_init;

  /* Are we using the cache? */
  int use_cache;

  /* Reader for each type of particles */
  PyParticleReader *part_reader[csds_type_count];

} PyObjectReader;

/**
 * @brief Deallocate the memory of a particle reader.
 */
INLINE static void particle_reader_dealloc(PyParticleReader *self) {
  Py_TYPE(self)->tp_free((PyObject *)self);
}

/**
 * @brief Deallocate the memory.
 */
static void Reader_dealloc(PyObjectReader *self) {
  if (self->basename != NULL) {
    free(self->basename);
    self->basename = NULL;
  }
  Py_TYPE(self)->tp_free((PyObject *)self);
}

/**
 * @brief Allocate the memory of a particle reader.
 */
INLINE static PyObject *particle_reader_new(PyTypeObject *type,
                                            ATTR_UNUSED PyObject *args,
                                            ATTR_UNUSED PyObject *kwds) {
  PyParticleReader *self = (PyParticleReader *)type->tp_alloc(type, 0);
  return (PyObject *)self;
}

/**
 * @brief Allocate the memory.
 */
static PyObject *Reader_new(PyTypeObject *type, ATTR_UNUSED PyObject *args,
                            ATTR_UNUSED PyObject *kwds) {
  PyObjectReader *self;

  self = (PyObjectReader *)type->tp_alloc(type, 0);
  self->ready = 0;
  self->basename = NULL;
  self->use_cache = 0;

  for (int i = 0; i < csds_type_count; i++) {
    self->part_reader[i] = NULL;
  }

  return (PyObject *)self;
}
/**
 * @brief Initializer of the Reader.
 *
 * @param basename The basename of the csds file.
 * @param verbose The verbosity level.
 *
 * @return The Reader object.
 */
static int Reader_init(PyObjectReader *self, PyObject *args, PyObject *kwds) {

  /* input variables. */
  char *basename = NULL;
  int verbose = 0;
  int number_index = 10;
  int restart_init = 0;
  int use_cache = 0;

  /* List of keyword arguments. */
  static char *kwlist[] = {"basename",     "verbose",   "number_index",
                           "restart_init", "use_cache", NULL};

  /* parse the arguments. */
  if (!PyArg_ParseTupleAndKeywords(args, kwds, "s|iipp", kwlist, &basename,
                                   &verbose, &number_index, &restart_init,
                                   &use_cache))
    return -1;

  /* Copy the arguments */
  self->verbose = verbose;
  self->number_index = number_index;
  self->restart_init = restart_init;
  self->use_cache = use_cache;

  size_t n_plus_null = strlen(basename) + 1;
  self->basename = (char *)malloc(n_plus_null * sizeof(char));
  if (self->basename == NULL) {
    error_python("Failed to allocate memory");
  }
  strcpy(self->basename, basename);

  return 0;
}

/**
 * @brief Read the minimal and maximal time.
 *
 * <b>returns</b> tuple containing min and max time.
 */
static PyObject *getTimeLimits(PyObject *self, ATTR_UNUSED PyObject *args) {
  if (!((PyObjectReader *)self)->ready) {
    error_python(
        "The CSDS is not ready yet."
        "Did you forget to open it with \"with\"?");
  }

  /* initialize the reader. */
  struct csds_reader *reader = &((PyObjectReader *)self)->reader;

  if (reader->verbose > 1) message("Reading time limits.");

  /* Get the time limits */
  double time_min = csds_reader_get_time_begin(reader);
  double time_max = csds_reader_get_time_end(reader);

  /* Create the output */
  PyObject *out = PyTuple_New(2);
  PyTuple_SetItem(out, 0, PyFloat_FromDouble(time_min));
  PyTuple_SetItem(out, 1, PyFloat_FromDouble(time_max));

  return (PyObject *)out;
}

/**
 * @brief Read the box size.
 *
 * <b>returns</b> tuple containing the box size.
 */
static PyObject *get_box_size(PyObject *self, ATTR_UNUSED PyObject *args) {
  if (!((PyObjectReader *)self)->ready) {
    error_python(
        "The CSDS is not ready yet."
        "Did you forget to open it with \"with\"?");
  }

  /* initialize the reader. */
  struct csds_reader *reader = &((PyObjectReader *)self)->reader;

  /* Create the output */
  PyObject *out = PyTuple_New(reader->params.dimension);
  for (int i = 0; i < reader->params.dimension; i++) {
    PyTuple_SetItem(out, i, PyFloat_FromDouble(reader->params.box_size[i]));
  }

  return (PyObject *)out;
}

/**
 * @brief Create a list of numpy array containing the fields.
 *
 * @param output A list of array of fields.
 * @param field_indices The indices (header ordering) of the requested fields.
 * @param n_fields The number of fields requested.
 * @param n_part The number of particles of each type.
 * @param n_tot The total number of particles.
 *
 * @return The python list of numpy array.
 */
INLINE static PyObject *csds_loader_create_output(void **output,
                                                  const enum field_enum *fields,
                                                  const int n_fields,
                                                  uint64_t n_tot) {

  /* Create the python dictionary */
  PyObject *dict = PyDict_New();

  /* Get all the requested fields */
  for (int i = 0; i < n_fields; i++) {
    /* Get all the required informations */
    struct csds_python_field current_field =
        field_enum_get_python_field(fields[i]);

    PyObject *array = NULL;
    /* Simple case where we only have normal type */
    if (current_field.typenum != CUSTOM_NPY_TYPE) {
      if (current_field.dimension > 1) {
        npy_intp dims[2] = {n_tot, current_field.dimension};
        array = PyArray_SimpleNewFromData(2, dims, current_field.typenum,
                                          output[i]);
      } else {
        npy_intp dims = n_tot;
        array = PyArray_SimpleNewFromData(1, &dims, current_field.typenum,
                                          output[i]);
      }
    }
    /* Now the complex types */
    else {
      /* Ensure that the field is properly defined */
      if (current_field.subfields_registred != current_field.dimension) {
        error_python(
            "It seems that you forgot to register a field (found %i and "
            "expecting %i)",
            current_field.subfields_registred, current_field.dimension);
      }

      /* Creates the dtype */
      PyObject *names = PyList_New(current_field.dimension);
      PyObject *formats = PyList_New(current_field.dimension);
      PyObject *offsets = PyList_New(current_field.dimension);

      /* Gather the data into the lists */
      for (int k = 0; k < current_field.dimension; k++) {
        struct csds_python_subfield *subfield = &current_field.subfields[k];

        /* Transform the information into python */
        PyObject *name = PyUnicode_FromString(subfield->name);
        PyObject *offset = PyLong_FromSize_t(subfield->offset);
        PyObject *format = csds_python_tools_get_format_string(
            subfield->typenum, subfield->dimension);

        /* Add everything into the lists */
        PyList_SetItem(names, k, name);
        PyList_SetItem(formats, k, format);
        PyList_SetItem(offsets, k, offset);
      }

      /* Now create the dtype dictionary */
      PyObject *dtype = PyDict_New();
      PyDict_SetItemString(dtype, "names", names);
      PyDict_SetItemString(dtype, "formats", formats);
      PyDict_SetItemString(dtype, "offsets", offsets);

      /* Cleanup a bit */
      Py_DECREF(names);
      Py_DECREF(formats);
      Py_DECREF(offsets);

      /* Now get the pyarray_descr */
      PyArray_Descr *descr = NULL;
      PyArray_DescrConverter(dtype, &descr);

      /* Create the list of flags */

      /* Create the array */
      npy_intp dims = n_tot;
      array = PyArray_NewFromDescr(&PyArray_Type, descr, /* nd */ 1, &dims,
                                   /* strides */ NULL, output[i],
                                   NPY_ARRAY_CARRAY, NULL);
    }

    csds_loader_python_field_free(&current_field);
    const char *field_name = field_enum_get_name(fields[i]);
    PyDict_SetItemString(dict, field_name, array);
  }

  return dict;
}

static PyObject *pyEnter(ATTR_UNUSED PyObject *self,
                         ATTR_UNUSED PyObject *args) {

  PyObjectReader *self_reader = (PyObjectReader *)self;
  csds_reader_init(&self_reader->reader, self_reader->basename,
                   self_reader->verbose, self_reader->number_index,
                   self_reader->restart_init, self_reader->use_cache);
  self_reader->ready = 1;

  /* Allocate the particle readers */
  PyObject *args_init = PyTuple_New(0);
  for (int i = 0; i < csds_type_count; i++) {
    PyParticleReader *part_reader = (PyParticleReader *)PyObject_CallObject(
        (PyObject *)&PyParticleReader_Type, args_init);

    /* Ensure that the object was created */
    if (part_reader == NULL) {
      return NULL;
    }
    part_reader->reader = (void *)self_reader;
    self_reader->part_reader[i] = part_reader;
    self_reader->part_reader[i]->type = i;
  }
  Py_DecRef(args_init);

  Py_INCREF(self);
  return self;
}

static PyObject *pyExit(ATTR_UNUSED PyObject *self,
                        ATTR_UNUSED PyObject *args) {
  PyObjectReader *self_reader = (PyObjectReader *)self;
  if (!self_reader->ready) {
    error_python("It seems that the reader was not initialized");
  }
  csds_reader_free(&self_reader->reader);
  self_reader->ready = 0;

  /* Do the same for the particle readers */
  for (int i = 0; i < csds_type_count; i++) {
    self_reader->part_reader[i]->reader = NULL;
    Py_DecRef((PyObject *)self_reader->part_reader[i]);
  }

  Py_RETURN_NONE;
}

static PyObject *pyGetListFields(ATTR_UNUSED PyObject *self, PyObject *args,
                                 PyObject *kwds) {
  PyObjectReader *self_reader = (PyObjectReader *)self;
  if (!self_reader->ready) {
    error_python(
        "The CSDS is not ready yet."
        "Did you forget to open it with \"with\"?");
  }

  /* input variables. */
  PyObject *types = Py_None;

  /* List of keyword arguments. */
  static char *kwlist[] = {"part_type", NULL};

  /* parse the arguments. */
  if (!PyArg_ParseTupleAndKeywords(args, kwds, "|O", kwlist, &types))
    return NULL;

  /* Get the type of particles to read. */
  int read_types[csds_type_count] = {0};
  /* By default, we read everything */
  if (types == Py_None) {
    for (int i = 0; i < csds_type_count; i++) {
      read_types[i] = 1;
    }
  }
  /* Deal with the case of a single int. */
  else if (PyLong_Check(types)) {
    const size_t type = PyLong_AsSize_t(types);
    if (type >= csds_type_count) {
      error_python("Unexpected particle type %zi", type);
    }
    read_types[type] = 1;
  }
  /* Deal with the case of a list */
  else if (PyList_Check(types)) {
    const size_t size = PyList_Size(types);
    for (size_t i = 0; i < size; i++) {
      PyObject *cur = PyList_GetItem(types, i);
      const size_t type = PyLong_AsSize_t(cur);
      if (type >= csds_type_count) {
        error_python("Unexpected particle type %zi", type);
      }
      read_types[type] = 1;
    }
  }

  /* initialize the reader. */
  struct csds_reader *reader = &self_reader->reader;
  const struct header *h = &reader->log.header;

  /* Create the array to check if a field is present. */
  int *field_present = (int *)malloc(field_enum_count * sizeof(int));
  if (field_present == NULL) {
    error("Failed to allocate the memory for the fields present.");
  }

  /* Find if the fields are present in all the types. */
  for (enum field_enum i = 0; i < field_enum_count; i++) {
    field_present[i] = 1;

    /* Check all the types */
    for (int type = 0; type < csds_type_count; type++) {
      /* Skip the types that are not required
         and the field not found */
      if (read_types[type] == 0 || field_present[i] == 0) continue;

      /* Search among all the fields of the particle type */
      int found = 0;
      for (int j = 0; j < h->number_fields[type]; j++) {
        if (i == h->fields[type][j].field.field) {
          found = 1;
          break;
        }
      }

      /* Set the field as not found */
      if (!found) field_present[i] = 0;
    }
  }

  /* Remove the special flags */
  field_present[field_enum_special_flags] = 0;

  /* Count the number of fields found */
  int number_fields = 0;
  for (int i = 0; i < field_enum_count; i++) {
    number_fields += field_present[i];
  }

  /* Create the python list for the output*/
  PyObject *list = PyList_New(number_fields);
  int current = 0;
  for (int i = 0; i < field_enum_count; i++) {
    /* Keep only the field present. */
    if (field_present[i] == 0) continue;

    const char *name = field_enum_get_name(i);
    PyObject *python_name = PyUnicode_FromString(name);
    PyList_SetItem(list, current, python_name);
    current += 1;
  }

  /* Free the memory. */
  free(field_present);

  return list;
}

/**
 * @brief shortcut to PyGetListFields for a single particle type
 */
static PyObject *pyParticleGetListFields(ATTR_UNUSED PyObject *self,
                                         PyObject *args, PyObject *kwds) {
  PyParticleReader *part_reader = (PyParticleReader *)self;
  PyObject *reader = (PyObject *)part_reader->reader;

  /* Check if we need to create the dictionary */
  const int dict_created = kwds == NULL;
  if (dict_created) {
    kwds = PyDict_New();
  }

  /* Set the particle type  */
  PyObject *part_type = PyLong_FromLong(part_reader->type);

  /* Add the list to the k-arguments */
  if (PyDict_SetItemString(kwds, "part_type", part_type) != 0) return NULL;

  /* Get the fields */
  PyObject *ret = pyGetListFields(reader, args, kwds);

  /* Cleanup */
  if (dict_created) {
    Py_DecRef(kwds);
    kwds = NULL;
  } else {
    PyDict_DelItemString(kwds, "part_type");
  }

  return ret;
}
/**
 * @brief Check the input data for #pyGetData.
 *
 * @param fields The list of fields to read.
 * @param types The type of particles to read.
 * @param part_ids The list of particle ids to read.
 * @param read_types (output) The list of particles' type to
 *   read (needs to be already allocated).
 */
void pyGetData_CheckInput(PyObject *fields, PyObject *types, PyObject *part_ids,
                          int *read_types) {

  /* Check the inputs. */
  if (!PyList_Check(fields)) {
    error_python("Expecting a list of fields");
  }

  /* Ensure that only the type or the ids are provided. */
  if (types != Py_None && part_ids != Py_None) {
    error_python(
        "Cannot request for both a type of particles and "
        "a list of ids.");
  }

  /* Ensure that a list of ids are provided. */
  if (part_ids != Py_None) {
    if (!PyList_Check(part_ids)) {
      error_python(
          "Expecting a list of numpy arrays containing the ids of the "
          "particles");
    }

    /* Check each element */
    const size_t size = PyList_Size(part_ids);
    if (size != csds_type_count) {
      error_python(
          "The list of ids should contain %i elements (one for each particle "
          "type)",
          csds_type_count);
    }
    for (size_t i = 0; i < size; i++) {
      PyArrayObject *el = (PyArrayObject *)PyList_GetItem(part_ids, i);

      /* Early abort */
      if ((PyObject *)el == Py_None) continue;

      /* Check object type */
      if (!PyArray_Check(el))
        error_python(
            "Expecting either None or a numpy array for each element in "
            "the particle ids");
      /* Check data type */
      if (PyArray_TYPE(el) != NPY_LONGLONG) {
        error_python("Expecting an array of long long integer for the ids.");
      }
    }
  }

  /* Get the type of particles to read. */
  /* By default, we read everything */
  if (types == Py_None) {
    for (int i = 0; i < csds_type_count; i++) {
      read_types[i] = 1;
    }
  }
  /* Deal with the case of a single int. */
  else if (PyLong_Check(types)) {
    const size_t type = PyLong_AsSize_t(types);
    if (type >= csds_type_count) {
      error_python("Unexpected particle type %zi", type);
    }
    read_types[type] = 1;
  }
  /* Deal with the case of a list */
  else if (PyList_Check(types)) {
    const size_t size = PyList_Size(types);
    for (size_t i = 0; i < size; i++) {
      PyObject *cur = PyList_GetItem(types, i);
      const size_t type = PyLong_AsSize_t(cur);
      if (type >= csds_type_count) {
        error_python("Unexpected particle type %zi", type);
      }
      read_types[type] = 1;
    }
  }
}

/**
 * @brief Read some fields at a given time.
 *
 * @param basename The basename of the csds files.
 * @param fields Python list containing the name of the fields (e.g.
 * Coordinates).
 * @param time The time of the fields.
 * @param verbose (Optional) The verbose level of the reader.
 *
 * @return Dictionary of numpy array containing the fields requested.
 */
static PyObject *pyGetData(ATTR_UNUSED PyObject *self, PyObject *args,
                           PyObject *kwds) {

  PyObjectReader *self_reader = (PyObjectReader *)self;
  if (!self_reader->ready) {
    error_python(
        "The CSDS is not ready yet."
        "Did you forget to open it with \"with\"?");
  }

  /* input variables. */
  PyObject *fields = NULL;
  double time = 0;
  PyObject *types = Py_None;
  PyObject *part_ids = Py_None;

  /* List of keyword arguments. */
  static char *kwlist[] = {"fields", "time", "filter_by_ids", "part_type",
                           NULL};

  /* parse the arguments. */
  if (!PyArg_ParseTupleAndKeywords(args, kwds, "Od|OO", kwlist, &fields, &time,
                                   &part_ids, &types))
    return NULL;

  int read_types[csds_type_count] = {0};
  pyGetData_CheckInput(fields, types, part_ids, read_types);

  /* initialize the reader. */
  struct csds_reader *reader = &self_reader->reader;

  /* Get the field enum from the header. */
  const int n_fields = PyList_Size(fields);
  enum field_enum *fields_wanted =
      (enum field_enum *)malloc(n_fields * sizeof(enum field_enum));
  for (int i = 0; i < n_fields; i++) {
    /* Get the an item in the list. */
    PyObject *field = PyList_GetItem(fields, i);
    if (!PyUnicode_Check(field)) {
      error_python("Expecting list of string for the fields");
    }

    /* Convert into C string. */
    Py_ssize_t size = 0;
    const char *field_name = PyUnicode_AsUTF8AndSize(field, &size);

    /* Set the enum */
    fields_wanted[i] = field_enum_get(field_name);
  }

  /* Set the time. */
  csds_reader_set_time(reader, time);

  /* Get the number of particles. */
  uint64_t n_part[csds_type_count];
  if (part_ids == Py_None) {
    /* If no ids provided, read everything from the index files */
    csds_reader_get_number_particles(reader, n_part, read_types);
  }
  /* If the ids are provided, use them to get the number of particles */
  else {
    for (int i = 0; i < csds_type_count; i++) {
      PyObject *el = PyList_GetItem(part_ids, i);
      n_part[i] = el == Py_None ? 0 : PyArray_Size(el);
    }
  }

  /* Count the total number of particles. */
  uint64_t n_tot = 0;
  for (int i = 0; i < csds_type_count; i++) {
    n_tot += n_part[i];
  }

  /* Print the number of particles */
  if (reader->verbose > 0) {
    message("Found %lu particles", n_tot);
  }

  /* Allocate the output memory. */
  void **output = malloc(n_fields * sizeof(void *));
  for (int i = 0; i < n_fields; i++) {
    unsigned int size = field_enum_get_size(fields_wanted[i]);
    output[i] = malloc(n_tot * size);
  }

  /* Read the particles. */
  if (part_ids == Py_None) {
    /* Enable multithreading */
    Py_BEGIN_ALLOW_THREADS;

    csds_reader_read_all_particles(reader, time, fields_wanted, n_fields,
                                   output, n_part);

    /* Disable multithreading */
    Py_END_ALLOW_THREADS;

  } else {

    /* Get the arrays in a C compatible way */
    PyObject *list[csds_type_count] = {NULL};
    long long *c_part_ids[csds_type_count];
    for (int i = 0; i < csds_type_count; i++) {
      PyObject *el = PyList_GetItem(part_ids, i);

      /* Check if a list is provided */
      if (el == Py_None) {
        c_part_ids[i] = NULL;
        continue;
      }

      /* Copy the ids in order to modify them */
      list[i] = PyArray_NewCopy((PyArrayObject *)el, NPY_CORDER);
      c_part_ids[i] = (long long *)PyArray_DATA((PyArrayObject *)list[i]);
    }

    /* Enable multithreading */
    Py_BEGIN_ALLOW_THREADS;

    /* Read the particles. */
    csds_reader_read_particles_from_ids(reader, time, fields_wanted, n_fields,
                                        output, n_part, c_part_ids);

    /* Disable multithreading */
    Py_END_ALLOW_THREADS;

    /* Free the memory and recompute n_tot */
    n_tot = 0;
    for (int i = 0; i < csds_type_count; i++) {
      if (list[i] != NULL) Py_DECREF(list[i]);

      n_tot += n_part[i];
    }
  }

  /* Create the python output. */
  PyObject *array =
      csds_loader_create_output(output, fields_wanted, n_fields, n_tot);

  /* Free the reader. */
  free(fields_wanted);

  return array;
}

/**
 * @brief Shortcut to #pyGetData for a single particle type.
 */
static PyObject *pyParticleGetData(ATTR_UNUSED PyObject *self, PyObject *args,
                                   PyObject *kwds) {

  PyParticleReader *part_reader = (PyParticleReader *)self;
  PyObject *reader = (PyObject *)part_reader->reader;

  /* input variables. */
  PyObject *fields = NULL;
  double time = 0;
  PyObject *part_ids = Py_None;

  /* List of keyword arguments. */
  static char *kwlist[] = {"fields", "time", "filter_by_ids", NULL};

  /* parse the arguments. */
  if (!PyArg_ParseTupleAndKeywords(args, kwds, "Od|O", kwlist, &fields, &time,
                                   &part_ids))
    return NULL;

  /* Check if we need to create the dictionary */
  const int dict_created = kwds == NULL;
  if (dict_created) {
    kwds = PyDict_New();
  }

  /* Adapt the inputs in case we are filtering */
  PyObject *part_ids_list = part_ids;
  if (part_ids != Py_None) {
    if (!PyArray_Check(part_ids)) {
      error("You need to provide a numpy array of ids");
    }

    part_ids_list = PyList_New(csds_type_count);
    for (int i = 0; i < csds_type_count; i++) {
      PyList_SetItem(part_ids_list, i, Py_None);
    }
    PyList_SetItem(part_ids_list, part_reader->type, part_ids);

    if (PyDict_SetItemString(kwds, "filter_by_ids", part_ids_list) != 0)
      return NULL;
  }
  /* Now the case without filtering */
  else {
    /* Set the particle type  */
    PyObject *part_type = PyLong_FromLong(part_reader->type);

    /* Add the list to the k-arguments */
    if (PyDict_SetItemString(kwds, "part_type", part_type) != 0) return NULL;
  }

  /* Get the fields */
  PyObject *ret = pyGetData(reader, args, kwds);

  /* Cleanup */
  if (dict_created) {
    Py_DecRef(kwds);
    kwds = NULL;
  } else {
    if (part_ids == Py_None) PyDict_DelItemString(kwds, "part_type");
  }

  return ret;
}

/**
 * @brief Update the particles in the cache.
 *
 * @return Dictionary of numpy array containing the updated fields.
 */
static PyObject *pyUpdateParticles(ATTR_UNUSED PyObject *self, PyObject *args,
                                   PyObject *kwds) {

  PyObjectReader *self_reader = (PyObjectReader *)self;
  if (!self_reader->ready) {
    error_python(
        "The CSDS is not ready yet."
        "Did you forget to open it with \"with\"?");
  }

  /* input variables. */
  PyObject *fields = NULL;
  double time = 0;

  /* List of keyword arguments. */
  static char *kwlist[] = {"fields", "time", NULL};

  /* parse the arguments. */
  if (!PyArg_ParseTupleAndKeywords(args, kwds, "Od", kwlist, &fields, &time))
    return NULL;

  /* Declare a few variables just for reusing a function */
  int read_types[csds_type_count] = {0};
  PyObject *types = Py_None;
  PyObject *part_ids = Py_None;
  pyGetData_CheckInput(fields, types, part_ids, read_types);

  /* initialize the reader. */
  struct csds_reader *reader = &self_reader->reader;

  /* Check the time */
  if (time < reader->time.time) {
    error("Cannot go backward in time");
  }

  /* Get the field enum from the header. */
  const int n_fields = PyList_Size(fields);
  enum field_enum *fields_wanted =
      (enum field_enum *)malloc(n_fields * sizeof(enum field_enum));
  for (int i = 0; i < n_fields; i++) {
    /* Get the an item in the list. */
    PyObject *field = PyList_GetItem(fields, i);
    if (!PyUnicode_Check(field)) {
      error_python("Expecting list of string for the fields");
    }

    /* Convert into C string. */
    Py_ssize_t size = 0;
    const char *field_name = PyUnicode_AsUTF8AndSize(field, &size);

    /* Set the enum */
    fields_wanted[i] = field_enum_get(field_name);
  }

  /* Set the time. */
  csds_reader_set_time(reader, time);

  /* Get the number of particles. */
  uint64_t n_part[csds_type_count];
  csds_reader_get_number_particles(reader, n_part, read_types);

  /* Count the total number of particles. */
  uint64_t n_tot = 0;
  for (int i = 0; i < csds_type_count; i++) {
    n_tot += n_part[i];
  }

  /* Check if we have enough cache */
  for (int i = 0; i < csds_type_count; i++) {
    if (n_part[i] != 0 && n_part[i] >= (uint64_t)reader->cache.capacity[i]) {
      error(
          "Not enough memory for particle type %s."
          " Are you sure that you are requesting the correct type?"
          " You can try to increase Cache:OverAllocation or"
          " Cache:InitialAllocation",
          part_type_names[i]);
    }
  }

  /* Print the number of particles */
  if (reader->verbose > 0) {
    message("Found %lu particles", n_tot);
  }

  /* Allocate the output memory. */
  void **output = malloc(n_fields * sizeof(void *));
  for (int i = 0; i < n_fields; i++) {
    unsigned int size = field_enum_get_size(fields_wanted[i]);
    output[i] = malloc(n_tot * size);
  }

  /* Enable multithreading */
  Py_BEGIN_ALLOW_THREADS;

  csds_reader_update_particles(reader, time, fields_wanted, n_fields, output,
                               n_part);

  /* Disable multithreading */
  Py_END_ALLOW_THREADS;

  /* Create the python output. */
  PyObject *array =
      csds_loader_create_output(output, fields_wanted, n_fields, n_tot);

  /* Free the reader. */
  free(fields_wanted);

  return array;
}

/* definition of the method table. */
static PyMethodDef libcsdsMethods[] = {
    {NULL, NULL, 0, NULL} /* Sentinel */
};

/* Definition of the Reader methods */
static PyMethodDef libcsdsReaderMethods[] = {
    {"get_time_limits", getTimeLimits, METH_NOARGS,
     "Read the time limits of the simulation.\n\n"
     "Returns\n"
     "-------\n\n"
     "times: tuple\n"
     "  time min, time max\n"},
    {"get_box_size", get_box_size, METH_NOARGS,
     "Gives the box size of the simulation.\n\n"
     "Returns\n"
     "-------\n\n"
     "box_size: tuple\n"
     "  The box size.\n"},
    {"get_data", (PyCFunction)pyGetData, METH_VARARGS | METH_KEYWORDS,
     "Read some fields from the logfile at a given time.\n\n"
     "Parameters\n"
     "----------\n\n"
     "fields: list\n"
     "  The list of fields (e.g. 'Coordinates', 'Entropies', ...)\n\n"
     "time: float\n"
     "  The time at which the fields must be read.\n\n"
     "filter_by_ids: bool\n"
     "  Read only the IDs provided.\n\n"
     "part_type: int, list"
     "  Types to read (default all). This can be a single int or a list"
     " of boolean where the index corresponds to the particle type.\n\n"
     "Returns\n"
     "-------\n\n"
     "list_of_fields: list\n"
     "  Each element is a numpy array containing the corresponding field.\n"},
    {"update_particles", (PyCFunction)pyUpdateParticles,
     METH_VARARGS | METH_KEYWORDS,
     "Update the particles read with get_data using the cache.\n\n"
     "Parameters\n"
     "----------\n\n"
     "fields: list\n"
     "  The list of fields (e.g. 'Coordinates', 'Entropies', ...)\n\n"
     "time: float\n"
     "  The time at which the fields must be read.\n\n"
     "Returns\n"
     "-------\n\n"
     "list_of_fields: list\n"
     "  Each element is a numpy array containing the corresponding field.\n"},
    {"get_list_fields", (PyCFunction)pyGetListFields,
     METH_VARARGS | METH_KEYWORDS,
     "Read the list of available fields in the logfile.\n\n"
     "Parameters\n"
     "----------\n\n"
     "type: int, list\n"
     "  The particle type for the list of fields\n\n"
     "Returns\n"
     "-------\n"
     "fields: tuple\n"
     "  The list of fields present in the logfile.\n"},
    {"__enter__", pyEnter, METH_VARARGS, ""},
    {"__exit__", pyExit, METH_VARARGS, ""},
    {NULL, NULL, 0, NULL} /* Sentinel */
};

static struct PyModuleDef libcsdsmodule = {
    PyModuleDef_HEAD_INIT,
    "libcsds",
    "Module reading a Continuous Simulation Data Stream",
    -1,
    libcsdsMethods,
    NULL, /* m_slots */
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL  /* m_free */
};

static PyMemberDef PyObjectReader_members[] = {
    {"gas", T_OBJECT_EX, offsetof(PyObjectReader, part_reader), READONLY},
    {"dark_matter", T_OBJECT_EX,
     offsetof(PyObjectReader, part_reader) + sizeof(PyParticleReader *),
     READONLY},
    {"dark_matter_background", T_OBJECT_EX,
     offsetof(PyObjectReader, part_reader) + 2 * sizeof(PyParticleReader *),
     READONLY},
    {"sinks", T_OBJECT_EX,
     offsetof(PyObjectReader, part_reader) + 3 * sizeof(PyParticleReader *),
     READONLY},
    {"stars", T_OBJECT_EX,
     offsetof(PyObjectReader, part_reader) + 4 * sizeof(PyParticleReader *),
     READONLY},
    {"black_holes", T_OBJECT_EX,
     offsetof(PyObjectReader, part_reader) + 5 * sizeof(PyParticleReader *),
     READONLY},
    {NULL}};

static PyTypeObject PyObjectReader_Type = {
    PyVarObject_HEAD_INIT(NULL, 0).tp_name = "libcsds.Reader",
    .tp_basicsize = sizeof(PyObjectReader),
    .tp_flags = Py_TPFLAGS_DEFAULT,
    .tp_doc =
        "Deals with a CSDS file and provides the data through"
        " its different methods. The reader is really open with the method "
        "__enter__"
        "(called by `with ...``). Outside the area delimited by with, "
        "the CSDS is not accessible.\n\n"
        "Parameters\n"
        "----------\n\n"
        "basename: str\n"
        "    Basename of the logfile.\n"
        "verbose: int (optional)\n"
        "    Verbose level\n\n"
        "number_index: int (optional)\n"
        "    Number of index files to generate\n\n"
        "restart_init: bool (optional)\n"
        "    Are we restarting an initialization?\n\n"
        "use_cache: bool (optional)\n"
        "    Use a cache for the evolution of particles?\n\n"
        "Methods\n"
        "-------\n\n"
        "get_time_limits\n"
        "    Provides the initial and final time of the simulation\n"
        "get_data\n"
        "    Reads some particle's data from the logfile\n\n"
        "Examples\n"
        "--------\n\n"
        ">>> import libcsds as csds\n"
        ">>> with csds.Reader(\"index_0000\") as reader:\n"
        ">>>    t0, t1 = reader.get_time_limits()\n"
        ">>>    fields = reader.get_list_fields()\n"
        ">>>    if \"Coordinates\" not in fields:\n"
        ">>>        raise Exception(\"Field Coordinates not present in the "
        "logfile.\")\n"
        ">>>    pos, ent = reader.get_data([\"Coordinates\", "
        "\"Entropies\"]"
        ", 0.5 * (t0 + t1))\n",
    .tp_init = (initproc)Reader_init,
    .tp_dealloc = (destructor)Reader_dealloc,
    .tp_new = Reader_new,
    .tp_methods = libcsdsReaderMethods,
    .tp_members = PyObjectReader_members,
};

/* Definition of the Reader methods */
static PyMethodDef libcsdsParticleReaderMethods[] = {
    {"get_data", (PyCFunction)pyParticleGetData, METH_VARARGS | METH_KEYWORDS,
     "Read some fields from the logfile at a given time.\n\n"
     "Parameters\n"
     "----------\n\n"
     "fields: list\n"
     "  The list of fields (e.g. 'Coordinates', 'Entropies', ...)\n\n"
     "time: float\n"
     "  The time at which the fields must be read.\n\n"
     "Returns\n"
     "-------\n\n"
     "list_of_fields: list\n"
     "  Each element is a numpy array containing the corresponding field.\n"},
    {"get_list_fields", (PyCFunction)pyParticleGetListFields,
     METH_VARARGS | METH_KEYWORDS,
     "Read the list of available fields in the logfile.\n\n"
     "Parameters\n"
     "----------\n\n"
     "type: int, list\n"
     "  The particle type for the list of fields\n\n"
     "Returns\n"
     "-------\n"
     "fields: tuple\n"
     "  The list of fields present in the logfile.\n"},
    {NULL, NULL, 0, NULL} /* Sentinel */
};

static PyTypeObject PyParticleReader_Type = {
    PyVarObject_HEAD_INIT(NULL, 0).tp_name = "libcsds.ParticleReader",
    .tp_basicsize = sizeof(PyParticleReader),
    .tp_flags = Py_TPFLAGS_DEFAULT,
    .tp_doc =
        "Implements libcsds.Reader for a single particle type.\n"
        "See the documentation of libcsds.Reader for more information.\n",
    .tp_init = NULL,
    .tp_dealloc = (destructor)particle_reader_dealloc,
    .tp_new = particle_reader_new,
    .tp_methods = libcsdsParticleReaderMethods,
};

PyMODINIT_FUNC PyInit_libcsds(void) {

  /* Create the module. */
  PyObject *m;
  m = PyModule_Create(&libcsdsmodule);
  if (m == NULL) return NULL;

  /* Deal with CSDS clock */
  clocks_set_cpufreq(0);

  /* Prepare the classes */
  /* Object reader */
  if (PyType_Ready(&PyObjectReader_Type) < 0) {
    return NULL;
  }
  Py_INCREF(&PyObjectReader_Type);

  /* Particle reader */
  if (PyType_Ready(&PyParticleReader_Type) < 0) {
    return NULL;
  }
  Py_INCREF(&PyParticleReader_Type);

  /* Add the reader object */
  PyModule_AddObject(m, "Reader", (PyObject *)&PyObjectReader_Type);
  PyModule_AddIntConstant(m, "gas", csds_type_gas);
  PyModule_AddIntConstant(m, "dark_matter", csds_type_dark_matter);
  PyModule_AddIntConstant(m, "dark_matter_background",
                          csds_type_dark_matter_background);
  PyModule_AddIntConstant(m, "sinks", csds_type_sink);
  PyModule_AddIntConstant(m, "stars", csds_type_stars);
  PyModule_AddIntConstant(m, "black_holes", csds_type_black_hole);

  /* Import numpy. */
  import_array();

  return m;
}

#endif  // HAVE_PYTHON
