/*******************************************************************************
 * This file is part of CSDS.
 * Copyright (c) 2019 Loic Hausammann (loic.hausammann@epfl.ch)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 ******************************************************************************/

/* Include corresponding header */
#include "csds_reader.h"

/* Include CSDS headers */
#include "csds_fields.h"
#include "csds_openmp.h"
#include "csds_parameters.h"

/* Include standard library */
#include <sys/sysinfo.h>
#include <unistd.h>

/**
 * @brief Initialize the reader.
 *
 * @param reader The #csds_reader.
 * @param basename The basename of the csds files.
 * @param verbose The verbose level.
 * @param number_index The number of index files to write.
 * @param restart_init Are we restarting the initial reading?
 * @param use_cache Are we using the cache?
 */
void csds_reader_init(struct csds_reader *reader, const char *basename,
                      int verbose, int number_index, int restart_init,
                      int use_cache) {
  if (verbose > 1) message("Initializing the reader.");

  /* Set the variable to the default values */
  reader->time.time = -1.;
  reader->time.int_time = 0;
  reader->time.time_offset = 0;
  reader->use_cache = use_cache;

  /* Copy the base name */
  strcpy(reader->basename, basename);

  /* Initialize the reader variables. */
  reader->verbose = verbose;
  csds_parameters_init(&reader->params, reader);

  /* Generate the logfile filename */
  char logfile_name[CSDS_STRING_SIZE];
  sprintf(logfile_name, "%s.dump", basename);

  /* Initialize the log file. */
  csds_logfile_init_from_file(&reader->log, logfile_name, reader,
                              /* only_header */ 0);

  /* Initialize the index files and creates them if needed */
  csds_reader_init_index(reader, number_index, restart_init);

  /* Save the time array */
  time_array_save(&reader->log.times, &reader->log);

  /* Initialize the cache */
  csds_cache_empty_init(&reader->cache);

  if (verbose > 1) message("Initialization done.");
}

/**
 * @brief Initialize the index part of the reader.
 *
 * @param reader The #csds_reader.
 * @param number_index Number of requested index files.
 * @param restart Are we restarting the generation of index files?
 */
void csds_reader_init_index(struct csds_reader *reader, int number_index,
                            int restart) {
  /* Initialize the csds_index */
  csds_index_init(&reader->index.index_prev, reader);
  csds_index_init(&reader->index.index_next, reader);

  /* Count the number of files */
  int count = 0;
  while (1) {
    char filename[CSDS_STRING_SIZE + 50];
    sprintf(filename, "%s_%04i.index", reader->basename, count);

    /* Check if file exists */
    if (access(filename, F_OK) != -1) {
      count++;
    } else {
      break;
    }
  }

  /* Ensures that we do not have an unexpected situation */
  if (number_index == 1 || (count == 1 && !restart)) {
    error(
        "The CSDS cannot run with only one index file."
        " Did you forget to restart their generation?");
  }

  /* Check if need to generate the index files or not. */
  if ((count != number_index && restart) || count == 0) {
    csds_reader_generate_index_files(reader, number_index, count);
    count = number_index;
  } else if (reader->verbose > 0 && count != number_index) {
    message(
        "Found %i index files. If you wish to have the"
        " requested %i files, please delete the old ones.",
        count, number_index);
  }

  reader->index.n_files = count;

  /* Initialize the arrays */
  if ((reader->index.times = (double *)malloc(count * sizeof(double))) ==
      NULL) {
    error_python("Failed to allocate the list of times");
  }
  if ((reader->index.int_times =
           (integertime_t *)malloc(count * sizeof(integertime_t))) == NULL) {
    error_python("Failed to allocate the list of times");
  }

  /* Get the information contained in the headers */
  for (int i = 0; i < reader->index.n_files; i++) {
    char filename[CSDS_STRING_SIZE + 50];
    sprintf(filename, "%s_%04i.index", reader->basename, i);

    /* Read the header */
    csds_index_read_header(&reader->index.index_prev, filename);

    /* Save the required information */
    reader->index.times[i] = reader->index.index_prev.time;
    reader->index.int_times[i] = reader->index.index_prev.integer_time;
  }
}

/**
 * @brief Free the reader.
 *
 * @param reader The #csds_reader.
 */
void csds_reader_free(struct csds_reader *reader) {
  /* Free the log. */
  csds_logfile_free(&reader->log);

  /* Free the index */
  if (reader->time.time != -1.) {
    csds_index_free(&reader->index.index_prev);
  }

  free(reader->index.int_times);
  free(reader->index.times);

  /* Free the cache */
  if (reader->use_cache) csds_cache_free(&reader->cache);
}

/**
 * @brief Set the reader to a given time and read the correct index file.
 *
 * @param reader The #csds_reader.
 * @param time The requested time.
 */
void csds_reader_set_time(struct csds_reader *reader, double time) {
  /* Set the time */
  reader->time.time = time;

  unsigned int left = 0;
  unsigned int right = reader->index.n_files - 1;

  /* Check if the time is meaningful */
  if (reader->index.times[left] > time || reader->index.times[right] < time)
    error("The requested time (%f) is not within the index limits [%f, %f]",
          time, reader->index.times[left], reader->index.times[right]);

  /* Find the correct index */
  while (left != right) {
    /* Do a ceil - division */
    unsigned int m = (left + right + 1) / 2;
    if (reader->index.times[m] > time) {
      right = m - 1;
    } else {
      left = m;
    }
  }

  /* Deal with the final time */
  if (left == (unsigned int)reader->index.n_files - 1) {
    left -= 1;
  }

  /* Generate the filename */
  char filename_prev[CSDS_STRING_SIZE + 50];
  sprintf(filename_prev, "%s_%04u.index", reader->basename, left);
  char filename_next[CSDS_STRING_SIZE + 50];
  sprintf(filename_next, "%s_%04u.index", reader->basename, left + 1);

  /* Check if the file is already mapped */
  if (reader->index.index_prev.index.map != NULL) {
    csds_index_free(&reader->index.index_prev);
  }
  if (reader->index.index_next.index.map != NULL) {
    csds_index_free(&reader->index.index_next);
  }

  /* Read the file */
  csds_index_read_header(&reader->index.index_prev, filename_prev);
  csds_index_map_file(&reader->index.index_prev, filename_prev,
                      /* sorted */ 1);

  csds_index_read_header(&reader->index.index_next, filename_next);
  csds_index_map_file(&reader->index.index_next, filename_next,
                      /* sorted */ 1);

  /* Get the offset of the time chunk */
  size_t ind = time_array_get_index_from_time(&reader->log.times, time);
  /* ind == 0 and ind == 1 are the same time, but when reading we need
     data before the initial time. */
  if (ind == 0) {
    ind = 1;
  }

  /* Check if we requested exactly a time step  */
  if (reader->log.times.records[ind].time != time) {
    /* In order to interpolate, we need to be above and not below the time */
    ind += 1;
  }

  /* Save the values */
  reader->time.index = ind;
  reader->time.int_time = reader->log.times.records[ind].int_time;
  reader->time.time_offset = reader->log.times.records[ind].offset;
}

/**
 * @brief Count the number of new particles at the time set since last index
 * file.
 *
 * @param reader The #csds_reader.
 * @param part_type The index corresponding to the particle type.
 *
 * @param The number of new particles.
 */
size_t csds_reader_count_number_new_particles(struct csds_reader *reader,
                                              enum part_type part_type) {

  const size_t threshold = reader->time.time_offset;

  /* Get the history of created particles. */
  struct index_data *data =
      csds_index_get_created_history(&reader->index.index_next, part_type);

  /* Do we have any new particle? */
  if (reader->index.index_next.nparts_created[part_type] == 0 ||
      threshold < data[0].offset) {
    return 0;
  }

  /* Perform a binary search */
  size_t right = reader->index.index_next.nparts_created[part_type] - 1;
  size_t left = 0;
  while (left <= right) {
    size_t m = (left + right) / 2;
    if (data[m].offset < threshold) {
      left = m + 1;
    } else if (data[m].offset > threshold) {
      right = m - 1;
    } else {
      return m;
    }
  }

  /* Compute the return value. */
  size_t ret = (left + right) / 2;

  /* Check if the binary search is correct. */
  if (data[ret].offset > threshold) {
    error_python("The binary search has failed.");
  }

  /* We need the count, not the index. */
  return ret + 1;
}

/**
 * @brief Count the number of removed particles at the time set since last index
 * file.
 *
 * @param reader The #csds_reader.
 * @param n_type (output) The number of particle type possible.
 * @param part_type The index corresponding to the particle type.
 *
 * @param The number of removed particles.
 */
size_t csds_reader_count_number_removed_particles(struct csds_reader *reader,
                                                  enum part_type part_type) {

  const size_t threshold = reader->time.time_offset;

  /* Get the history of created particles. */
  struct index_data *data =
      csds_index_get_removed_history(&reader->index.index_next, part_type);

  /* Do we have any new particle? */
  if (reader->index.index_next.nparts_removed[part_type] == 0 ||
      threshold < data[0].offset) {
    return 0;
  }

  /* Perform a binary search */
  size_t right = reader->index.index_next.nparts_removed[part_type] - 1;
  size_t left = 0;
  while (left <= right) {
    size_t m = (left + right) / 2;
    if (data[m].offset < threshold) {
      left = m + 1;
    } else if (data[m].offset > threshold) {
      right = m - 1;
    } else {
      return m;
    }
  }

  /* Compute the return value. */
  size_t ret = (left + right) / 2;

  /* Check if the binary search is correct. */
  if (data[ret].offset > threshold) {
    error_python("The binary search has failed.");
  }

  /* We need the count, not the index. */
  return ret + 1;
}

/**
 * @brief Provides the number of particle (per type) from the index file.
 *
 * @param reader The #csds_reader.
 * @param n_parts (out) Number of particles at the time set in the reader.
 * @param read_types Should we read this type of particles?
 */
void csds_reader_get_number_particles(struct csds_reader *reader,
                                      uint64_t *n_parts,
                                      const int *read_types) {
  for (enum part_type i = (enum part_type)0; i < csds_type_count; i++) {
    /* Should we skip this type of particles? */
    if (read_types[i] == 0) {
      n_parts[i] = 0;
      continue;
    }

    /* Count the number of particles present in the last index file. */
    const uint64_t count_prev = reader->index.index_prev.nparts[i];
    /* Count the number of particles that have been created since last index. */
    const uint64_t count_new =
        csds_reader_count_number_new_particles(reader, i);
    /* Count the number of particles that have been removed since last index. */
    const uint64_t count_removed =
        csds_reader_count_number_removed_particles(reader, i);
    n_parts[i] = (count_prev + count_new) - count_removed;
  }
}

/**
 * @brief Convert the fields from global indexes to local.
 *
 * @param reader The #csds_reader.
 * @param fields_wanted_wanted The fields to convert.
 * @param fields_info_wanted (out) #field_information for the wanted fields
 * (need to be allocated).
 * @param n_fields_wanted Number of elements in fields_wanted,
 * local_fields_wanted and its derivatives.
 * @param type The type of the particle.
 */
void csds_reader_get_fields_wanted(
    const struct csds_reader *reader, const enum field_enum *fields_wanted,
    const struct field_information **fields_found, const int n_fields_wanted,
    enum part_type type) {

  const struct header *h = &reader->log.header;

  /* Mark fields_found in order to check that the fields are found. */
  for (int i = 0; i < n_fields_wanted; i++) {
    fields_found[i] = NULL;
  }

  /* Find the corresponding fields */
  for (int i = 0; i < n_fields_wanted; i++) {

    for (int j = 0; j < h->number_fields[type]; j++) {
      /* Copy the structure if the field is found.  */
      if (h->fields[type][j].field.field == fields_wanted[i]) {
        fields_found[i] = &h->fields[type][j];
        break;
      }
    }
  }

  /* Check that we found the fields */
  for (int i = 0; i < n_fields_wanted; i++) {
    if (fields_found[i] == NULL) {
      const char *name = field_enum_get_name(fields_wanted[i]);
      error_python("Field %s not found in particle type %s", name,
                   part_type_names[type]);
    }
  }
}

/**
 * @brief Read a single particle.
 *
 * @param reader The #csds_reader
 * @param fields_wanted The list of fields to return.
 * @param n_fields_wanted The number of fields to return
 * @param output The pointers to the output arrays
 * @param offset_from_index The offset of the particle to read
 * @param output_index The index of the particle within the output arrays
 * @param type The type of particle
 * @parma number_jumps (out) The number of jumps required to get the particle
 */
int csds_reader_read_particle(struct csds_reader *reader, double time,
                              const struct field_information **fields_wanted,
                              int n_fields_wanted, void **output,
                              size_t offset_from_index, size_t output_index,
                              enum part_type type, int *number_jumps) {

  /* Get global variables */
  const struct header *h = &reader->log.header;
  size_t offset_time = reader->time.time_offset;

  /* Index in the cache */
  int64_t cache_index = -1;

  /* Check if the offsets are correct. */
  if (offset_from_index > offset_time) {
    error_python("Last offset is larger than the requested time.");
  }

  /* Record's header information */
  mask_type mask = 0;
  size_t h_offset;

  /* The offset used for all the manipulations */
  size_t offset = offset_from_index;
  *number_jumps = 0;

  /* Find the data for the previous record.
     As we start from a full record,
     no need to check if the field is found.
  */
  while (offset < offset_time) {
    *number_jumps += 1;
    /* Read the particle. */
    csds_mapped_file_read_mask(&reader->log.log, offset, &mask, &h_offset);

    /* Is the particle removed from the logfile? */
    if (mask & CSDS_SPECIAL_FLAGS_MASK) {
      int data = 0;
      int part_type = 0;
      enum csds_special_flags flag = csds_particle_read_special_flag(
          offset, &mask, &h_offset, &data, &part_type, &reader->log.log);

#ifdef CSDS_DEBUG_CHECKS
      if (part_type != type) {
        error("Found particles that do not correspond.");
      }
#endif

      if (flag == csds_flag_change_type || flag == csds_flag_mpi_exit ||
          flag == csds_flag_delete) {
        return 1;
      }
    }

    /* Check if there is a next record (avoid infinite loop). */
    if (h_offset == 0) {
      error_python("No next record found.");
    }

    /* Go to the next record. */
    offset += h_offset;
  }

  /* Set the next record */
  size_t offset_next = offset;
  /* Go back before offset_time */
  offset -= h_offset;

  /* Loop over each field. */
  for (int field = 0; field < n_fields_wanted; field++) {
    const struct field_information *current_field = fields_wanted[field];
    void *current_output =
        (char *)output[field] + output_index * current_field->field.size;

    /* Read the field */
    csds_particle_read_field(offset, current_output, current_field,
                             /* derivative */ 0, &mask, &h_offset,
                             h->fields[type], h->number_fields[type],
                             &reader->log.log);

    /* Deal with the first derivative. */
    int first_found = current_field->first.field != field_enum_none &&
                      mask & current_field->first.mask;
    char first_deriv[current_field->first.size];

    if (first_found) {
      /* Read the first derivative */
      csds_particle_read_field(offset, first_deriv, current_field,
                               /* derivative */ 1, &mask, &h_offset,
                               h->fields[type], h->number_fields[type],
                               &reader->log.log);
    }

    /* Deal with the second derivative. */
    int second_found = current_field->second.field != field_enum_none &&
                       mask & current_field->second.mask;
    char second_deriv[current_field->second.size];

    if (second_found) {
      /* Read the first derivative */
      csds_particle_read_field(offset, second_deriv, current_field,
                               /* derivative */ 2, &mask, &h_offset,
                               h->fields[type], h->number_fields[type],
                               &reader->log.log);
    }

    /* Get the time. */
    // TODO reduce search interval
    double time_before = time_array_get_time(&reader->log.times, offset);

    /* Get the mask */
    csds_mapped_file_read_mask(&reader->log.log, offset_next, &mask, &h_offset);

    /* Output after the requested time. */
    char output_after[current_field->field.size];

    /* Read the field */
    csds_particle_read_field(offset_next, output_after, current_field,
                             /* derivative */ 0, &mask, &h_offset,
                             h->fields[type], h->number_fields[type],
                             &reader->log.log);

    /* Deal with the first derivative. */
    char first_deriv_after[current_field->first.size];

    /* Did we find the derivative before and in this record? */
    first_found = current_field->first.field != field_enum_none &&
                  first_found && mask & current_field->first.mask;
    if (first_found) {
      /* Read the first derivative */
      csds_particle_read_field(offset_next, first_deriv_after, current_field,
                               /*derivative*/ 1, &mask, &h_offset,
                               h->fields[type], h->number_fields[type],
                               &reader->log.log);
    }

    /* Deal with the second derivative. */
    char second_deriv_after[current_field->second.size];

    /* Did we find the derivative before and in this record? */
    second_found = current_field->second.field != field_enum_none &&
                   second_found && mask & current_field->second.mask;
    if (second_found) {
      /* Read the second derivative */
      csds_particle_read_field(offset_next, second_deriv_after, current_field,
                               /* derivative */ 2, &mask, &h_offset,
                               h->fields[type], h->number_fields[type],
                               &reader->log.log);
    }

    /* Get the time. */
    // TODO reduce search interval
    double time_after = time_array_get_time(&reader->log.times, offset_next);

    /* Deal with the derivatives */
    struct csds_reader_field before;
    struct csds_reader_field after;
    before.field = current_output;
    before.first_deriv = first_found ? first_deriv : NULL;
    before.second_deriv = second_found ? second_deriv : NULL;
    after.field = output_after;
    after.first_deriv = first_found ? first_deriv_after : NULL;
    after.second_deriv = second_found ? second_deriv_after : NULL;

    /* Interpolate the data. */
    csds_particle_interpolate_field(time_before, &before, time_after, &after,
                                    current_output, time, current_field,
                                    &reader->params);

    /* Now deal with the cache */
    if (reader->use_cache) {
      cache_index = csds_cache_save(&reader->cache, cache_index, type,
                                    time_before, &before, time_after, &after,
                                    offset_next, current_field->field.field);
    }
  }

  return 0;
}

/**
 * @brief Read all the particles of a given type from the index file.
 *
 * @param reader The #csds_reader.
 * @param time The requested time for the particle.
 * @param fields_wanted The fields requested.
 * @param n_fields_wanted Number of field requested.
 * @param output Pointer to the output array. Size: (n_fields_wanted,
 * sum(n_part)).
 * @param n_part Number of particles of each type.
 * @param type The particle type
 */
void csds_reader_read_all_particles_single_type(
    struct csds_reader *reader, double time,
    const enum field_enum *fields_wanted, const int n_fields_wanted,
    void **output, const uint64_t *n_part, enum part_type type) {

  /* Allocate temporary memory. */
  const struct field_information **fields_info_wanted =
      (const struct field_information **)malloc(
          sizeof(struct field_information *) * n_fields_wanted);
  if (fields_info_wanted == NULL) {
    error_python("Failed to allocate the field information.");
  }

  /* Convert fields into the local array. */
  csds_reader_get_fields_wanted(reader, fields_wanted, fields_info_wanted,
                                n_fields_wanted, type);

  /* Count the number of previous parts for the shift in output */
  uint64_t prev_npart = 0;
  for (int i = 0; i < type; i++) {
    prev_npart += n_part[i];
  }

  /* Create some information from the index files */
  const size_t size_index = reader->index.index_prev.nparts[type];
  const size_t size_history =
      csds_reader_count_number_new_particles(reader, type);

  struct index_data *data =
      csds_index_get_data(&reader->index.index_prev, type);
  struct index_data *data_created =
      csds_index_get_created_history(&reader->index.index_next, type);

  /* Current index to read */
  size_t current_index = 0;
  size_t particles_read = 0;
  int local_counter = 0;
  size_t total_number_jumps = 0;
  const int local_update_limit = 1000;
  const ticks tic2 = getticks();

  /* Read the particles */
#pragma omp parallel for firstprivate(local_counter) reduction(+:total_number_jumps)
  for (size_t i = 0; i < n_part[type]; i++) {

    /* Try to find a particle that is not removed */
    int particle_removed = 1;
    while (particle_removed) {

      size_t local_index = 0;
#pragma omp atomic capture
      local_index = current_index++;

      /* Are we reading the history? */
      int reading_history = local_index >= size_index;

      /* Check if we still have some particles available. */
      if (reading_history && local_index == size_history + size_index) {
        error_python("The CSDS was not able to find enough particles.");
      }

      /* Get the offset */
      const size_t current =
          reading_history ? local_index - size_index : local_index;
      size_t offset =
          reading_history ? data_created[current].offset : data[current].offset;

      /* Get the offset of the particle */
      int number_jumps = 0;
      particle_removed = csds_reader_read_particle(
          reader, time, fields_info_wanted, n_fields_wanted, output, offset, i,
          type, &number_jumps);

      /* Update the number of jumps */
      if (!particle_removed) {
        total_number_jumps += number_jumps;
      }
    }

    /* Do we need to print something? */
    if (!(reader->verbose > 0)) continue;

    /* Update the counters */
    local_counter++;
    if (local_counter < local_update_limit) continue;

      /* Add the local counter to the global one */
#pragma omp atomic
    particles_read += local_counter;
    local_counter = 0;

    /* Only the master is printing */
    int current_thread = csds_get_thread_num();
    if (current_thread != 0) continue;

    float ratio = (float)particles_read / (float)n_part[type];

    /* Print the message */
    tools_print_progress(100 * ratio, tic2, "Reading particles");
  }

  /* Print the number of jumps if required */
  if (reader->verbose > 0 || reader->verbose == CSDS_VERBOSE_TIMERS) {
    float avg =
        (float)total_number_jumps / (float)(n_fields_wanted * n_part[type]);
    message("Average number of jumps for %s: %.1f", part_type_names[type], avg);
  }

  /* Free the memory */
  free(fields_info_wanted);
}

/**
 * @brief Read all the particles from the index file.
 *
 * @param reader The #csds_reader.
 * @param time The requested time for the particle.
 * @param fields_wanted The fields requested.
 * @param n_fields_wanted Number of field requested.
 * @param output Pointer to the output array. Size: (n_fields_wanted,
 * sum(n_part)).
 * @param n_part Number of particles of each type.
 */
void csds_reader_read_all_particles(struct csds_reader *reader, double time,
                                    const enum field_enum *fields_wanted,
                                    const int n_fields_wanted, void **output,
                                    const uint64_t *n_part) {
  const ticks tic = getticks();

  /* Prepare the cache */
  if (reader->use_cache) {
    csds_cache_free(&reader->cache);
    csds_cache_init(&reader->cache, n_part, fields_wanted, n_fields_wanted,
                    reader->params.cache_over_alloc, &reader->log.header,
                    reader->params.cache_init_size, reader->time.time_offset);
  }

  /* Read the gas */
  if (n_part[csds_type_gas] != 0) {
    csds_reader_read_all_particles_single_type(reader, time, fields_wanted,
                                               n_fields_wanted, output, n_part,
                                               csds_type_gas);
  }

  /* Read the dark matter. */
  if (n_part[csds_type_dark_matter] != 0) {
    csds_reader_read_all_particles_single_type(reader, time, fields_wanted,
                                               n_fields_wanted, output, n_part,
                                               csds_type_dark_matter);
  }
  /* Read the dark matter background. */
  if (n_part[csds_type_dark_matter_background] != 0) {
    csds_reader_read_all_particles_single_type(
        reader, time, fields_wanted, n_fields_wanted, output, n_part,
        csds_type_dark_matter_background);
  }
  /* Read the neutrino dark matter. */
  if (n_part[csds_type_neutrino] != 0) {
    csds_reader_read_all_particles_single_type(reader, time, fields_wanted,
                                               n_fields_wanted, output, n_part,
                                               csds_type_neutrino);
  }
  /* Read the stars. */
  if (n_part[csds_type_stars] != 0) {
    csds_reader_read_all_particles_single_type(reader, time, fields_wanted,
                                               n_fields_wanted, output, n_part,
                                               csds_type_stars);
  }

  /* Print the time */
  if (reader->verbose > 0 || reader->verbose == CSDS_VERBOSE_TIMERS)
    message("took %.3f %s", clocks_from_ticks(getticks() - tic),
            clocks_getunit());
}

/**
 * @brief Read the particles of a given type from the index file and their ids.
 *
 * @param reader The #csds_reader.
 * @param time The requested time for the particle.
 * @param fields_wanted The fields requested.
 * @param n_fields_wanted Number of field requested.
 * @param output Pointer to the output array. Size: (n_fields_wanted,
 * sum(n_part)).
 * @param n_part Number of particles of each type.
 * Updated with the number of particles found.
 * @param type The particle type
 * @param ids The particles' ids.
 * The ids not found are removed from this array.
 */
void csds_reader_read_particles_from_ids_single_type(
    struct csds_reader *reader, double time,
    const enum field_enum *fields_wanted, const int n_fields_wanted,
    void **output, uint64_t *n_part, enum part_type type, long long *ids) {

  /* Count the number of previous parts for the shift in output */
  uint64_t prev_npart = 0;
  for (int i = 0; i < type; i++) {
    prev_npart += n_part[i];
  }

  /* Allocate temporary memory. */
  const struct field_information **fields_info_wanted =
      (const struct field_information **)malloc(
          sizeof(struct field_information *) * n_fields_wanted);
  if (fields_info_wanted == NULL) {
    error_python("Failed to allocate the field information.");
  }

  /* Convert fields into the local array. */
  csds_reader_get_fields_wanted(reader, fields_wanted, fields_info_wanted,
                                n_fields_wanted, type);

  /* Read the particles */
  size_t total_number_jumps = 0;
  for (size_t i = 0; i < n_part[type]; i++) {

    /* Get the offset */
    size_t offset = 0;
    int found =
        csds_index_get_offset(&reader->index.index_prev, ids[i], type, &offset);

    /* Deal with the particles not found */
    if (!found) {
      if (i != n_part[type] - 1) ids[i] = ids[i + 1];
      n_part[type] -= 1;
      i -= 1;
      continue;
    }

    /* Read the particle */
    size_t output_index = i + prev_npart;
    int number_jumps = 0;
    int particle_removed = csds_reader_read_particle(
        reader, time, fields_info_wanted, n_fields_wanted, output, offset,
        output_index, type, &number_jumps);

    total_number_jumps += number_jumps;

    /* Did we found one? */
    if (particle_removed) {
      if (i != n_part[type] - 1) ids[i] = ids[i + 1];
      n_part[type] -= 1;
      i -= 1;
    }
  }

  /* Print the number of jumps if required */
  if (reader->verbose > 1 || reader->verbose == CSDS_VERBOSE_TIMERS) {
    float avg =
        (float)total_number_jumps / (float)(n_fields_wanted * n_part[type]);
    message("Average number of jumps for %s: %.1f", part_type_names[type], avg);
  }

  /* Free the memory */
  free(fields_info_wanted);
}

/**
 * @brief Read the particles from the index file and their ids.
 *
 * @param reader The #csds_reader.
 * @param time The requested time for the particle.
 * @param fields_wanted The fields requested.
 * @param n_fields_wanted Number of field requested.
 * @param output Pointer to the output array. Size: (n_fields_wanted,
 * sum(n_part)).
 * @param n_part Number of particles of each type.
 * Updated with the number of particles found.
 * @param ids The particles' ids.
 * The ids not found are removed from this array.
 */
void csds_reader_read_particles_from_ids(struct csds_reader *reader,
                                         double time,
                                         const enum field_enum *fields_wanted,
                                         const int n_fields_wanted,
                                         void **output, uint64_t *n_part,
                                         long long **ids) {

  const ticks tic = getticks();

  /* Prepare the cache */
  if (reader->use_cache) {
    csds_cache_free(&reader->cache);
    csds_cache_init(&reader->cache, n_part, fields_wanted, n_fields_wanted,
                    reader->params.cache_over_alloc, &reader->log.header,
                    reader->params.cache_init_size, reader->time.time_offset);
  }

  /* Read the gas */
  if (n_part[csds_type_gas] != 0) {
    csds_reader_read_particles_from_ids_single_type(
        reader, time, fields_wanted, n_fields_wanted, output, n_part,
        csds_type_gas, ids[csds_type_gas]);
  }

  /* Read the dark matter. */
  if (n_part[csds_type_dark_matter] != 0) {
    csds_reader_read_particles_from_ids_single_type(
        reader, time, fields_wanted, n_fields_wanted, output, n_part,
        csds_type_dark_matter, ids[csds_type_dark_matter]);
  }
  /* Read the dark matter background. */
  if (n_part[csds_type_dark_matter_background] != 0) {
    csds_reader_read_particles_from_ids_single_type(
        reader, time, fields_wanted, n_fields_wanted, output, n_part,
        csds_type_dark_matter_background,
        ids[csds_type_dark_matter_background]);
  }
  /* Read the stars. */
  if (n_part[csds_type_stars] != 0) {
    csds_reader_read_particles_from_ids_single_type(
        reader, time, fields_wanted, n_fields_wanted, output, n_part,
        csds_type_stars, ids[csds_type_stars]);
  }

  /* Print the time */
  if (reader->verbose > 0 || reader->verbose == CSDS_VERBOSE_TIMERS)
    message("took %.3f %s", clocks_from_ticks(getticks() - tic),
            clocks_getunit());
}

/**
 * @brief Get the simulation initial time.
 *
 * @param reader The #csds_reader.
 *
 * @return The initial time
 */
double csds_reader_get_time_begin(struct csds_reader *reader) {
  return reader->log.times.records[0].time;
}

/**
 * @brief Get the simulation final time.
 *
 * @param reader The #csds_reader.
 *
 * @return The final time
 */
double csds_reader_get_time_end(struct csds_reader *reader) {
  const size_t ind = reader->log.times.size;
  return reader->log.times.records[ind - 1].time;
}

/**
 * @brief Get the offset of the last timestamp before a given time.
 *
 * @param reader The #csds_reader.
 * @param time The requested time.
 *
 * @return The offset of the timestamp.
 */
size_t csds_reader_get_next_offset_from_time(struct csds_reader *reader,
                                             double time) {
  size_t ind = time_array_get_index_from_time(&reader->log.times, time);
  /* We do not want to have the sentiel */
  if (reader->log.times.size - 2 == ind) {
    ind -= 1;
  }
  return reader->log.times.records[ind + 1].offset;
}

/**
 * @brief Read a record without knowing if it is a particle or a timestamp.
 *
 * WARNING This function asssumes that all the particles are hydro particles.
 * Thus it should be used only for testing the code.
 *
 * @param reader The #csds_reader.
 * @param output The already allocated buffer containing all the fields possible
 * for an hydro particle. (out) The particle if the record is a particle
 * @param time (out) The time if the record is a timestamp.
 * @param is_particle (out) 1 if the record is a particle 0 otherwise.
 * @param offset The offset of the record to read.
 *
 * @return The offset after the record.
 */
size_t csds_reader_read_record(struct csds_reader *reader, void **output,
                               double *time, int *is_particle, size_t offset) {

  /* Get a few pointers. */
  const struct header *h = &reader->log.header;

  mask_type mask = 0;
  size_t h_offset = 0;

  /* Read the record's mask. */
  csds_mapped_file_read_mask(&reader->log.log, offset, &mask, &h_offset);

  *is_particle = !(mask & CSDS_TIMESTAMP_MASK);
  /* The record is a particle. */
  if (*is_particle) {
    size_t offset_tmp = offset;
    for (int i = 0; i < h->number_fields[csds_type_gas]; i++) {
      offset = csds_particle_read_field(
          offset_tmp, output[i], &h->fields[csds_type_gas][i],
          /* derivative */ 0, &mask, &h_offset, h->fields[csds_type_gas],
          h->number_fields[csds_type_gas], &reader->log.log);
    }
  }
  /* The record is a timestamp. */
  else {
    struct time_record time_record;
    offset = time_read(&time_record, reader, offset);
    *time = time_record.time;
  }

  return offset;
}

/**
 * @brief Update a particle in the cache
 *
 * @param cache The #csds_cache.
 * @param index The index of the particle.
 * @param type The type of particle.
 * @param time The time for the interpolation
 * @param offset_time The offset of the interpolation time
 * @param field_index The index of the field in the cache
 * @param output The buffer to write the output
 *
 * @return 0 if the particle was successfully updated
 */
int csds_reader_update_single_particle(struct csds_reader *reader, size_t index,
                                       enum part_type type, double time,
                                       size_t offset_time, size_t field_index,
                                       void *output) {

  struct csds_cache *cache = &reader->cache;

  /* Check the index */
  if (index >= (size_t)cache->capacity[type]) {
    error("Requesting an index too large");
  }

  /* Get the time */
  double time_before = cache->last.time[type][index];
  double time_after = cache->next.time[type][index];
  size_t offset_next = cache->next.offset[type][index];

  /* Update the cache if needed */
  if (offset_next < offset_time) {
    mask_type mask = 0;
    size_t h_offset = 0;

    /* Get the new offset */
    while (offset_next < offset_time) {
      /* Read the particle. */
      csds_mapped_file_read_mask(&reader->log.log, offset_next, &mask,
                                 &h_offset);

      /* Is the particle removed? */
      if (mask & CSDS_SPECIAL_FLAGS_MASK) {
        int data = 0;
        int part_type = 0;
        enum csds_special_flags flag = csds_particle_read_special_flag(
            offset_next, &mask, &h_offset, &data, &part_type, &reader->log.log);

#ifdef CSDS_DEBUG_CHECKS
        if (part_type != type) {
          error("Found particles that do not correspond.");
        }
#endif

        if (flag == csds_flag_change_type || flag == csds_flag_mpi_exit ||
            flag == csds_flag_delete) {
          return 1;
        }
      }

      /* Check if there is a next record (avoid infinite loop). */
      if (h_offset == 0) {
        error_python("No next record found.");
      }

      offset_next += h_offset;
    }

    /* Get a few variables for later */
    size_t offset_before = offset_next - h_offset;
    time_before = time_array_get_time(&reader->log.times, offset_before);
    time_after = time_array_get_time(&reader->log.times, offset_next);

    /* Update the cache */
    cache->last.time[type][index] = time_before;
    cache->next.offset[type][index] = offset_next;
    cache->next.time[type][index] = time_after;
    memcpy(cache->last.fields[type] + cache->total_size * index,
           cache->next.fields[type] + cache->total_size * index,
           cache->total_size);

    /* Get the fields for the next record */
    const struct header *h = &reader->log.header;
    for (int i = 0; i < cache->n_fields; i++) {
      struct field_information *info = &cache->fields[type][i];

      /* Read the field */
      char *current_output = cache->next.fields[type] +
                             cache->total_size * index + cache->offset[i];

      csds_particle_read_field(offset_next, current_output, info,
                               /* derivative */ 0, &mask, &h_offset,
                               h->fields[type], h->number_fields[type],
                               &reader->log.log);
      current_output += info->field.size;

      /* Read the first derivative */
      if (info->first.field != field_enum_none) {
        csds_particle_read_field(offset_next, current_output, info,
                                 /* derivative */ 1, &mask, &h_offset,
                                 h->fields[type], h->number_fields[type],
                                 &reader->log.log);
        current_output += info->first.size;
      }

      /* Read the second derivative */
      if (info->second.field != field_enum_none) {
        csds_particle_read_field(offset_next, current_output, info,
                                 /* derivative */ 2, &mask, &h_offset,
                                 h->fields[type], h->number_fields[type],
                                 &reader->log.log);
      }
    }
  }

  /* Get the derivatives */
  struct field_information *field = &cache->fields[type][field_index];

  /* Generate the interpolation structure */
  struct csds_reader_field before;
  struct csds_reader_field after;
  size_t shift = index * cache->total_size + cache->offset[field_index];

  before.field = cache->last.fields[type] + shift;
  after.field = cache->next.fields[type] + shift;
  shift += field->field.size;

  if (field->first.field != field_enum_none) {
    before.first_deriv = cache->last.fields[type] + shift;
    after.first_deriv = cache->next.fields[type] + shift;
    shift += field->first.size;
  } else {
    before.first_deriv = NULL;
    after.first_deriv = NULL;
  }
  if (field->second.field != field_enum_none) {
    before.second_deriv = cache->last.fields[type] + shift;
    after.second_deriv = cache->next.fields[type] + shift;
    shift += field->second.size;
  } else {
    before.first_deriv = NULL;
    after.first_deriv = NULL;
  }

  /* Interpolate the data. */
  csds_particle_interpolate_field(time_before, &before, time_after, &after,
                                  output, time, field, &reader->params);

  return 0;
}

/**
 * @brief Update the particles of a given type from the cache.
 *
 * @param reader The #csds_reader.
 * @param time The requested time for the particle.
 * @param fields_wanted The fields requested.
 * @param n_fields_wanted Number of field requested.
 * @param output Pointer to the output array. Size: (n_fields_wanted,
 * sum(n_part)).
 * @param n_part Number of particles of each type.
 * @param type The particle type
 */
void csds_reader_update_particles_single_type(
    struct csds_reader *reader, double time,
    const enum field_enum *fields_wanted, const int n_fields_wanted,
    void **output, const uint64_t *n_part, enum part_type type) {

  /* Allocate temporary memory. */
  const struct field_information **fields_info_wanted =
      (const struct field_information **)malloc(
          sizeof(struct field_information *) * n_fields_wanted);
  if (fields_info_wanted == NULL) {
    error_python("Failed to allocate the field information.");
  }

  /* Convert fields into the local array. */
  csds_reader_get_fields_wanted(reader, fields_wanted, fields_info_wanted,
                                n_fields_wanted, type);

  /* Count the number of previous parts for the shift in output */
  uint64_t prev_npart = 0;
  for (int i = 0; i < type; i++) {
    prev_npart += n_part[i];
  }

  /* Get the time offset of the current time */
  size_t offset_time = time_array_get_index_from_time(&reader->log.times, time);
  offset_time = reader->log.times.records[offset_time].offset;

  /* Get the field index */
  int *field_index_in_cache = (int *)malloc(n_fields_wanted * sizeof(int));
  if (field_index_in_cache == NULL) {
    error("Failed to allocate the field index");
  }
  for (int field = 0; field < n_fields_wanted; field++) {
    field_index_in_cache[field] =
        csds_cache_get_field_index(&reader->cache, fields_wanted[field], type);
  }

  /* Update the particles */
  for (int64_t part = 0; part < reader->cache.size[type]; part++) {
    /* Get the current particle index */
    size_t ind = reader->cache.current_index[type];
    reader->cache.current_index[type]++;

    for (int field = 0; field < n_fields_wanted; field++) {
      /* Get the output */
      const struct field_information *current_field = fields_info_wanted[field];
      char current_output[current_field->field.size];

      /* Update the particle */
      int particle_removed = csds_reader_update_single_particle(
          reader, ind, type, time, offset_time, field_index_in_cache[field],
          current_output);

      /* Move to the next particle if this one is removed */
      if (particle_removed) {
        csds_cache_remove_particle(&reader->cache, ind, type);
        break;
      }

      /* Write the particle back into the array */
      void *buffer = (char *)output[field] + part * current_field->field.size;
      memcpy(buffer, current_output, current_field->field.size);
    }
  }

  /* Now look for the new particles */

  /* Check if the index files are compatible */
  size_t offset_index = time_array_get_index_from_time(
      &reader->log.times, reader->index.index_prev.time);
  offset_index = reader->log.times.records[offset_index].offset;
  if (offset_index > reader->cache.last_time_offset) {
    message("%zi %zi", offset_index, reader->cache.last_time_offset);
    error("Requesting an update over more than one index file");
  }
  offset_index = time_array_get_index_from_time(&reader->log.times,
                                                reader->index.index_next.time);
  offset_index = reader->log.times.records[offset_index].offset;
  const int do_previous = reader->cache.last_time_offset < offset_index;

  /* Get the histories */
  struct index_data *prev_created =
      csds_index_get_created_history(&reader->index.index_prev, type);
  struct index_data *next_created =
      csds_index_get_created_history(&reader->index.index_next, type);

  /* Go to the last particle created */
  struct index_data *index_data = do_previous ? prev_created : next_created;
  size_t size_max = do_previous ? reader->index.index_prev.nparts_created[type]
                                : reader->index.index_next.nparts_created[type];

  // TODO use interpolation search
  size_t left = 0;
  size_t right = size_max;
  size_t center = 0;
  int contains_particle = right != 0;
  while (contains_particle && left <= right) {
    center = (left + right) / 2;
    const size_t offset_center = index_data[center].offset;

    if (offset_time > offset_center) {
      left = center + 1;
    } else if (offset_time < offset_center) {
      right = center - 1;
    } else {
      break;
    }
  }

  /* The binary search provides the index below the value */
  center++;

  /* Get the size of the fields */
  int fields_size = 0;
  for (int field = 0; field < n_fields_wanted; field++) {
    fields_size += fields_info_wanted[field]->field.size;
  }

  /* Now read all the missing particles */
  for (size_t i = center; i < size_max; i++) {
    /* Stop as soon as we reach the required offset */
    if (index_data[i].offset > offset_time) break;

    /* Now read and store the particle in the cache */
    char buffer[fields_size];
    char *p_buffer[n_fields_wanted];
    p_buffer[0] = (char *)&buffer;
    for (int field = 1; field < n_fields_wanted; field++) {
      p_buffer[field] =
          p_buffer[field - 1] + fields_info_wanted[field]->field.size;
    }
    int number_jumps = 0;
    int removed = csds_reader_read_particle(
        reader, time, fields_info_wanted, n_fields_wanted, (void **)p_buffer,
        index_data[i].offset, /* output_index */ 0, type, &number_jumps);

    /* Deal with the case where the particle is removed */
    if (removed) {
      reader->cache.current_index[type]--;
    }
  }

  /* Do the same for the next index file */
  if (do_previous) {
    index_data = next_created;
    size_max = reader->index.index_next.nparts_created[type];
    for (size_t i = 0; i < size_max; i++) {
      /* Stop as soon as we reach the required offset */
      if (index_data[i].offset > offset_time) break;

      /* Now read and store the particle in the cache */
      char buffer[fields_size];
      char *p_buffer[n_fields_wanted];
      p_buffer[0] = (char *)&buffer;
      for (int field = 1; field < n_fields_wanted; field++) {
        p_buffer[field] =
            p_buffer[field - 1] + fields_info_wanted[field]->field.size;
      }
      int number_jumps = 0;
      int removed = csds_reader_read_particle(
          reader, time, fields_info_wanted, n_fields_wanted, (void **)p_buffer,
          index_data[i].offset, /* output_index */ 0, type, &number_jumps);

      /* Deal with the case where the particle is removed */
      if (removed) {
        reader->cache.current_index[type]--;
      }
    }
  }

  /* Free the memory */
  free(field_index_in_cache);
  free(fields_info_wanted);
}

/**
 * @brief Update the particles from the index file.
 *
 * @param reader The #csds_reader.
 * @param time The requested time for the particle.
 * @param fields_wanted The fields requested.
 * @param n_fields_wanted Number of field requested.
 * @param output Pointer to the output array. Size: (n_fields_wanted,
 * sum(n_part)).
 * @param n_part Number of particles of each type.
 */
void csds_reader_update_particles(struct csds_reader *reader, double time,
                                  const enum field_enum *fields_wanted,
                                  const int n_fields_wanted, void **output,
                                  const uint64_t *n_part) {
  const ticks tic = getticks();

  /* Check if the cache is enabled */
  if (!reader->use_cache) {
    error("Cannot update the particles without cache");
  }

  /* Prepare the cache */
  csds_cache_reset_current_index(&reader->cache);

  /* Update the gas */
  if (n_part[csds_type_gas] != 0) {
    csds_reader_update_particles_single_type(reader, time, fields_wanted,
                                             n_fields_wanted, output, n_part,
                                             csds_type_gas);
  }

  /* Update the dark matter */
  if (n_part[csds_type_dark_matter] != 0) {
    csds_reader_update_particles_single_type(reader, time, fields_wanted,
                                             n_fields_wanted, output, n_part,
                                             csds_type_dark_matter);
  }
  /* Update the dark matter background. */
  if (n_part[csds_type_dark_matter_background] != 0) {
    csds_reader_update_particles_single_type(reader, time, fields_wanted,
                                             n_fields_wanted, output, n_part,
                                             csds_type_dark_matter_background);
  }

  /* Update the neutrino dark matter. */
  if (n_part[csds_type_neutrino] != 0) {
    csds_reader_update_particles_single_type(reader, time, fields_wanted,
                                             n_fields_wanted, output, n_part,
                                             csds_type_neutrino);
  }
  /* Update the stars. */
  if (n_part[csds_type_stars] != 0) {
    csds_reader_update_particles_single_type(reader, time, fields_wanted,
                                             n_fields_wanted, output, n_part,
                                             csds_type_stars);
  }

  /* Update the last offset read */
  reader->cache.last_time_offset = reader->time.time_offset;

  /* Print the time */
  if (reader->verbose > 0 || reader->verbose == CSDS_VERBOSE_TIMERS)
    message("took %.3f %s", clocks_from_ticks(getticks() - tic),
            clocks_getunit());
}
