/*******************************************************************************
 * This file is part of CSDS.
 * Copyright (c) 2019 Loic Hausammann (loic.hausammann@epfl.ch)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 ******************************************************************************/
#include "csds_time.h"

/* System includes */
#include <unistd.h>

/* Local includes */
#include "csds_logfile.h"
#include "csds_mapped_file.h"
#include "csds_reader.h"

/**
 * @brief Check if enough space is available and increase it if required.
 *
 * @param t The #time_array.
 */
void time_array_ensure_size(struct time_array *t) {
  /* Check if we still have some place. */
  if (t->size < t->capacity) return;

  /* Increase the size */
  t->capacity *= 2;

  /* Allocate the new array */
  struct time_record *tmp = malloc(sizeof(struct time_record) * t->capacity);
  if (tmp == NULL) error_python("Failed to allocate the time records.");

  /* Copy the memory */
  memcpy(tmp, t->records, sizeof(struct time_record) * t->size);

  /* Cleanup the memory */
  free(t->records);

  /* Set the pointer to the new array */
  t->records = tmp;
}

/**
 * @brief Add an element to the #time_array.
 *
 * @param t The #time_array.
 * @param int_time The time in integer.
 * @param time The time in double.
 * @param offset The offset of the record.
 */
void time_array_append(struct time_array *t, const integertime_t int_time,
                       const double time, const size_t offset) {

  /* Increase the available space if required */
  time_array_ensure_size(t);

  /* Copy the values */
  t->records[t->size].time = time;
  t->records[t->size].int_time = int_time;
  t->records[t->size].offset = offset;

  /* Increase the size used. */
  t->size += 1;
}

/**
 * @brief read a time record.
 *
 * @param time_record (output) The time read.
 * @param reader The #csds_reader.
 * @param offset The offset to read
 *
 * @return The offset after the time record
 */
size_t time_read(struct time_record *time_record,
                 const struct csds_reader *reader, size_t offset) {

  /* Initialize variables. */
  const struct header *h = &reader->log.header;
  struct mapped_file *log = &h->log->log;

  mask_type mask = 0;
  size_t prev_offset = 0;
  time_record->int_time = 0;
  time_record->time = 0;

  /* read record header. */
  offset = csds_mapped_file_read_mask(log, offset, &mask, &prev_offset);

  /* check if reading a time record. */
  if (CSDS_TIMESTAMP_MASK != mask) error_python("Not a time record.");

  /* read the record. */
  offset = csds_mapped_file_read_data(
      log, offset, sizeof(unsigned long long int), &time_record->int_time);
  offset = csds_mapped_file_read_data(log, offset, sizeof(double),
                                      &time_record->time);

  return offset;
}

/**
 * @brief get offset of first time record
 *
 * @param h file #header
 * @return offset of first time record
 *
 */
size_t time_offset_first_record(const struct header *h) {

  /* Initialize a few variables. */
  size_t offset = h->offset_first_record;
  struct mapped_file *map = &h->log->log;

  mask_type mask = 0;
  csds_mapped_file_read_mask(map, offset, &mask, NULL);

  if (mask != CSDS_TIMESTAMP_MASK)
    error_python("Log file should begin by timestep.");

  return h->offset_first_record;
}

/**
 * @brief Initialize an empty time array.
 *
 * @param t #time_array to initialize.
 */
void time_array_init(struct time_array *t, size_t initial_size) {
  /* Allocate the arrays */
  t->records = malloc(sizeof(struct time_record) * initial_size);
  if (t->records == NULL)
    error_python("Failed to initialize the time records.");

  /* Initialize the sizes */
  t->size = 0;
  t->capacity = initial_size;
}

/**
 * @brief Get the filename of the index file used to save the time array.
 *
 * @param log The #csds_logfile.
 * @param filename (output) The filename of the index file.
 *
 * @return Does the file exists?
 */
int time_array_get_filename_savefile(const struct csds_logfile *log,
                                     char *filename) {
  /* Generate the name */
  sprintf(filename, "%s_0000.index", log->reader->basename);

  /* Check if the file exists. */
  return access(filename, F_OK) == 0;
}

/**
 * @brief Save the array into a file.
 *
 * @param t The #time_array.
 * @param log The #csds_logfile.
 */
void time_array_save(struct time_array *t, const struct csds_logfile *log) {
  /* Get the filename of the saved file. */
  char filename[CSDS_STRING_SIZE + 50];
  int savefile = time_array_get_filename_savefile(log, filename);

  if (!savefile) {
    error("Cannot save the time array without the first index file.");
  }

  /* Open the file */
  FILE *f = fopen(filename, "ab");
  if (f == NULL) {
    error("Failed to open the file %s", filename);
  }

  /* Dump the array */
  uint64_t size = t->size;
  fwrite(&size, sizeof(uint64_t), 1, f);
  fwrite(t->records, sizeof(struct time_record), t->size, f);

  /* Close the file */
  int failed = fclose(f);
  if (failed) {
    error("Failed to close the file %s", filename);
  }
}

/**
 * @brief Load the array from a file
 *
 * @param t The #time_array.
 * @param index The #csds_index that contains the time array.
 */
void time_array_load(struct time_array *t, struct csds_index *index) {
  /* Ensure that the array is empty */
  if (t->capacity != 0) time_array_free(t);

  /* Get the position of the time array. */
  char *map = (char *)csds_index_get_removed_history(index, csds_type_count);

  /* Read the array */
  uint64_t size = 0;
  memcpy(&size, map, sizeof(uint64_t));
  map += sizeof(uint64_t);

  /* Allocate the array */
  time_array_init(t, size);
  t->size = size;

  /* Read the elements */
  memcpy(t->records, map, t->size * sizeof(struct time_record));
}

/**
 * @brief Restore the time array from the savefile.
 *
 * @param t The #time_array.
 * @param log The #csds_logfile.
 *
 * @return Is the time array restored?
 */
int time_array_restore(struct time_array *t, struct csds_logfile *log) {
  /* Get the filename of the saved file. */
  char filename[CSDS_STRING_SIZE + 50];
  int savefile = time_array_get_filename_savefile(log, filename);

  if (!savefile) return 0;

  /* Initialize the index file that might contain the time array */
  struct csds_index index;
  csds_index_init(&index, log->reader);
  csds_index_read_header(&index, filename);
  csds_index_map_file(&index, filename, /* sorted */ 1);

  /* Check if the index file contains the time array */
  if (csds_index_contains_time_array(&index)) {
    if (log->reader->verbose > 0) {
      message("Restoring the time array from %s", filename);
    }

    /* Load the time array */
    time_array_load(t, &index);
    csds_index_free(&index);
    return 1;
  }

  /* Free the memory. */
  csds_index_free(&index);

  return 0;
}

/**
 * @brief Initialize a time array from a file.
 *
 * @param t #time_array to initialize.
 * @param log The #csds_logfile.
 */
void time_array_populate(struct time_array *t, struct csds_logfile *log) {

  /* Try restoring the time array */
  if (time_array_restore(t, log)) {
    return;
  }

  /* get file size. */
  size_t file_size = log->log.mmap_size;

  /* get first timestamp. */
  size_t offset = time_offset_first_record(&log->header);
  while (offset < file_size) {
    /* read current time record and store it. */
    size_t tmp_offset = offset;
    struct time_record time;
    time_read(&time, log->reader, tmp_offset);
    time_array_append(t, time.int_time, time.time, offset);

    /* get next record. */
    int test = tools_get_next_record(&log->header, &log->log, &offset,
                                     log->log.mmap_size);
    if (test == -1) break;
  }
}

/**
 * @brief access the time of a given record (by its offset).
 *
 * @param t #time_array to access.
 * @param offset offset of the record.
 *
 * @return integer time of the record.
 */
integertime_t time_array_get_integertime(struct time_array *t,
                                         const size_t offset) {
  size_t ind = time_array_get_index(t, offset);
  return t->records[ind].int_time;
}

/**
 * @brief access the time of a given record (by its offset).
 *
 * @param t #time_array to access.
 * @param offset offset of the record.
 *
 * @return time of the record.
 */
double time_array_get_time(const struct time_array *t, const size_t offset) {
  size_t ind = time_array_get_index(t, offset);
  return t->records[ind].time;
}

/**
 * @brief Find the index of the last time record written before a given offset.
 *
 * @param t #time_array to access.
 * @param offset offset of the record.
 *
 * @return The index of the last time record.
 */
size_t time_array_get_index(const struct time_array *t, const size_t offset) {

#ifdef CSDS_DEBUG_CHECKS
  if (!t) error_python("NULL pointer.");

  if (offset < t->records[0].offset || offset > t->records[t->size - 1].offset)
    error_python("Offset outside of range. %zi > %zi > %zi",
                 t->records[t->size - 1].offset, offset, t->records[0].offset);
#endif

  /* right will contain the index at the end of the loop */
  size_t left = 0;
  size_t right = t->size - 1;

  /* Find the time_array with the correct offset through a bisection method. */
  // TODO use interpolation search (same for the other binary searches)
  while (left <= right) {
    size_t center = (left + right) / 2;
    const size_t offset_center = t->records[center].offset;

    if (offset > offset_center) {
      left = center + 1;
    } else if (offset < offset_center) {
      right = center - 1;
    } else {
      return center;
    }
  }

  /* Avoid the sentinel */
  if (right == t->size - 1) {
    right = right - 1;
  }

#ifdef CSDS_DEBUG_CHECKS
  if (t->records[right].offset > offset ||
      t->records[right + 1].offset <= offset) {
    error_python("Found the wrong element");
  }

#endif

  return right;
}

/**
 * @brief Find the index of the last time record written before a given time.
 *
 * @param t #time_array to access.
 * @param time The time requested.
 *
 * @return The index of the last time record.
 */
size_t time_array_get_index_from_time(const struct time_array *t,
                                      const double time) {

#ifdef CSDS_DEBUG_CHECKS
  if (!t) error_python("NULL pointer.");

  if (time < t->records[0].time || time > t->records[t->size - 1].time)
    error_python("Time outside of range (%g < %g < %g).", t->records[0].time,
                 time, t->records[t->size - 1].time);
#endif

  /* right will contain the index at the end of the loop */
  size_t left = 0;
  size_t right = t->size - 1;

  /* Find the time_array with the correct time through a bisection method. */
  while (left <= right) {
    size_t center = (left + right) / 2;
    const double time_center = t->records[center].time;

    if (time > time_center) {
      left = center + 1;
    } else if (time < time_center) {
      right = center - 1;
    } else {
      return center;
    }
  }

  /* Avoid the sentinel */
  if (right == t->size - 1) {
    right = right - 1;
  }

#ifdef CSDS_DEBUG_CHECKS
  if (t->records[right].time > time || t->records[right + 1].time <= time) {
    error_python("Found the wrong element");
  }

#endif

  return right;
}

/**
 * @brief free memory of a #time_array
 *
 * @param t #time_array to free
 */
void time_array_free(struct time_array *t) {
  /* Free the arrays */
  free(t->records);
  t->records = NULL;

  /* Reset the counters */
  t->size = 0;
  t->capacity = 0;
}

/**
 * @brief print a #time_array
 *
 * @param t #time_array to print
 */
void time_array_print(const struct time_array *t) {
  const size_t threshold = 5;

  size_t n = t->size;
  size_t up_threshold = n - threshold;

  printf("Times (size %lu): [%lli (%g)", n, t->records[0].int_time,
         t->records[0].time);

  /* Loop over all elements. */
  for (size_t i = 1; i < n; i++) {
    /* Skip the times at the center of the array. */
    if (i < threshold || i > up_threshold)
      printf(", %zi (%g)", t->records[i].offset, t->records[i].time);

    if (i == threshold) printf(", ...");
  }

  printf("]\n");
}

/**
 * @brief print a #time_array (offset)
 *
 * @param t #time_array to print
 */
void time_array_print_offset(const struct time_array *t) {
  const size_t threshold = 4;

  size_t n = t->size;
  size_t up_threshold = n - threshold;

  printf("Times (size %lu): [%lu", n, t->records[0].offset);

  /* Loop over all elements. */
  for (size_t i = 1; i < n; i++) {
    /* Skip the offset in the middle of the array. */
    if (i < threshold || i > up_threshold)
      printf(", %lu", t->records[i].offset);

    if (i == threshold) printf(", ...");
  }

  printf("]\n");
}
