/*******************************************************************************
 * This file is part of CSDS.
 * Copyright (c) 2019 Loic Hausammann (loic.hausammann@epfl.ch)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 ******************************************************************************/
#include "csds_tools.h"

#include "csds_header.h"
#include "csds_mapped_file.h"
#include "csds_particle.h"
#include "csds_reader.h"

#include <stdio.h>

/**
 * @brief get the offset of the next corresponding record.
 *
 * @param h #header structure of the file
 * @param map The logfile
 * @param offset In: initial offset, Out: offset of the next record
 * @param file_size The file size.
 *
 * @return -1 if no next record, otherwise 0
 */
int tools_get_next_record(const struct header *h, struct mapped_file *map,
                          size_t *offset, size_t file_size) {
  if (header_is_forward(h)) return _tools_get_next_record_forward(map, offset);
  if (header_is_backward(h))
    return _tools_get_next_record_backward(h, map, offset, file_size);
  else
    error_python("Offsets are corrupted.");
}

/**
 * @brief internal function of #tools_get_next_record. Should not be used
 * outside.
 *
 * @param map The logfile
 * @param offset (Out) offset of the next record
 *
 * @return error code, -1 if no next record
 */
int _tools_get_next_record_forward(struct mapped_file *map, size_t *offset) {
  size_t diff_offset = 0;

  /* Read the offset. */
  csds_mapped_file_read_mask(map, *offset, NULL, &diff_offset);

  if (diff_offset == 0) return -1;

  /* Set the absolute offset. */
  *offset += diff_offset;
  return 0;
}

/**
 * @brief internal function of #tools_get_next_record. Should not be used (very
 * slow)
 *
 * @param h #header structure of the file
 * @param map The logfile
 * @param offset In: initial offset, Out: offset of the next record
 * @param file_size The file size.
 *
 * @return error code, -1 if no next record
 */
int _tools_get_next_record_backward(const struct header *h,
                                    struct mapped_file *map, size_t *offset,
                                    size_t file_size) {
#ifndef CSDS_DEBUG_CHECKS
  error_python("Should not be used, method too slow");
#endif
  size_t current_offset = *offset;
  size_t record_header = CSDS_MASK_SIZE + CSDS_OFFSET_SIZE;

  while (current_offset < file_size) {
    mask_type mask = 0;
    size_t prev_offset;
    csds_mapped_file_read_mask(map, current_offset, &mask, &prev_offset);

    prev_offset = current_offset - prev_offset - record_header;
    if (*offset == prev_offset) {
      *offset = current_offset - record_header;
      return 0;
    }

    current_offset += header_get_record_size_from_mask(h, mask);
  }

  return -1;
}

/**
 * @brief switch side offset.
 *
 * From current record, switch side of the offset of the previous one.
 * @param h #header structure of the file.
 * @param map file mapping.
 * @param offset position of the record.
 *
 * @return position after the record.
 */
size_t tools_reverse_offset(const struct header *h, struct mapped_file *map,
                            size_t offset) {
  mask_type mask = 0;
  size_t prev_offset = 0;
  const size_t cur_offset = offset;

  /* read mask + offset. */
  offset = csds_mapped_file_read_mask(map, offset, &mask, &prev_offset);

  /* write offset of zero (in case it is the last record). */
  const size_t zero = 0;
  offset -= CSDS_OFFSET_SIZE;
  offset = csds_mapped_file_write_data(map, offset, CSDS_OFFSET_SIZE, &zero);

  /* set offset after current record. */
  offset += header_get_record_size_from_mask(h, mask);
  const size_t after_current_record = offset;

  /* first records do not have a previous partner. */
  if (prev_offset == cur_offset) return after_current_record;
  if (prev_offset > cur_offset)
    error_python("Unexpected offset: header %lu, current %lu.", prev_offset,
                 cur_offset);

  /* modify previous offset. */
  offset = cur_offset - prev_offset + CSDS_MASK_SIZE;
  offset =
      csds_mapped_file_write_data(map, offset, CSDS_OFFSET_SIZE, &prev_offset);

#ifdef CSDS_DEBUG_CHECKS
  mask_type prev_mask = 0;
  offset -= CSDS_MASK_SIZE + CSDS_OFFSET_SIZE;
  csds_mapped_file_read_mask(map, offset, &prev_mask, NULL);

  /* Check if we are not mixing timestamp and particles */
  if ((prev_mask != CSDS_TIMESTAMP_MASK && mask == CSDS_TIMESTAMP_MASK) ||
      (prev_mask == CSDS_TIMESTAMP_MASK && mask != CSDS_TIMESTAMP_MASK))
    error_python("Unexpected mask: %i, got %i.", mask, prev_mask);

#endif  // CSDS_DEBUG_CHECKS

  return after_current_record;
}

/**
 * @brief debugging function checking the offset and the mask of a record.
 *
 * Compare the mask with the one pointed by the header.
 * if the record is a particle, check the id too.
 *
 * @param reader The #csds_reader.
 * @param offset position of the record.
 *
 * @return position after the record.
 */
size_t tools_check_record_consistency(struct csds_reader *reader,
                                      size_t offset) {
#ifndef CSDS_DEBUG_CHECKS
  error_python("Should not check in non debug mode.");
#endif

  const struct header *h = &reader->log.header;
  struct mapped_file *map = &reader->log.log;
  const size_t init_offset = offset;

  mask_type mask;
  size_t pointed_offset;

  /* read mask + offset. */
  offset = csds_mapped_file_read_mask(map, offset, &mask, &pointed_offset);

  /* set offset after current record. */
  offset += header_get_record_size_from_mask(h, mask);
  const size_t offset_ret = offset;

  /* If something happened, skip the check. */
  if (mask & CSDS_SPECIAL_FLAGS_MASK) {
    return offset_ret;
  }

  /* get absolute offset. */
  if (header_is_forward(h))
    pointed_offset += init_offset;
  else if (header_is_backward(h)) {
    if (init_offset < pointed_offset)
      error_python("Offset too large (%lu) at %lu with mask %i.",
                   pointed_offset, init_offset, mask);
    pointed_offset = init_offset - pointed_offset;
  } else {
    error_python("Offset are corrupted.");
  }

  if (pointed_offset == init_offset || pointed_offset == 0) return offset_ret;

  /* read mask of the pointed record. */
  mask_type pointed_mask = 0;
  csds_mapped_file_read_mask(map, pointed_offset, &pointed_mask, NULL);

  /* check if not mixing timestamp and particles. */
  if ((pointed_mask != CSDS_TIMESTAMP_MASK && mask == CSDS_TIMESTAMP_MASK) ||
      (pointed_mask == CSDS_TIMESTAMP_MASK && mask != CSDS_TIMESTAMP_MASK))
    error_python("Error in the offset (mask %i at %lu != %i at %lu).", mask,
                 init_offset, pointed_mask, pointed_offset);

  return offset_ret;
}

/**
 * @brief Remove the previous text and print the progress of current task.
 *
 * @param percent The current progress.
 * @param tic The initial time
 * @param message The message to display before the progress message.
 */
void tools_print_progress(float percent, const ticks tic, const char *message) {
  /* Get the remaining time */
  const int current_time =
      clocks_diff_ticks(getticks(), clocks_start_ticks) / 1000.0;
  const int init_time = clocks_diff_ticks(tic, clocks_start_ticks) / 1000.0;

  int remaining_time = (current_time - init_time) * (100. - percent) / percent;

  /* Compute the time */
  int hour = remaining_time / (60 * 60);
  remaining_time -= hour * 60 * 60;
  int min = remaining_time / 60;
  int sec = remaining_time - min * 60;

  /* Allocate the output string */
  char output[300];

  /* Write the message */
  char *current =
      output +
      sprintf(output, "%s: %2.1f%% done, Remaining time: ", message, percent);

  /* Write the hour */
  if (hour == 0)
    current += sprintf(current, "    ");
  else
    current += sprintf(current, "%.2ih ", hour);

  /* Write the minutes */
  if (hour == 0 && min == 0)
    current += sprintf(current, "      ");
  else
    current += sprintf(current, "%.2imin ", min);

  /* Write the seconds */
  current += sprintf(current, "%.2is", sec);

  /* Print the string */
  printf("\r%s %s", clocks_get_timesincestart(), output);
  fflush(stdout);
}
