/*
 * The file was obtained from https://github.com/tidwall/hashmap.c
 * and was slightly adapted.
 */
// Copyright 2020 Joshua J Baker. All rights reserved.
// Use of this source code is governed by an MIT-style
// license that can be found in the LICENSE file

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

/* CSDS header */
#include "csds_error.h"
#include "csds_hashmap.h"

#define N 10000

static void shuffle(struct index_data *array, size_t numels) {
  struct index_data tmp;
  for (size_t i = 0; i < numels - 1; i++) {
    int j = i + rand() / (RAND_MAX / (numels - i) + 1);
    tmp = array[j];
    array[j] = array[i];
    array[i] = tmp;
  }
}

static size_t deepcount(struct csds_hashmap *map) {
  size_t count = 0;
  for (size_t i = 0; i < map->nbuckets; i++) {
    if (bucket_at(map, i)->dib) {
      count++;
    }
  }
  return count;
}

static void all(void) {
  struct index_data *vals;
  vals = (struct index_data *)malloc(N * sizeof(struct index_data));
  if (vals == NULL) {
    error("Failed to allocate the index array");
  }
  for (int i = 0; i < N; i++) {
    vals[i].id = i;
    vals[i].offset = i;
  }

  struct csds_hashmap *map;

  map = csds_hashmap_new(0);
  if (map == NULL) error("Failed to allocate the hashmap");

  shuffle(vals, N);

  /* Test addition of particles */
  for (size_t i = 0; i < N; i++) {
    assert(map->count == i);
    assert(map->count == csds_hashmap_count(map));
    assert(map->count == deepcount(map));

    assert(!csds_hashmap_get(map, vals[i].id));
    assert(!csds_hashmap_delete(map, vals[i].id));
    assert(!csds_hashmap_set(map, &vals[i]));
  }

  /* Test getting the particles */
  for (size_t i = 0; i < N; i++) {
    struct index_data *data = csds_hashmap_get(map, vals[i].id);
    assert(data);
    assert(data->offset == vals[i].offset);
    assert(data->id == vals[i].id);
  }

  /* Write the particles inside a file */
  const char *filename = "testHashmap.bin";
  FILE *f = fopen(filename, "wb");
  if (f == NULL) error("Failed to open file %s", filename);
  csds_hashmap_write(map, f);
  fclose(f);

  /* Read the particles from the file */
  f = fopen(filename, "rb");
  if (f == NULL) error("Failed to open file %s", filename);
  struct index_data *test =
      (struct index_data *)malloc(N * sizeof(struct index_data));
  if (test == NULL) error("Failed to allocate array");
  size_t count = fread(test, sizeof(struct index_data), N, f);
  assert(count == N);
  fclose(f);

  /* Test the i/o */
  for (size_t i = 0; i < N; i++) {
    struct index_data *data = csds_hashmap_get(map, test[i].id);
    assert(data);
    assert(data->id == test[i].id);
    assert(data->offset == test[i].offset);
  }

  /* Test deleting the particles */
  for (size_t i = 0; i < N; i++) {
    assert(csds_hashmap_delete(map, vals[i].id));
  }

  free(test);
  csds_hashmap_free(map);
  free(vals);
}

#define bench(name, N, code)                                                  \
  {                                                                           \
    {                                                                         \
      if (strlen(name) > 0) {                                                 \
        printf("%-14s ", name);                                               \
      }                                                                       \
      clock_t begin = clock();                                                \
      for (int i = 0; i < N; i++) {                                           \
        (code);                                                               \
      }                                                                       \
      clock_t end = clock();                                                  \
      double elapsed_secs = (double)(end - begin) / CLOCKS_PER_SEC;           \
      printf("%d ops in %.3f secs, %.0f ns/op, %.0f op/sec", N, elapsed_secs, \
             elapsed_secs / (double)N * 1e9, (double)N / elapsed_secs);       \
      printf("\n");                                                           \
    }                                                                         \
  }

static void benchmarks(void) {
  struct index_data *vals = malloc(N * sizeof(struct index_data));
  for (int i = 0; i < N; i++) {
    vals[i].id = i;
    vals[i].offset = i;
  }

  struct csds_hashmap *map;
  map = csds_hashmap_new(0);
  if (map == NULL) {
    error("Failed to initialize the hashmap");
  }

  shuffle(vals, N);
  bench("set", N, {
    struct index_data *v = csds_hashmap_set(map, &vals[i]);
    assert(!v);
  });

  shuffle(vals, N);
  bench("get", N, {
    struct index_data *v = csds_hashmap_get(map, vals[i].id);
    assert(v && v->id == vals[i].id);
  });

  shuffle(vals, N);
  bench("delete", N, {
    struct index_data *v = csds_hashmap_delete(map, vals[i].id);
    assert(v && v->id == vals[i].id);
  });

  csds_hashmap_free(map);

  map = csds_hashmap_new(N);

  shuffle(vals, N);
  bench("set (cap)", N, {
    struct index_data *v = csds_hashmap_set(map, &vals[i]);
    assert(!v);
  });

  shuffle(vals, N);
  bench("get (cap)", N, {
    struct index_data *v = csds_hashmap_get(map, vals[i].id);
    assert(v && v->id == vals[i].id);
  });

  shuffle(vals, N);
  bench("delete (cap)", N, {
    struct index_data *v = csds_hashmap_delete(map, vals[i].id);
    assert(v && v->id == vals[i].id);
  });

  csds_hashmap_free(map);

  free(vals);
}

int main(void) {
  all();
  benchmarks();
}
