/*******************************************************************************
 * This file is part of CSDS.
 * Copyright (C) 2019 Loic Hausammann (loic.hausammann@epfl.ch).
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 ******************************************************************************/

/* Local header */
#include "csds_header.h"
#include "csds_mapped_file.h"
#include "csds_particle.h"
#include "csds_reader.h"

/* Tests header */
#include "generate_log.h"

#include <assert.h>

#define number_parts 100

/** Count the number of active particles. */
int get_number_active_particles(int step, struct csds_part *p) {
  int count = 0;
  for (int i = 0; i < number_parts; i++) {
    if (step % p[i].time_bin == 0) count += 1;
  }
  return count;
}

/**
 * @brief Check that the reader contains the correct data
 *
 * @param reader The #csds_reader.
 */
void check_data(struct csds_reader *reader, struct csds_part *parts) {

  /* No need to check the header, this is already done in testHeader.c */

  /* Get required structures. */
  struct csds_logfile *logfile = &reader->log;
  struct header *h = &reader->log.header;

  /* Create a particle */
  void **output = malloc(h->number_fields[csds_type_gas] * sizeof(void *));
  for (int i = 0; i < h->number_fields[csds_type_gas]; i++) {
    output[i] = malloc(h->fields[csds_type_gas][i].field.size);
    message("%i %i %i", i, h->fields[csds_type_gas][i].field.field,
            h->fields[csds_type_gas][i].field.position);
  }

  /* Define a few variables */
  double time = get_double_time(0);
  int is_particle = 0;
  int step = 0;
  int init_log_all_done = -1;

  /* Number of particle found during this time step. */
  int count = 0;
  /* Set it to an impossible value in order to flag it. */
  const uint64_t id_flag = 5 * number_parts;
  uint64_t previous_id = id_flag;

  /* Get all the fields */
  const struct field_information *field_id =
      header_get_field_from_name(h, "ParticleIDs", csds_type_gas);
  const struct field_information *field_pos =
      header_get_field_from_name(h, "Coordinates", csds_type_gas);

  /* Loop over each record. */
  for (size_t offset =
           csds_reader_read_record(reader, output, &time, &is_particle,
                                   logfile->header.offset_first_record);
       offset < logfile->log.mmap_size;
       offset = csds_reader_read_record(reader, output, &time, &is_particle,
                                        offset)) {

    /* Do the particle case */
    if (is_particle) {
      count += 1;

      /*
        Check that we are really increasing the id in the logfile.
        See the writing part to see that we are always increasing the id.
      */
      const uint64_t current_id = *(uint64_t *)output[field_id->field.position];
      if (previous_id != id_flag && previous_id >= current_id) {
        error("Wrong particle found");
        previous_id = current_id;
      }

      /* Get the corresponding particle */
      if (current_id >= number_parts) error("Wrong id %li", current_id);

      struct csds_part *p = &parts[current_id];
      const double *pos = (double *)output[field_pos->field.position];

      /* Check the record's data. */
      for (int i = 0; i < 3; i++) {
        /* in the first index, we are storing the step information. */
        if (i == 0) {
          double tmp = step;
          /* At the end, we are not updating the particle */
          if (step >= number_steps) {
            int n_1 = number_steps - 1;
            tmp = n_1 - (n_1 % p->time_bin);
          }
          assert(tmp == pos[i]);
        } else
          assert(p->x[i] == i);
      }

    }
    /* Time stamp case. */
    else {
      message("Step: %i", step);
      /* Check if we have the current amount of particles in previous step. */
      if (step != 0 && count != get_number_active_particles(step, parts))
        error(
            "The reader did not find the correct number of particles during "
            "step %i: %i != %i",
            step, count, get_number_active_particles(step, parts));

      /* Avoid the initial log */
      if (init_log_all_done > 0) {
        step += 1;
      }

      init_log_all_done += 1;

      /* Reset some variables. */
      previous_id = id_flag;
      count = 0;

      /* Check the record's data. */
      const int tmp_step = step >= number_steps ? number_steps : step;
      assert(time == get_double_time(tmp_step));
    }
  }

  /* Cleanup */
  for (int i = 0; i < h->number_fields[csds_type_gas]; i++) {
    free(output[i]);
  }
  free(output);
}

int main(void) {

  /*
    First generate the file.
  */

  message("Generating the dump.");

  /* Create required structures. */
  char filename[200] = "test_reader_0000.dump";

  /* Delete the index files */
  remove("test_reader_0000_0000.index");
  remove("test_reader_0000_0001.index");
  remove("test_reader_0000_0002.index");
  remove("test_reader_0000_0003.index");
  remove("test_reader_0000_0004.index");

  /* Initialize the particles. */
  struct csds_part *parts;
  if ((parts = (struct csds_part *)malloc(sizeof(struct csds_part) *
                                          number_parts)) == NULL)
    error("Failed to allocate particles array.");

  /* Write a 'simulation' */
  generate_log(parts, number_parts, filename);

  /*
    Then read the file.
  */

  message("Reading the header.");

  /* Generate required structure for reading. */
  struct csds_reader reader;

  /* Set verbose level. */
  reader.verbose = 1;

  /* Read the header. */
  char basename[200] = "test_reader_0000";
  csds_reader_init(&reader, basename, /* verbose */ 1,
                   /* number_index*/ 5,
                   /* restart */ 0,
                   /* use_cache */ 0);

  /*
    Finally check everything.
  */

  check_data(&reader, parts);

  /* Do some cleanup. */
  free(parts);
  csds_reader_free(&reader);

  return 0;
}
