/*******************************************************************************
 * This file is part of CSDS.
 * Copyright (c) 2019 Loic Hausammann (loic.hausammann@epfl.ch)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 ******************************************************************************/

/* Local headers */
#include "csds_definitions.h"
#include "csds_parser.h"

/* Standard headers */
#include <assert.h>
#include <math.h>
#include <string.h>

int main(void) {

  struct csds_parser parser;
  parser_init("test_parser.yml", &parser);
  parser_read_file("test_parser.yml", &parser);
  parser_print_params(&parser);

  /* scalar reader */
  assert(parser_get_param_char(&parser, "Test:char") == 'c');
  assert(parser_get_param_int(&parser, "Test:int") == 10);
  float ftest = parser_get_param_float(&parser, "Test:float");
  assert(fabsf(ftest - 0.2f) < 1e-6);
  assert(parser_get_param_double(&parser, "Test:double") == 0.2);
  assert(parser_get_param_longlong(&parser, "Test:longlong") == 1254108);

  char ret_param[CSDS_STRING_SIZE];
  parser_get_param_string(&parser, "Test:string", ret_param);
  assert(strcmp(ret_param, "test") == 0);

  /* Optional scalar reader (failed) */
  assert(parser_get_opt_param_char(&parser, "Test:test", 'd') == 'd');
  assert(parser_get_opt_param_int(&parser, "Test:test", 2) == 2);
  ftest = parser_get_opt_param_float(&parser, "Test:test", 0.1);
  assert(fabsf(ftest - 0.1f) < 1e-6);
  assert(parser_get_opt_param_double(&parser, "Test:test", 0.1) == 0.1);
  assert(parser_get_opt_param_longlong(&parser, "Test:test", 1) == 1);

  parser_get_opt_param_string(&parser, "Test:test", ret_param, "test1");
  assert(strcmp(ret_param, "test1") == 0);

  /* Optional scalar reader (pass) */
  assert(parser_get_opt_param_char(&parser, "Test:char", 'd') == 'c');
  assert(parser_get_opt_param_int(&parser, "Test:int", 1) == 10);
  ftest = parser_get_opt_param_float(&parser, "Test:float", 0.1);
  assert(fabsf(ftest - 0.2f) < 1e-6);
  assert(parser_get_opt_param_double(&parser, "Test:double", 0.1) == 0.2);
  assert(parser_get_opt_param_longlong(&parser, "Test:longlong", 1) == 1254108);

  parser_get_opt_param_string(&parser, "Test:string", ret_param, "test1");
  assert(strcmp(ret_param, "test") == 0);

  /* Array reader */
  const int nval = 3;
  char c[nval];
  int i[nval];
  float f[nval];
  double d[nval];
  long long l[nval];

  parser_get_param_char_array(&parser, "Test:char_array", nval, c);
  parser_get_param_int_array(&parser, "Test:int_array", nval, i);
  parser_get_param_float_array(&parser, "Test:float_array", nval, f);
  parser_get_param_double_array(&parser, "Test:double_array", nval, d);
  parser_get_param_longlong_array(&parser, "Test:longlong_array", nval, l);

  /* Check the values */
  assert(c[0] == 'a');
  assert(c[1] == 'b');
  assert(c[2] == 'c');

  for (int k = 0; k < nval; k++) {
    int m = k + 1;
    assert(i[k] == m);
    assert(fabsf(f[k] - m * 0.1f) < 1e-6);
    assert(fabs(d[k] - m * 0.1) < 1e-12);
    if (k == 0) {
      assert(l[k] == 1234567890);
    } else {
      assert(l[k] == m);
    }
  }

  /* Optional array reader (failed) */
  c[0] = 'e';
  c[1] = 'f';
  c[2] = 'g';
  i[0] = 3;
  i[1] = 4;
  i[2] = 5;
  f[0] = 0.3;
  f[1] = 0.4;
  f[2] = 0.5;
  d[0] = 0.3;
  d[1] = 0.4;
  d[2] = 0.5;
  l[0] = 9876543210;
  l[1] = 4;
  l[2] = 5;

  parser_get_opt_param_char_array(&parser, "Test:test", nval, c);
  parser_get_opt_param_int_array(&parser, "Test:test", nval, i);
  parser_get_opt_param_float_array(&parser, "Test:test", nval, f);
  parser_get_opt_param_double_array(&parser, "Test:test", nval, d);
  parser_get_opt_param_longlong_array(&parser, "Test:test", nval, l);

  /* Check the values */
  assert(c[0] == 'e');
  assert(c[1] == 'f');
  assert(c[2] == 'g');

  for (int k = 0; k < nval; k++) {
    int m = k + nval;
    assert(i[k] == m);
    assert(fabsf(f[k] - m * 0.1f) < 1e-6);
    assert(fabs(d[k] - m * 0.1) < 1e-12);
    if (k == 0) {
      assert(l[k] == 9876543210);
    } else {
      assert(l[k] == m);
    }
  }

  /* Optional array reader (success) */
  c[0] = 'e';
  c[1] = 'f';
  c[2] = 'g';

  i[0] = 3;
  i[1] = 4;
  i[2] = 5;

  f[0] = 0.3;
  f[1] = 0.4;
  f[2] = 0.5;

  d[0] = 0.3;
  d[1] = 0.4;
  d[2] = 0.5;

  l[0] = 9876543210;
  l[1] = 4;
  l[2] = 5;

  parser_get_opt_param_char_array(&parser, "Test:char_array", nval, c);
  parser_get_opt_param_int_array(&parser, "Test:int_array", nval, i);
  parser_get_opt_param_float_array(&parser, "Test:float_array", nval, f);
  parser_get_opt_param_double_array(&parser, "Test:double_array", nval, d);
  parser_get_opt_param_longlong_array(&parser, "Test:longlong_array", nval, l);

  /* Check the values */
  assert(c[0] == 'a');
  assert(c[1] == 'b');
  assert(c[2] == 'c');

  for (int k = 0; k < nval; k++) {
    int m = k + 1;
    assert(i[k] == m);
    assert(fabsf(f[k] - m * 0.1f) < 1e-6);
    assert(fabs(d[k] - m * 0.1) < 1e-12);
    if (k == 0) {
      assert(l[k] == 1234567890);
    } else {
      assert(l[k] == m);
    }
  }
}
