/*******************************************************************************
 * This file is part of CSDS.
 * Copyright (C) 2019 Loic Hausammann (loic.hausammann@epfl.ch)
 *                    Florian Cabot (florian.cabot@epfl.ch)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 ******************************************************************************/

/* Include configuration */
#include "config.h"

/* Standard include */
#include <stdio.h>
#include <stdlib.h>

/* Local include */
#include "csds_reader.h"
#include "generate_log.h"

#define number_parts 100

/**
 * This function test the csds in the Virtual Reality mode.
 * The idea is to simply read a snapshot at a given time and
 * then simply advance in time the particles.
 */
int main(void) {
  /* Create required structures. */
  char filename[200] = "testvr_0000.dump";

  /* Delete the index files */
  remove("testvr_0000_0000.index");
  remove("testvr_0000_0001.index");
  remove("testvr_0000_0002.index");
  remove("testvr_0000_0003.index");
  remove("testvr_0000_0004.index");

  /* Initialize the particles. */
  struct csds_part *parts;
  if ((parts = (struct csds_part *)malloc(sizeof(struct csds_part) *
                                          number_parts)) == NULL)
    error("Failed to allocate particles array.");

  /* Write a 'simulation' */
  generate_log(parts, number_parts, filename);

  /* Initialize the reader */
  struct csds_reader reader;
  char basename[200] = "testvr_0000";
  csds_reader_init(&reader, basename,
                   /* Verbose */ 2,
                   /* number_index */ 5,
                   /* restart */ 0,
                   /* use_cache */ 1);

  /* Read the time limits */
  double begin = csds_reader_get_time_begin(&reader);
  double end = csds_reader_get_time_end(&reader);

  /* Set the time */
  message("Time begin: %f end: %f", begin, end);
  csds_reader_set_time(&reader, begin);

  /* Create the variables for the number of particles */
  const int n_type = csds_type_count;
  uint64_t *n_parts = (uint64_t *)malloc(n_type * sizeof(uint64_t));
  int *read_types = (int *)malloc(n_type * sizeof(int));
  if (read_types == NULL || n_parts == NULL) {
    error("Failed to allocate arrays.");
  }

  /* Set the flags in order to read everything */
  for (int i = 0; i < n_type; i++) {
    read_types[i] = 1;
  }

  /* Get the number of particles */
  csds_reader_get_number_particles(&reader, n_parts, read_types);

  uint64_t n_tot = 0;
  for (int i = 0; i < n_type; i++) {
    n_tot += n_parts[i];
  }

  /* Allocate the particles memory */
  double *pos = (double *)malloc(n_tot * 3 * sizeof(double));
  long long *ids = (long long *)malloc(n_tot * sizeof(long long));

  /* Create the list of fields. */
  const int n_fields = 2;
  enum field_enum *required_fields =
      (enum field_enum *)malloc(n_fields * sizeof(enum field_enum));
  required_fields[0] = field_enum_coordinates;
  required_fields[1] = field_enum_particles_ids;

  /* Create the output */
  void **output = malloc(n_fields * sizeof(void *));
  output[0] = (void *)pos;
  output[1] = (void *)ids;

  /* Set the time of the next reading */
  csds_reader_set_time(&reader, begin);

  /* Read the next time */
  csds_reader_read_all_particles(&reader, begin, required_fields, n_fields,
                                 output, n_parts);

  /* Loop over time for a single particle */
  int part_ind = 0;
  for (double t = begin; t < end; t += (end - begin) / number_steps) {

    csds_reader_update_particles(&reader, t, required_fields, n_fields, output,
                                 n_parts);

    message("Particle %lli: %f %f %f %f", ids[part_ind], pos[3 * part_ind + 0],
            pos[3 * part_ind + 1], pos[3 * part_ind + 2], t);
  }

  /* Cleanup the memory */
  free(required_fields);
  free(ids);
  free(pos);
  free(parts);
  free(n_parts);
  free(read_types);
  csds_reader_free(&reader);
  free(output);
  return 0;
}
